package core;
/*
 * Implementation of the Line Tracer algorithm, modified for filament tracing.
 * 
 * Relevant research articles:
 * Carsten Gottschlich, Preda Mihailescu, and Axel Munk,
 * "Robust Orientation Field Estimation and Extrapolation Using Semilocal Line Sensors",
 * IEEE Transactions on Information Forensics and Security, vol. 4, no. 4, pp. 802-811, Dec. 2009.
 * dx.doi.org/10.1109/TIFS.2009.2033219
 *
 * Copyright 2009-2010 Carsten Gottschlich
 *           2011-2013 Julian Rüger
 *           2013-2015 Benjamin Eltzner
 *
 * This class is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This class is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this class; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.Point;

import ui.FilamentSensor;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.lang.Math;

class CurveTracer extends Tracer
{
    // m_directions should be a multiple of 8.
    private static final int m_direction_step = 3;
    private static final int m_directions = 360 / m_direction_step;
    private static final int m_orientations = m_directions / 2;
//    private static final int m_sense_depth = 20;
//    private static final int m_step_size = 10;
    private static final double m_tolerance = 0.3;

    private static boolean[][][] m_circle_masks;
    private static Point[][] m_lattice_sense, m_lattice_step;
    private static double[][] m_gauss_mask;

    private int m_max_width;

    private int[][] m_orientation_field, m_width_map;
    private List<SingleFilament> m_filament_list;

    CurveTracer()
    {
        if (m_circle_masks == null)
        {
            initCircleMasks(50);
        }
    }

    private void reset()
    {
        m_filament_list = new ArrayList<>();
        m_orientation_field = null;
        m_lattice_sense = null;
        m_lattice_step = null;
    }

    int[][] getOrientationField()
    {
        return m_orientation_field;
    }

    boolean[][] getPixelMask()
    {
        final int x_size = m_orientation_field.length,
                  y_size = m_orientation_field[0].length;
        boolean[][] out = new boolean[x_size][y_size];
        for (int x = 0; x < x_size; x++)
        {
            for (int y = 0; y < y_size; y++)
            {
                out[x][y] = m_orientation_field[x][y] > -1;
            }
        }
        return out;
    }

    int[][] getWidthMap()
    {
        return m_width_map;
    }

    List<SingleFilament> scanFilaments(boolean[][] bin, double tolerance, int min_length,
                                       int min_angle, int step)
    {
        reset();
        initSpokeLattices(step, Math.max(2*step, 20));
        // refresh map and get max_radius
        Map<Integer, List<Point>> width_list = getWidthMap(bin, tolerance);
        final int x_size = bin.length,
                  y_size = bin[0].length;
        // TODO: Mask sigma = 0.5 * step?
        m_gauss_mask = Calc.gaussianQuarterMask((3 * m_max_width + 1) / 2, 0.5 * m_max_width);

        if (m_orientation_field == null)
        {
            // initialize orientation field
            m_orientation_field = new int[x_size][y_size];

            for (int x = 0; x < x_size; x++)
            {
                for (int y = 0; y < y_size; y++)
                {
                    m_orientation_field[x][y] = -1;
                }
            }
        }

        Map<Integer, List<Gatherer>> scored_lines = new HashMap<>();
        int max_score = 0, line_counter = 0;
        FilamentSensor.debugMessage("Start tracing, max width: " + m_max_width);
        for (int width = m_max_width; width > 0; width--)
        {
            FilamentSensor.debugMessage("Tracing filaments, width " + width + "... ");

            List<Point> points = width_list.get(width);
            if (points == null)
            {
                continue;
            }
            for (Point p : points)
            {
                if (m_orientation_field[p.x][p.y] > -1)
                {
                    continue;
                }

                // Test lines for all directions.
                Gatherer[] gatherers = new Gatherer[m_directions];
                int[] scores = new int[m_directions];
                for (int dir = 0; dir < m_directions; dir++)
                {
                    gatherers[dir] = new Gatherer(p.x, p.y, width, dir);
                    scores[dir] = gatherers[dir].sense(m_width_map);
                }

                // Identify highest scoring pair of directions.
                int[] best_pair = null;
                int highscore = 0;
                // Opposite directions
                for (int i = 0; i < m_orientations; i++)
                {
                    int score = scores[i] + scores[i + m_orientations];
                    if (score > highscore)
                    {
                        highscore = score;
                        best_pair = new int[]{i, i + m_orientations};
                    }
                }
                // Direction change in start point.
                for (int i = 0; i < m_directions; i++)
                {
                    int j = (i + m_orientations - 1) % m_directions;
                    int score = scores[i] + scores[j];
                    if (score > highscore)
                    {
                        highscore = score;
                        best_pair = new int[]{i, j};
                    }
                }

                // Gather line by following best initial orientations.
                gatherers = new Gatherer[]{gatherers[best_pair[0]], gatherers[best_pair[1]]};
                for (int i : new int[]{0,1})
                {
                    gatherers[i].gather();//m_width_map);
                }
                Gatherer line = gatherers[0];
                line.join(gatherers[1]);

                // Add line to map.
                int score = line.score();
                if (!scored_lines.containsKey(score))
                {
                    scored_lines.put(score, new ArrayList<Gatherer>());
                }
                if (score > max_score)
                {
                    max_score = score;
                }
                scored_lines.get(score).add(line);
                if (++line_counter >= 10000)
                {
                    FilamentSensor.debugMessage(line_counter + " lines. Maximal score: " + max_score);
                    line_counter = cleanUpLines(scored_lines, max_score, min_length, min_angle, bin);
                    for (int m = max_score; m >= 0; m--)
                    {
                        if (scored_lines.get(m) != null)
                        {
                            max_score = m;
                            break;
                        }
                    }
                    FilamentSensor.debugMessage(line_counter + " lines. Maximal score: " + max_score);
                }
            }
        }
        markLines(scored_lines, max_score, min_length, min_angle, bin);

        return m_filament_list;
    }

    int calcWidthMap(boolean[][] bin, double tolerance)
    {
        return Collections.max(getWidthMap(bin, tolerance).keySet());
    }

    /**
     * For each white point in the binary image the width map entry
     * is the highest diameter at which the ratio of black pixels
     * to all pixels exceeds the tolerance, zero everywhere else.
     */
    private Map<Integer, List<Point>> getWidthMap(boolean[][] bin, double tolerance)
    {
        m_max_width = 0;

        int x_size = bin.length,
            y_size = bin[0].length;
        m_width_map = new int[x_size][y_size];
        Map<Integer, List<Point>> width_list = new HashMap<>();

        for (int x = 0; x < x_size; x++)
        {
            for (int y = 0; y < y_size; y++)
            {
                if (bin[x][y]) // false is white. If black, m_width_map[i][j] remains 0.
                {
                    continue;
                }
                int diameter = 1, count = 0, misses = 0;
                double ratio = 0.0;

                while (ratio <= tolerance)
                {
                    diameter++;
                    if (diameter > m_max_width)
                    {
                        m_max_width = diameter;
                    }

                    if (diameter >= m_circle_masks.length)
                    {
                        initCircleMasks(m_circle_masks.length * 2);
                        return getWidthMap(bin, tolerance);
                    }

                    boolean[][] mask = m_circle_masks[diameter];
                    int range = mask.length / 2;

                    for (int dx = -range; dx < range + 1; dx++)
                    {
                        if (x + dx < 0 || x + dx > x_size - 1)
                        {
                            continue;
                        }

                        for (int dy = -range; dy < range + 1; dy++)
                        {
                            if (!mask[range + dx][range + dy] || y + dy < 0 || y + dy > y_size - 1)
                            {
                                continue;
                            }
                            count++;
                            // Count white points as misses.
                            if (bin[x + dx][y + dy])
                            {
                                misses++;
                            }
                        }
                    }

                    // Ratio of misses.
                    ratio = (double) misses / (double) count;
                }
                // Radius at which black pixel ratio exceeded the tolerance.
                m_width_map[x][y] = diameter - 1;
                if (!width_list.containsKey(diameter - 1))
                {
                    width_list.put(diameter - 1, new ArrayList<Point>());
                }
                width_list.get(diameter -1).add(new Point(x, y));
            }
        }
        return width_list;
    }

    int[][] makeFingerprint(double axis)
    {
        if (m_orientation_field == null)
        {
            return null;
        }
        // Make List of pixels with orientation
        List<int[]> points = new ArrayList<>();
        for (int x = 0; x < m_orientation_field.length; x++)
        {
            for (int y = 0; y < m_orientation_field[0].length; y++)
            {
                if (m_orientation_field[x][y] >= 0)
                {
                    points.add(new int[]{x, y, m_orientation_field[x][y]});
                }
            }
        }
        // Make fingerprint image
        int[][] fingerprint = new int[101][91];
        for (int[] p : points)
        {
            for (int[] q : points)
            {
                if (q.equals(p))
                {
                    break;
                }
                int length = (int) Math.round(100 * Math.sqrt((p[0] - q[0]) * (p[0] - q[0]) +
                                                              (p[1] - q[1]) * (p[1] - q[1])) / axis);
                if (length > 100)
                {
                    continue;
                }
                int ori_diff = Math.abs(90 - Math.abs(p[2] - q[2]));
                fingerprint[length][ori_diff]++;
            }
        }
        return fingerprint;
    }

    private void markLines(Map<Integer, List<Gatherer>> scored_lines, int max_score,
                           int min_length, int min_angle, boolean[][] binary_image)
    {
        // Run through the lines, longest first, and mark them.
        for (int score = max_score; score >= 0; score--)
        {
            if (!scored_lines.containsKey(score))
            {
                continue;
            }
            List<Gatherer> list = scored_lines.get(score);

            for (int j = list.size() - 1; j >= 0; j--)
            {
                Gatherer line = list.get(j);
               
                int old_length = line.length();
                if (old_length >= min_length)
                {
                    drawLine(line, min_angle, binary_image, true);
                }

                // Shortened lines may be 3/4 the minimal length.
                if (line.length() < old_length && line.length() >= 0.75 * min_length)
                {
                    int new_score = line.score();
                    if (scored_lines.get(new_score) == null)
                    {
                        scored_lines.put(new_score, new ArrayList<Gatherer>());
                    }

                    scored_lines.get(new_score).add(line);
                }
                else
                {
                    line = null;
                }
            }
        }
    }

    private int cleanUpLines(Map<Integer, List<Gatherer>> scored_lines, int max_score,
                             int min_length, int min_angle, boolean[][] binary_image)
    {
        // Run through the longest lines and mark them.
        if (!scored_lines.containsKey(max_score))
        {
            return -1;
        }
        List<Gatherer> list = scored_lines.remove(max_score);

        for (int j = list.size() - 1; j >= 0; j--)
        {
            Gatherer line = list.remove(j);

            int old_length = line.length();
            if (old_length >= min_length)
            {
                drawLine(line, min_angle, binary_image, true);
            }
            if (line.length() < old_length)
            {
                // Draw lines that were only shortened by zero-pixels.
                if (line.score() == max_score)
                {
                    drawLine(line, min_angle, binary_image, true);
                }
                // Really shortened lines may be 3/4 the minimal length.
                else if (line.length() < old_length && line.length() >= 0.75 * min_length)
                {
                    int new_score = line.score();
                    if (scored_lines.get(new_score) == null)
                    {
                        scored_lines.put(new_score, new ArrayList<Gatherer>());
                    }

                    scored_lines.get(new_score).add(line);
                }
            }
        }

        // Cleanup lists of shorter lines.
        int line_counter = 0;
        for (int score = max_score - 1; score >= 0; score--)
        {
            if (!scored_lines.containsKey(score))
            {
                continue;
            }
            list = scored_lines.remove(score);

            for (int j = list.size() - 1; j >= 0; j--)
            {
                Gatherer line = list.remove(j);

                int old_length = line.length();
                if (old_length >= min_length)
                {
                    drawLine(line, min_angle, binary_image, false);
                }

                // Shortened lines may be 3/4 the minimal length.
                if (line.length() >= 0.75 * min_length)
                {
                    int new_score = line.score();
                    if (scored_lines.get(new_score) == null)
                    {
                        scored_lines.put(new_score, new ArrayList<Gatherer>());
                    }

                    scored_lines.get(new_score).add(line);
                    line_counter++;
                }
            }
        }
        return line_counter;
    }

    private void drawLine(Gatherer line, int min_angle, boolean[][] binary_image, boolean draw)
    {
        List<int[]> points = line.points();
        final boolean xcross = line.xcross();

        // Shorten line from both ends.
        boolean shortened = false;
        while (!points.isEmpty())
        {
            int[] p0 = points.get(0);
            if (m_orientation_field[p0[0]][p0[1]] < 0 && p0[2] > 0)
            {
                break;
            }
            line.remove(0);
            shortened = true;
        }
        while (!points.isEmpty())
        {
            int end = points.size()-1;
            int[] p_end = points.get(end);
            if (m_orientation_field[p_end[0]][p_end[1]] < 0 && p_end[2] > 0)
            {
                break;
            }
            line.remove(end);
            shortened = true;
        }

        // Return if line was shortened.
        if (shortened)
        {
            line.rescore();
            return;
        }

        // Check that less than m_tolerance of the line is already covered.
        // FIXME: Restrict fraction of line that may be covered by non-conflicting orientations?
        int already_present = 0;
        for (int[] p : line.points())
        {
            int ori = (p[3] * m_direction_step) % 180,
                pre_ori = m_orientation_field[p[0]][p[1]];
            if (pre_ori != -1)
            {
                if (angle(ori, pre_ori) < min_angle)
                {
                    already_present++;
                }
            }
        }
        if (already_present > m_tolerance * line.length())
        {
            line.abandon();
            return;
        }

        // Return if we are only on a shortening run.
        if (!draw)
        {
            line.rescore();
            return;
        }

        // Make binary line mask using point positions, widths and binary image.
        int x_min = Integer.MAX_VALUE, x_max = 0, y_min = Integer.MAX_VALUE, y_max = 0, w_max = 0;
        for (int[] p : line.points())
        {
            if (p[0] < x_min) { x_min = p[0]; }
            if (p[0] > x_max) { x_max = p[0]; }
            if (p[1] < y_min) { y_min = p[1]; }
            if (p[1] > y_max) { y_max = p[1]; }
            if (p[2] > w_max) { w_max = p[2]; }
        }
        x_min -= w_max; x_max += w_max; y_min -= w_max; y_max += w_max;

        boolean[][] canvas = new boolean[x_max - x_min + 1][y_max - y_min + 1];
        for (int[] p : line.points())
        {
            if (p[2] == 0)
            {
                continue;
            }
            boolean[][] circle = m_circle_masks[p[2]+1];
            int shift = circle.length / 2;
            int dx1 = p[0] - shift, dx2 = dx1 - x_min, dy1 = p[1] - shift, dy2 = dy1 - y_min;
            for (int x = 0; x < circle.length; x++)
            {
                for (int y = 0; y < circle.length; y++)
                {
                    canvas[x + dx2][y + dy2] = canvas[x + dx2][y + dy2] ||
                                               circle[x][y] && !binary_image[x + dx1][y + dy1];
                }
            }
        }

        // Make line mask with orientations via Gaussian weighted mean over centerline pixels.
        int[][] ori_canvas = new int[x_max - x_min + 1][y_max - y_min + 1];

        int shift = m_gauss_mask.length - 1;
        for (int x = 0; x < canvas.length; x++)
        {
            for (int y = 0; y < canvas[x].length; y++)
            {
                if (canvas[x][y])
                {
                    int qx = x + x_min, qy = y + y_min, turn = m_orientations / 2;
                    double gauss_sum = 0, sum = 0;
                    
                    if (xcross)
                    {
                        for (int[] p : line.points())
                        {
                            int dx = Math.abs(qx - p[0]), dy = Math.abs(qy - p[1]);
                            if (dx > shift || dy > shift)
                            {
                                continue;
                            }
                            gauss_sum += m_gauss_mask[dx][dy] * ((p[3] + turn) % m_orientations);
                            sum += m_gauss_mask[dx][dy];
                        }
                        ori_canvas[x][y] = ((int)Math.round((gauss_sum / sum + m_orientations - turn) *
                                                            m_direction_step) % 180) + 1;
                    }
                    else
                    {
                        for (int[] p : line.points())
                        {
                            int dx = Math.abs(qx - p[0]), dy = Math.abs(qy - p[1]);
                            if (dx > shift || dy > shift)
                            {
                                continue;
                            }
                            gauss_sum += m_gauss_mask[dx][dy] * p[3];
                            sum += m_gauss_mask[dx][dy];
                        }
                        ori_canvas[x][y] = ((int)Math.round(gauss_sum * m_direction_step /
                                                            sum) % 180) + 1;
                    }
                }
            }
        }

        // Draw line mask to orientation field.
        for (int x = 0; x < ori_canvas.length; x++)
        {
            for (int y = 0; y < ori_canvas[x].length; y++)
            {
                if (ori_canvas[x][y] > 0)
                {
                    m_orientation_field[x+x_min][y+y_min] = ori_canvas[x][y] - 1;
                }
                ori_canvas[x][y] -= 1;
            }
        }
        m_filament_list.add(line.asFilament());
    }

    private static int angle(int a, int b)
    {
        int diff = a - b;
        diff = (diff < 0 ? -diff : diff);
        diff = (diff > 90 ? 180 - diff : diff);
        return diff;
    }

    private static void initCircleMasks(int size)
    {
        m_circle_masks = new boolean[size+1][][];
        for (int i = 1; i <= size; i++)
        {
            m_circle_masks[i] = Calc.circleMask(i);
        }
    }

    // This method defines a "square" lattice.
    private static void initSpokeLattices(int step_size, int sense_depth)
    {
        Point[][] pre_lattice = new Point[m_directions][sense_depth];
        
        final int eighth = m_directions / 8;

        for (int j = 0; j < sense_depth; j++)
        {
            pre_lattice[0][j] = new Point(j + 1, 0);
            pre_lattice[2 * eighth][j] = new Point(0, j + 1);
            pre_lattice[4 * eighth][j] = new Point(-j - 1, 0);
            pre_lattice[6 * eighth][j] = new Point(0, -j - 1);

            pre_lattice[1 * eighth][j] = new Point(j + 1, j + 1);
            pre_lattice[3 * eighth][j] = new Point(-j - 1, j + 1);
            pre_lattice[5 * eighth][j] = new Point(-j - 1, -j - 1);
            pre_lattice[7 * eighth][j] = new Point(j + 1, -j - 1);
        }

        for (int i = 1; i < eighth; i++)
        {

            double slope = Math.tan(Math.PI * i / ((double) m_orientations));

            for (int j = 0; j < sense_depth; j++)
            {
                int x = j + 1, y = (int) (Math.round(x * slope));

                pre_lattice[i][j] = new Point(x, y);

                pre_lattice[(2 * eighth) - i][j] = new Point(y, x);
                pre_lattice[(2 * eighth) + i][j] = new Point(-y, x);

                pre_lattice[(4 * eighth) - i][j] = new Point(-x, y);
                pre_lattice[(4 * eighth) + i][j] = new Point(-x, -y);

                pre_lattice[(6 * eighth) - i][j] = new Point(-y, -x);
                pre_lattice[(6 * eighth) + i][j] = new Point(y, -x);

                pre_lattice[(8 * eighth) - i][j] = new Point(x, -y);

            }
        }

        Point[][][] lattices = new Point[2][m_directions][];
        for (int i = 0; i < 2; i++)
        {
            int depth = (i == 0 ? sense_depth : step_size);
            for (int j = 0; j < m_directions; j++)
            {
                List<Point> ray = new ArrayList<>();
                double old_distance = 0, distance = 0;
                for (int k = 0; k < sense_depth; k++)
                {
                    Point p = pre_lattice[j][k];
                    distance = Math.sqrt(p.x * p.x + p.y * p.y);
                    if (Math.abs(distance - depth) > Math.abs(old_distance - depth))
                    {
                        break;
                    }
                    old_distance = distance;

                    ray.add(p);
                }
                lattices[i][j] = ray.toArray(new Point[0]);
            }
        }
        m_lattice_sense = lattices[0];
        m_lattice_step = lattices[1];
    }

    private class Gatherer
    {
        private List<int[]> m_points;
        private int[] m_start;
        private int m_score;
        private boolean m_xcross;
        private int m_direction;
        
        private Gatherer(int x_pos, int y_pos, int width, int direction)
        {
            m_start = new int[]{x_pos, y_pos, width, direction % m_orientations};
            m_direction = direction;
            m_xcross = false;
        }

        private int sense(int[][] width_map)
        {
            m_points = new ArrayList<>();
            m_points.add(m_start);
            m_score = m_start[2];
            int dir = m_direction;
            Point[] ray = m_lattice_sense[dir];
            int step_length = m_lattice_step[dir].length;
            int x = m_start[0], y = m_start[1], score = m_score;
            int i = 0;
            boolean one_black_pixel = false;
            for (Point p : ray)
            {
                int px = x + p.x, py = y + p.y, w = width_map[px][py];
                // Accept at most one 0 in a row.
                if (w == 0)
                {
                    if (one_black_pixel)
                    {
                        break;
                    }
                    else
                    {
                        one_black_pixel = true;
                    }
                }
                else if (one_black_pixel)
                {
                    one_black_pixel = false;
                }
                if (i < step_length)
                {
                    m_points.add(new int[]{px, py, w, dir % m_orientations});
                    m_score += w;
                }
                score += w;
                i++;
            }
            // Eliminate empty points from end of list.
            for (i = length(); i >= 0; i--)
            {
                int w = m_points.get(i)[2];
                if (w > 0 && m_score >= m_points.size())
                {    
                    break;
                }
                m_points.remove(i);
                m_score -= w;
            }
            return score;
        }

        private void gather()//int[][] width_map)
        {
            gather(m_points.get(m_points.size()-1));//width_map, 
        }

        private void gather(int[] start)//int[][] width_map, 
        {
            int dir = m_direction;

            // Current direction and both neighboring directions.
            int[] dirs = new int[]{dir, (dir+1) % m_directions,
                                   (dir+m_directions-1) % m_directions};
            double best_score = 0, best_small_score = 0;
            int full_score = 0;
            List<int[]> best_path = new ArrayList<>(), path = new ArrayList<>();
            for (int d : dirs)
            {
                path.clear();
                Point[] ray = m_lattice_sense[d];
                int step_length = m_lattice_step[d].length;
                int x = start[0], y = start[1];
                int small_score = 0;
                double mean_score = 0, mean_small_score = 0;
                int i = 0;
                boolean one_black_pixel = false;
                for (Point p : ray)
                {
                    int px = x + p.x, py = y + p.y, w = m_width_map[px][py];
                    // Accept at most one 0 in a row.
                    if (w == 0)
                    {
                        if (one_black_pixel)
                        {
                            break;
                        }
                        else
                        {
                            one_black_pixel = true;
                        }
                    }
                    else if (one_black_pixel)
                    {
                        one_black_pixel = false;
                    }
                    
                    if (i < step_length)
                    {
                        path.add(new int[]{px, py, w, d % m_orientations});
                        small_score += w;
                    }
                    mean_score += w;
                    i++;
                }
                mean_score /= (double)ray.length;
                mean_small_score = (double)small_score / (double)step_length;

                if ((mean_score > best_score && mean_small_score > 1) ||
                    (mean_small_score <= 1 && mean_small_score > best_small_score))
                {
                    best_score = mean_score;
                    best_small_score = mean_small_score;
                    best_path.clear();
                    best_path.addAll(path);
                    full_score = small_score;
                    m_direction = d;
                }
            }
            // Set flag, if line orientation crosses x-axis.
            if (!best_path.isEmpty())
            {
                int new_ori = best_path.get(0)[3];
                if (Math.abs(new_ori - (dir % m_orientations)) > 1)
                {
                    m_xcross = true;
                }
            }
            if (best_small_score > 1)
            {
                m_points.addAll(best_path);
                m_score += full_score;
                gather(m_points.get(m_points.size()-1));
            }
            else
            {
                // Eliminate empty points from end of list.
                for (int i = best_path.size() -1; i >= 0; i--)
                {
                    if (best_path.get(i)[2] > 0)
                    {
                        break;
                    }
                    best_path.remove(i);
                }
                if (!best_path.isEmpty())
                {
                    int score = 0;
                    for (int[] p : best_path)
                    {
                        score += p[2];
                    }
                    if (score >= best_path.size())
                    {
                        m_points.addAll(best_path);
                        m_score += score;
                    }
                    else
                    {
                        List<int[]> rest = new ArrayList<>();

                        score = 0;
                        for (int[] p : best_path)
                        {
                            if (p[2] == 0)
                            {
                                break;
                            }
                            rest.add(p);
                            score += p[2];
                        }
                        m_points.addAll(rest);
                        m_score += score;
                    }
                }
            }
        }

        private void rescore()
        {
            m_score = 0;
            for (int[] p : m_points)
            {
                m_score += p[2];
            }
        }

        private int score()
        {
            return m_score;
        }

        private int length()
        {
            // TODO: Return real length?
            return m_points.size() - 1;
        }

        private List<int[]> points()
        {
            return m_points;
        }

        private CurvedFilament asFilament()
        {
            List<Point> bend_points = new ArrayList<>();
            int[] p0 = m_points.get(0);
            bend_points.add(new Point(p0[0], p0[1]));
            int ori = p0[3], abs_curv = 0, signed_curv = 0;
            for (int[] p : m_points)
            {
                if (p[3] != ori)
                {
                    abs_curv += m_direction_step;
                    bend_points.add(new Point(p[0], p[1]));
                    if (p[3] == (ori + 1) % m_orientations)
                    {
                        signed_curv += m_direction_step;
                    }
                    else if (ori == (p[3] + 1) % m_orientations)
                    {
                        signed_curv -= m_direction_step;
                    }
                    ori = p[3];
                }
            }
            int[] tmp = m_points.get(m_points.size() -1);
            Point p_last = new Point(tmp[0], tmp[1]);
            if (!bend_points.get(bend_points.size() - 1).equals(p_last))
            {
                bend_points.add(p_last);
            }

            return new CurvedFilament(bend_points, Math.round(m_score * Const.MF / (double)m_points.size()),
                                      signed_curv, abs_curv, true);
        }

        private boolean xcross()
        {
            return m_xcross;
        }

        private void abandon()
        {
            m_points = new ArrayList<>();
            m_start = new int[4];
            m_score = 0;
        }

        private int[] remove(int index)
        {
            return m_points.remove(index);
        }

        private void join(Gatherer other)
        {
            List<int[]> other_points = other.points();
            if (Math.abs(other_points.get(0)[3] - m_points.get(0)[3]) > 1 || other.m_xcross)
            {
                m_xcross = true;
            }
            List<int[]> new_list = new ArrayList<>();
            for (int i = other_points.size() - 1; i > 0; i--)
            {
                new_list.add(other_points.get(i));
            }
            new_list.addAll(m_points);
            m_points = new_list;
            // add scores, correcting for starting point.
            m_score += other.score() - other_points.get(0)[2];
        }
    }

}