package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Point;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import ui.FilamentSensor;

public class CurvedFilament extends SingleFilament
{
    private List<Point> m_points;
    private boolean m_keep;

    private Point m_center; // center is scaled by a factor of 10.
    // length, angle, width and curvatures are longs that are scaled
    // by 1,000,000 to avoid the problems of floating point arithmetics.
    private long m_length;
    private long m_orientation; // 0 =< angle < 180,000,000
    private long m_mean_signed_curvature;
    private long m_mean_absolute_curvature;
    private long m_width;

    public CurvedFilament(List<Point> points, long width, int signed_curvature,
                          int absolute_curvature, boolean keep)
    {
        m_points = points;
        m_points = straighten();
        Point a = m_points.get(0), b = m_points.get(m_points.size() - 1);
        m_length = 0;
        long[] partial_lengths = new long[m_points.size()];
        for (int i = 0; i < m_points.size() - 1; i++)
        {
            m_length += Calc.distanceM(m_points.get(i), m_points.get(i+1));
            partial_lengths[i+1] = m_length;
        }
        m_center = calcCenter(partial_lengths, m_points);
        m_orientation = Calc.orientation(a, b);
        m_width = width;
        long[] curvatures = calcCurvatures(m_points, m_length);
        m_mean_signed_curvature = curvatures[0];
        m_mean_absolute_curvature = curvatures[1];
        m_keep = keep;
    }

    public CurvedFilament(List<Point> points, long width)
    {
        m_points = points;
        Point a = points.get(0), b = points.get(points.size() - 1);
        m_length = 0;
        long[] partial_lengths = new long[points.size()];
        for (int i = 0; i < points.size() - 1; i++)
        {
            m_length += Calc.distanceM(points.get(i), points.get(i+1));
            partial_lengths[i+1] = m_length;
        }
        m_center = calcCenter(partial_lengths, points);
        m_orientation = Calc.orientation(a, b);
        m_width = width;
        long[] curvatures = calcCurvatures(points, m_length);
        m_mean_signed_curvature = curvatures[0];
        m_mean_absolute_curvature = curvatures[1];
        m_keep = true;
    }
    
    public CurvedFilament(List<Point> points)
    {
        this(points, Const.M);
    }

    public CurvedFilament(Element element)
    {
        m_points = pointsFromElement(element);
        m_center = new Point(Integer.parseInt(element.getAttribute("center_x")),
                             Integer.parseInt(element.getAttribute("center_y")));
        m_length = Long.parseLong(element.getAttribute("length"));
        m_orientation = Long.parseLong(element.getAttribute("angle"));
        m_width = Long.parseLong(element.getAttribute("width"));
        m_mean_signed_curvature = zeroOld(element.getAttribute("signed_curv"));
        m_mean_absolute_curvature = zeroOld(element.getAttribute("abs_curv"));
        m_keep = element.getAttribute("keep").equals("yes");
    }

    public CurvedFilament(SingleFilament fil)
    {
        m_points = new ArrayList<>();
        if (fil instanceof Filament)
        {
            Filament filament = (Filament)fil;
            m_points.add(filament.getHead());
            m_points.add(filament.getTail());
        }
        else if (fil instanceof CurvedFilament)
        {
            CurvedFilament filament = (CurvedFilament)fil;
            for (Point p : filament.m_points)
            {
                m_points.add(new Point(p));
            }
        }

        m_center = fil.getCenter();
        m_length = fil.getLength();
        m_orientation = fil.getOrientation();
        m_width = fil.getWidth();
        m_mean_signed_curvature = fil.getSignedCurvature();
        m_mean_absolute_curvature = fil.getAbsoluteCurvature();
        m_keep = fil.getKeep();
    }

    private CurvedFilament(CurvedFilament that)
    {
        m_points = new ArrayList<>();
        for (Point p : that.m_points)
        {
            m_points.add(new Point(p));
        }
        m_center = new Point(that.m_center);
        m_length = that.m_length;
        m_orientation = that.m_orientation;
        m_width = that.m_width;
        m_mean_signed_curvature = that.m_mean_signed_curvature;
        m_mean_absolute_curvature = that.m_mean_absolute_curvature;
        m_keep = that.m_keep;
    }

    @Override
    public ClusterableFilament makeClusterable(int time)
    {
        return new ClusterableFilament(this, time);
    }

    private static long zeroOld(String string)
    {
        return (string == "" ? 0 : Long.parseLong(string));
    }

    private static List<Point> pointsFromElement(Element element)
    {
        List<Point> points = new ArrayList<>();
        if (element.getAttribute("start_x").isEmpty())
        {
            int i = 0;
            while (true)
            {
                String x = element.getAttribute("point" + i + "_x");
                if (x.isEmpty())
                {
                    break;
                }
                points.add(new Point(Integer.parseInt(x),
                                     Integer.parseInt(element.getAttribute("point" + i + "_y"))));
                i++;
            }
        }
        else
        {
            points.add(new Point(Integer.parseInt(element.getAttribute("start_x")),
                                 Integer.parseInt(element.getAttribute("start_y"))));
            points.add(new Point(Integer.parseInt(element.getAttribute("end_x")),
                                 Integer.parseInt(element.getAttribute("end_y"))));
        }
        
        return points;
    }

    // Getters

    public List<Point> getPoints()
    {
        return m_points;
    }

    @Override
    public List<Filament> splitIntoLinearPieces()
    {
        straighten();
        List<Filament> out = new ArrayList<>();
        if (m_points.size() < 2)
        {
            return out;
        }
        if (m_points.size() == 2)
        {
            out.add(new Filament(m_points.get(0), m_points.get(1), m_width, m_keep));
            return out;
        }

        out.add(new Filament(m_points.get(0), m_points.get(1), m_width, m_keep));
        for (int i = 1; i < m_points.size() -1; i++)
        {
            Point p1 = m_points.get(i-1), q1 = m_points.get(i);

            // Avoid "touching end points".
            double vx = q1.x - p1.x, vy = q1.y - p1.y,
                   v = Math.max(Math.abs(vx), Math.abs(vy));
            int dx = (int)Math.round(vx / v), dy = (int)Math.round(vy / v);
            
            out.add(new Filament(new Point(q1.x + dx, q1.y + dy), m_points.get(i+1),
                                 m_width, m_keep));
        }

        return out;
    }

    /*
    public Point getHead()
    {
        return new Point(m_points.get(0));
    }

    public Point getTail()
    {
        return new Point(m_points.get(m_points.size() - 1));
    }*/

    @Override
    public Point getCenter()
    {
        return new Point(m_center);
    }

    @Override
    public boolean getKeep()
    {
        return m_keep;
    }

    @Override
    public long getLength()
    {
        return m_length;
    }

    public double getMass()
    {
        return m_length * m_width / (Const.MF * Const.MF);
    }

    @Override
    public long getOrientation()
    {
        return m_orientation;
    }

    @Override
    public long getWidth()
    {
        return m_width;
    }

    @Override
    public long getSignedCurvature()
    {
        return m_mean_signed_curvature;
    }

    @Override
    public long getAbsoluteCurvature()
    {
        return m_mean_absolute_curvature;
    }

    @Override
    public void setKeep(boolean keep)
    {
        this.m_keep = keep;
    }

    @Override
    public void setWidth(long width)
    {
        this.m_width = width;
    }

    @Override
    public void invert()
    {
        Collections.reverse(m_points);
    }

    @Override
    public void remedyOldAngles()
    {
        FilamentSensor.debugError("This is not longer supported in FilamentSensor 0.2. Please use an older version of FilamentSensor.");
    }

    @Override
    public Element toXML(Document document, int chain_number, int filament_number)
    {
        Element element = document.createElement("curved_filament" + chain_number + "_" + filament_number);

        int i = 0;
        for (Point p : m_points)
        {
            element.setAttribute("point" + i + "_x", "" + p.x);
            element.setAttribute("point" + i + "_y", "" + p.y);
            i++;
        }
        element.setAttribute("center_x", "" + m_center.x);
        element.setAttribute("center_y", "" + m_center.y);
        element.setAttribute("length", "" + m_length);
        element.setAttribute("angle", "" + m_orientation);
        element.setAttribute("width", "" + m_width);
        element.setAttribute("signed_curv", "" + m_mean_signed_curvature);
        element.setAttribute("abs_curv", "" + m_mean_absolute_curvature);
        element.setAttribute("keep", m_keep ? "yes" : "no");

        return element;
    }

    // Note that the coordinates of the center are scaled by a factor of 10
    // in comparison to the coordinates of the corner points.
    private static Point calcCenter(long[] partial_lengths, List<Point> points)
    {
        if (points.size() == 2)
        {
            Point a = points.get(0), b = points.get(1);
            return new Point((a.x + b.x) * 5, (a.y + b.y) * 5);
        }
        long full_length = partial_lengths[partial_lengths.length - 1];
        int i;
        for (i = 0; i < partial_lengths.length; i++)
        {
            if (2 * partial_lengths[i] > full_length)
            {
                break;
            }
        }
        Point a = points.get(i-1), b = points.get(i);
        double ratio = (0.5 * full_length - partial_lengths[i-1]) /
                       (partial_lengths[i] - partial_lengths[i-1]);
        return new Point((int)Math.round(10 * (a.x + ratio * (b.x - a.x))),
                         (int)Math.round(10 * (a.y + ratio * (b.y - a.y))));
    }

    private static long[] calcCurvatures(List<Point> points, long length)
    {
        if (points.size() < 3)
        {
            return new long[]{0,0};
        }
        double[][] directions = new double[points.size() - 1][2];
        for (int i = 0; i < points.size() - 1; i++)
        {
            Point a = points.get(i), b = points.get(i+1);
            double distance = a.distance(b);
            directions[i][0] = (b.x - a.x) / distance;
            directions[i][1] = (b.y - a.y) / distance;
        }
        double abs_sum = 0;
        int n = directions.length - 1;
        for (int i = 0; i < n; i++)
        {
            double angle = Math.acos(Math.min(1, directions[i][0] * directions[i+1][0] +
                                                 directions[i][1] * directions[i+1][1])) / Const.RAD;
            
            abs_sum += Math.abs(angle);
        }

        double cosine = (directions[0][0] * directions[n][0] +
                         directions[0][1] * directions[n][1]),
               sine = (directions[0][0] * directions[n][1] -
                      directions[0][1] * directions[n][0]),
               angle = Math.atan2(sine, cosine) / Const.RAD,
               factor = Const.MF * Const.MF / (double) length;

        return new long[]{Math.round(factor * angle), Math.round(factor * abs_sum)};
    }

    private List<Point> straighten()
    {
        if (m_points.size() < 3)
        {
            return m_points;
        }

        List<Integer> split_list = new ArrayList<>();
        split_list.add(0);
        int end = m_points.size()-1;
        split_list = bestSplits(0, end, split_list);
        split_list.add(end);
        Collections.sort(split_list);

        List<Point> out = new ArrayList<>();
        for (int i : split_list)
        {
            out.add(m_points.get(i));
        }
        return out;
    }

    private List<Integer> bestSplits (int start, int end, List<Integer> list)
    {
        if (maxDistance(start, end) < 1)
        {
            return list;
        }

        double max_length = 0;
        boolean which_end = false; // false = start, true = end
        int best_split = start;
        for (int i = start+1; i < end; i++)
        {
            double dist1 = maxDistance(start, i), dist2 = maxDistance(i, end);
            if (dist1 < 1 && dist2 < 1)
            {
                list.add(i);
                return list;
            }
            if (dist1 < 1)
            {
                double length = m_points.get(start).distanceSq(m_points.get(i));
                if (length > max_length)
                {
                    best_split = i;
                    max_length = length;
                    which_end = false;
                }
            }
            if (dist2 < 1)
            {
                double length = m_points.get(i).distanceSq(m_points.get(end));
                if (length > max_length)
                {
                    best_split = i;
                    max_length = length;
                    which_end = true;
                }
            }
        }
        list.add(best_split);
        if (which_end)
        {
            return bestSplits(start, best_split, list);
        }
        return bestSplits(best_split, end, list);
    }

    private double maxDistance (int start, int end)
    {
        if (end < start + 2)
        {
            return 0;
        }
        Point p1 = m_points.get(start), q1 = m_points.get(end);
        double max_dist = 0;
        for (int i = start + 1; i < end; i++)
        {
            double dist = pointLineDistanceSquare(p1, q1, m_points.get(i));
            if (dist > max_dist)
            {
                max_dist = dist;
            }
        }
        return max_dist;
    }

    @Override
    public Object clone()
    {
        return new CurvedFilament(this);
    }

    @Override
    public boolean equals(Object object)
    {
        if (object == null || !(object.getClass().equals(this.getClass())))
        {
            return false;
        }

        CurvedFilament fil = (CurvedFilament) object;
        if (fil.m_points.size() != m_points.size())
        {
            return false;
        }
        for (int i = 0; i < m_points.size(); i++)
        {
            if (!fil.m_points.get(i).equals(m_points.get(i)))
            {
                return false;
            }
        }
        return (m_keep == fil.m_keep && m_width == fil.m_width && m_mean_signed_curvature == fil.m_mean_signed_curvature &&
                m_mean_absolute_curvature == fil.m_mean_absolute_curvature && m_center.x == fil.m_center.x &&
                m_center.y == fil.m_center.y && m_length == fil.m_length &&
                m_orientation % (180 * Const.M) == fil.m_orientation % (180 * Const.M));
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((m_center == null) ? 0 : m_center.hashCode());
        result = prime * result + (m_keep ? 1231 : 1237);
        result = prime * result + (int) (m_length ^ (m_length >>> 32));
        result = prime * result + (int) (m_mean_absolute_curvature ^ (m_mean_absolute_curvature >>> 32));
        result = prime * result + (int) (m_mean_signed_curvature ^ (m_mean_signed_curvature >>> 32));
        result = prime * result + (int) (m_orientation ^ (m_orientation >>> 32));
        result = prime * result + ((m_points == null) ? 0 : m_points.hashCode());
        result = prime * result + (int) (m_width ^ (m_width >>> 32));
        return result;
    }

}