package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import core.DataParameters.Any;
import core.DataParameters.Trace;
import filters.FilterCloseHoles;
import ui.FilamentSensor;

public class DataFilaments
{
    private boolean m_changed_flag = false;
    private ChangeListener m_listener;

    private List<FilamentChain> m_filament_chain_list;
    private List<AbstractFilament> m_selection;

    private List<ChainOperation> m_undo_chain_list;
    private List<ChainOperation> m_redo_chain_list;

    private int m_tracing_parameters_hash;

    public DataFilaments(List<FilamentChain> filament_chain_list, boolean flag)
    {
        m_filament_chain_list = (filament_chain_list == null ? new ArrayList<FilamentChain>() :
                                                               filament_chain_list);
        AbstractFilament.sortList(m_filament_chain_list);
        m_selection = new ArrayList<AbstractFilament>();

        m_undo_chain_list = new ArrayList<ChainOperation>();
        m_redo_chain_list = new ArrayList<ChainOperation>();

        m_tracing_parameters_hash = 0;
    }

    public DataFilaments(List<SingleFilament> filament_list)
    {
        m_filament_chain_list = (filament_list == null ? new ArrayList<FilamentChain>() :
                                                         toChain(filament_list));
        AbstractFilament.sortList(m_filament_chain_list);
        m_selection = new ArrayList<AbstractFilament>();

        m_undo_chain_list = new ArrayList<ChainOperation>();
        m_redo_chain_list = new ArrayList<ChainOperation>();

        m_tracing_parameters_hash = 0;
    }

    DataFilaments()
    {
        this(null);
    }

    void cleanTracedFilaments(boolean remember)
    {
        if (m_filament_chain_list == null || m_filament_chain_list.isEmpty())
        {
            return;
        }

        for (int i = m_filament_chain_list.size() - 1; i >= 0; i--)
        {
            FilamentChain chain = m_filament_chain_list.get(i);

            boolean removed = false;

            List<SingleFilament> chain_content = chain.getFilaments();
            for (int j = chain_content.size() - 1; j >= 0; j--)
            {
                SingleFilament filament = chain_content.get(j);
                if (!filament.getKeep())
                {
                    chain_content.remove(filament);
                    removed = true;
                }
            }
            if (removed)
            {
                removeFilamentChain(chain, remember);
                for (SingleFilament filament : chain_content)
                {
                    addFilamentChain(new FilamentChain(filament), remember);
                }
            }
        }
        AbstractFilament.sortList(m_filament_chain_list);
    }

    void keepDiscardAllFilaments()
    {
        if (m_filament_chain_list == null || m_filament_chain_list.isEmpty())
        {
            return;
        }

        boolean discard_all = false;
        for (FilamentChain chain : m_filament_chain_list)
        {
            if (chain.getKeepAny())
            {
                discard_all = true;
                break;
            }
        }
        if (discard_all)
        {
            for (FilamentChain chain : m_filament_chain_list)
            {
                chain.setKeepAll(false);
            }
        }
        else
        {
            for (FilamentChain chain : m_filament_chain_list)
            {
                chain.setKeepAll(true);
            }
        }
    }

    void remedyOldAngles()
    {
        if (m_filament_chain_list == null || m_filament_chain_list.isEmpty())
        {
            return;
        }

        for (FilamentChain chain : m_filament_chain_list)
        {
            chain.remedyOldAngles();
        }
    }

    boolean[][] scanFilaments(Tracer sensor, boolean[][] binary_image, boolean[][] mask, Map<Any, Integer> trace_params)
    {
        m_selection.clear();
        cleanTracedFilaments(false);
        keepDiscardAllFilaments();
        cleanTracedFilaments(false);
        List<SingleFilament> fils = sensor.scanFilaments(binary_image,
                                                         trace_params.get(Trace.tolerance) / 100.0,
                                                         trace_params.get(Trace.minlen),
                                                         trace_params.get(Trace.minangle),
                                                         trace_params.get(Trace.step));

        if (trace_params.get(Trace.no_boundary) != 0)
        {
            boolean[][] alt_mask = Calc.erode(FilterCloseHoles.apply(Calc.close(sensor.getPixelMask(),
                                                                            Calc.circleMask(50))),
                                          Calc.circleMask(30));
            mask = Calc.erode(mask, Calc.circleMask(30));
            final int image_width = alt_mask.length, height = alt_mask[0].length;
            for (int x = 0; x < image_width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    mask[x][y] = alt_mask[x][y] || mask[x][y];
                }
            }
            for (int i = fils.size() - 1; i >= 0; i--)
            {
                int counter = 0;
                List<Point> points = fils.get(i).getPoints();
                for (Point p : points)
                {
                    if (!mask[p.x][p.y])
                    {
                        counter++;
                    }
                }
                if (counter > points.size() / 2)
                {
                    long width = fils.remove(i).getWidth();
                    if (counter < points.size() - 1)
                    {
                        for (int j = points.size() - 1; j > 0; j--)
                        {
                            Point p = points.get(j);
                            if (mask[p.x][p.y])
                            {
                                break;
                            }
                            points.remove(j);
                        }
                        while (true)
                        {
                            Point p = points.get(0);
                            if (mask[p.x][p.y])
                            {
                                break;
                            }
                            points.remove(0);
                        }
                        fils.add(i, new CurvedFilament(points, width));
                    }
                }
            }
        }
        addFilamentChains(toChain(fils), false);
        return mask;
    }

    void splitToLinear()
    {
        List<SingleFilament> old_list = getFilaments();
        List<SingleFilament> new_list = new ArrayList<>();
        for (SingleFilament f : old_list)
        {
            new_list.addAll(f.splitIntoLinearPieces());
        }
        m_filament_chain_list = toChain(new_list);
    }

    void updateTraceParams(Map<Any, Integer> trace_params)
    {
        if (trace_params.hashCode() == m_tracing_parameters_hash)
        {
            return;
        }

        m_tracing_parameters_hash = trace_params.hashCode();
        setChanged(true);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////////////////    Change and callback handling    //////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    void setChanged(boolean changed)
    {
        m_changed_flag = changed;

        if (m_listener == null)
        {
            return;
        }

        m_listener.stateChanged(new ChangeEvent(this));
    }

    boolean hasChanged()
    {
        return m_changed_flag;
    }

    void addChangeListener(ChangeListener cl)
    {
        m_listener = cl;
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////////   Filament operations and undo/redo support   /////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    List<SingleFilament> getSelectedFilaments()
    {
        return toBare(m_selection);
    }

    public List<AbstractFilament> getSelection()
    {
        List<AbstractFilament> selection = new ArrayList<>();

        for (AbstractFilament chain : m_selection)
        {
            selection.add(chain);
        }

        return selection;
    }

    void setSelection(List<? extends AbstractFilament> list)
    {
        m_selection.clear();

        if (list != null)
        {
            m_selection.addAll(list);
        }

        setChanged(m_changed_flag);
    }

    void setSelection(AbstractFilament fil)
    {
        m_selection.clear();

        if (fil != null)
        {
            m_selection.add(fil);
        }

        setChanged(m_changed_flag);
    }

    void setSelection(int index)
    {
        this.setSelection(getFilamentChain(index));
    }

    void clearSelection()
    {
        m_selection.clear();
        setChanged(m_changed_flag);
    }

    int countChains()
    {
        return m_filament_chain_list.size();
    }

    int countFilaments()
    {
        return toBare(m_filament_chain_list).size();
    }

    int countRows()
    {
        int rows = m_filament_chain_list.size();
        for (FilamentChain chain : m_filament_chain_list)
        {
            int size = chain.getListSize();
            if (size > 1)
            {
                rows += size;
            }
        }
        return rows;
    }

    /**
     * This method walks through the FilamentChains and Filaments in the order
     * they are displayed in the table. It returns a clone of the
     * AbstractFilament at position row.
     */
    AbstractFilament getEntryInRow(int row)
    {
        int counter = 0;
        for (int i = 0; i < m_filament_chain_list.size(); i++)
        {
            if (row == counter)
            {
                FilamentChain chain = (FilamentChain) m_filament_chain_list.get(i).clone();
                return chain;
            }

            int size = m_filament_chain_list.get(i).getListSize();

            if (size > 1)
            {
                for (int j = 0; j < size; j++)
                {
                    counter++;
                    if (row == counter)
                    {
                        SingleFilament filament = m_filament_chain_list.get(i).getFilament(j);
                        return filament;
                    }
                }
            }
            counter++;
        }
        return null;
    }

    /**
     * This method walks through the FilamentChains and Filaments in the order
     * they are displayed in the table. It returns the AbstractFilament at
     * position row.
     */
    AbstractFilament changeEntryInRow(int row)
    {
        int counter = 0;
        for (int i = 0; i < m_filament_chain_list.size(); i++)
        {
            if (row == counter)
            {
                FilamentChain chain = m_filament_chain_list.get(i);
                return chain;
            }

            int size = m_filament_chain_list.get(i).getListSize();

            if (size > 1)
            {
                for (int j = 0; j < size; j++)
                {
                    counter++;
                    if (row == counter)
                    {
                        SingleFilament filament = m_filament_chain_list.get(i).changeFilament(j);
                        return filament;
                    }
                }
            }
            counter++;
        }
        return null;
    }

    /**
     * This method walks through the FilamentChains and Filaments in the order
     * they are displayed in the table. It returns the position of
     * AbstractFilament fil.
     */
    int getRowOfEntry(AbstractFilament fil)
    {
        int counter = 0;
        for (int i = 0; i < m_filament_chain_list.size(); i++)
        {
            FilamentChain chain = (FilamentChain) m_filament_chain_list.get(i).clone();
            if (chain.equals(fil))
            {
                return counter;
            }
            int size = chain.getListSize();
            if (size == 1)
            {
                if (chain.getFilament(0).equals(fil))
                {
                    return counter;
                }
            }

            if (size > 1)
            {
                for (int j = 0; j < size; j++)
                {
                    counter++;
                    if (chain.getFilament(j).equals(fil))
                    {
                        return counter;
                    }
                }
            }
            counter++;
        }
        return -1;
    }

    AbstractFilament wrap(AbstractFilament fil)
    {
        return changeEntryInRow(getRowOfEntry(fil));
    }

    FilamentChain getFilamentChain(int index)
    {
        if (index >= m_filament_chain_list.size())
        {
            return null;
        }

        return m_filament_chain_list.get(index);
    }

    FilamentChain getChainContainingFilament(SingleFilament filament)
    {
        for (FilamentChain chain : m_filament_chain_list)
        {
            if (chain.contains(filament))
            {
                return chain;
            }
        }
        return null;
    }

    List<SingleFilament> getFilaments()
    {
        return toBare(m_filament_chain_list);
    }

    List<FilamentChain> getFilamentChains()
    {
        List<FilamentChain> chain_list = new ArrayList<FilamentChain>();

        for (FilamentChain chain : m_filament_chain_list)
        {
            chain_list.add((FilamentChain) chain.clone());
        }

        return chain_list;
    }

    void addFilamentChain(FilamentChain chain, boolean undoable)
    {
        if (undoable)
        {
            add_undoChain(new ChainOperation(ChainOperation.ADD, null, chain, null));
        }
        m_selection.clear();
        m_selection.add(chain);
        m_filament_chain_list.add(chain);
        AbstractFilament.sortList(m_filament_chain_list);

        setChanged(true);
    }

    private void addFilamentChains(List<FilamentChain> list, boolean undoable)
    {
        if (undoable)
        {
            add_undoChain(new ChainOperation(ChainOperation.ADD, null, null, list));
        }

        m_filament_chain_list.addAll(list);
        AbstractFilament.sortList(m_filament_chain_list);

        setChanged(true);
    }

    boolean modifyAbstractFilament(AbstractFilament before, AbstractFilament after)
    {
        if (!before.getClass().equals(after.getClass()))
        {
            System.out.println("Illegal replacement: " + before.getClass().getCanonicalName() +
                               " to " + after.getClass().getCanonicalName());
            return false;
        }
        if (before.getClass().equals(FilamentChain.class))
        {
            add_undoChain(new ChainOperation(ChainOperation.MODIFY, (FilamentChain) before,
                                             (FilamentChain) after, null));
            m_selection.clear();
            m_selection.add((FilamentChain) after);
            m_filament_chain_list.remove((FilamentChain) before);
            m_filament_chain_list.add((FilamentChain) after);
        }
        else if (before instanceof SingleFilament)
        {
            FilamentChain before_chain = getChainContainingFilament((SingleFilament) before);
            FilamentChain after_chain = (FilamentChain) before_chain.clone();

            if (after_chain.removeFilament((SingleFilament) before))
            {
                after_chain.add((SingleFilament) after);
                // TODO: Check for integrity of chain!
            }
            else
            {
                return false;
            }
            add_undoChain(new ChainOperation(ChainOperation.MODIFY, before_chain, after_chain, null));
            m_selection.clear();
            m_selection.add(after_chain);
            m_filament_chain_list.remove(before_chain);
            m_filament_chain_list.add(after_chain);
        }
        else
        {
            System.out.println("Illegal class: " + before.getClass().getCanonicalName());
            return false;
        }

        AbstractFilament.sortList(m_filament_chain_list);

        setChanged(true);
        return true;
    }

    private void removeFilamentChain(FilamentChain chain, boolean undoable)
    {
        if (undoable)
        {
            add_undoChain(new ChainOperation(ChainOperation.REMOVE, chain, null, null));
        }
        m_selection.clear();
        m_filament_chain_list.remove(chain);
        AbstractFilament.sortList(m_filament_chain_list);

        setChanged(true);
    }

    void deleteSelected()
    {
        for (AbstractFilament fil : m_selection)
        {
            if (fil instanceof SingleFilament)
            {
                for (FilamentChain chain : m_filament_chain_list)
                {
                    if (chain.contains((SingleFilament) fil))
                    {
                        m_filament_chain_list.remove(chain);

                        List<SingleFilament> filament_list = chain.getFilaments();
                        filament_list.remove(fil);
                        m_filament_chain_list.addAll(toChain(filament_list));
                    }
                }
            }
            else if (fil instanceof FilamentChain)
            {
                m_filament_chain_list.remove(fil);
            }
            else
            {
                System.out.println("AbstractFilament of unknown type: " + fil.getClass().getCanonicalName());
            }
        }
        AbstractFilament.sortList(m_filament_chain_list);
    }

    private void add_undoChain(ChainOperation cop)
    {
        // if (too long)
        // TODO config option

        m_undo_chain_list.add(cop);
        m_redo_chain_list.clear();

        FilamentSensor.debugMessage(m_undo_chain_list.size() + " filament operations in undo list.");
    }

    void undoChain()
    {
        if (m_undo_chain_list == null || m_undo_chain_list.isEmpty())
        {
            return;
        }

        ChainOperation cop = m_undo_chain_list.remove(m_undo_chain_list.size() - 1);

        switch (cop.getType())
        {
        case ChainOperation.ADD:
            m_selection.clear();
            if (cop.getChainList() != null)
            {
                for (FilamentChain chain : cop.getChainList())
                {
                    m_filament_chain_list.remove(chain);
                }
            }
            else
            {
                m_filament_chain_list.remove(cop.getNewChain());
            }
            break;

        case ChainOperation.REMOVE:
            m_selection.clear();
            if (cop.getChainList() != null)
            {
                for (FilamentChain chain : cop.getChainList())
                {
                    m_filament_chain_list.add(chain);
                    m_selection.add(chain);
                }
            }
            else
            // TODO: check for null?
            {
                m_filament_chain_list.add(cop.getOldChain());
                m_selection.add(cop.getOldChain());
            }
            break;

        case ChainOperation.MODIFY: // TODO: check for null?
            m_selection.clear();
            int index = m_filament_chain_list.indexOf(cop.getNewChain());
            m_filament_chain_list.remove(cop.getNewChain());
            m_filament_chain_list.add(index, cop.getOldChain());
            m_selection.add(cop.getOldChain());
            break;

        default:
            System.out.println("Someone is messing with us... cop.type = " + cop.getType());
        }

        m_redo_chain_list.add(cop);

        AbstractFilament.sortList(m_filament_chain_list);

        setChanged(true);

        if (FilamentSensor.DEBUG())
        {
            String operation = "??";
            int count = cop.getChainList() == null ? 1 : cop.getChainList().size();

            switch (cop.getType())
            {
            case ChainOperation.ADD:
                operation = "Adding";
                break;
            case ChainOperation.REMOVE:
                operation = "Removing";
                break;
            case ChainOperation.MODIFY:
                operation = "Modifying";
            }

            System.out.println(operation + " " + count + " filament chain(s) undone.");
        }
    }

    public boolean canUndoChain()
    {
        return !m_undo_chain_list.isEmpty();
    }

    void redoChain()
    {
        if (m_redo_chain_list.isEmpty())
        {
            return;
        }

        ChainOperation cop = m_redo_chain_list.remove(m_redo_chain_list.size() - 1);

        switch (cop.getType())
        {
        case ChainOperation.ADD:
            m_selection.clear();
            if (cop.getChainList() != null)
            {
                for (FilamentChain chain : cop.getChainList())
                {
                    m_filament_chain_list.add(chain);
                    m_selection.add(chain);
                }
            }
            else
            // TODO: check for null?
            {
                m_filament_chain_list.add(cop.getNewChain());
                m_selection.add(cop.getNewChain());
            }
            break;

        case ChainOperation.REMOVE:
            m_selection.clear();
            if (cop.getChainList() != null)
            {
                for (FilamentChain chain : cop.getChainList())
                {
                    m_filament_chain_list.remove(chain);
                }
            }
            else
            {
                m_filament_chain_list.remove(cop.getOldChain());
            }
            break;

        case ChainOperation.MODIFY: // TODO: check for null?
            m_selection.clear();
            int index = m_filament_chain_list.indexOf(cop.getOldChain());
            m_filament_chain_list.remove(cop.getOldChain());
            m_filament_chain_list.add(index, cop.getNewChain());
            m_selection.add(cop.getNewChain());
            break;

        default:
            System.out.println("Someone is messing with us... cop.type = " + cop.getType());
        }

        m_undo_chain_list.add(cop);

        AbstractFilament.sortList(m_filament_chain_list);

        setChanged(true);

        if (FilamentSensor.DEBUG())
        {
            String operation = "??";
            int count = cop.getChainList() == null ? 1 : cop.getChainList().size();

            switch (cop.getType())
            {
            case ChainOperation.ADD:
                operation = "Adding";
                break;
            case ChainOperation.REMOVE:
                operation = "Removing";
                break;
            case ChainOperation.MODIFY:
                operation = "Modifying";
            }

            System.out.println(operation + " " + count + " filament chain(s) redone.");
        }
    }

    public boolean canRedoChain()
    {
        return !m_redo_chain_list.isEmpty();
    }

    private static List<SingleFilament> toBare(List<? extends AbstractFilament> abstract_filament_list)
    {
        if (abstract_filament_list == null)
        {
            return null;
        }

        List<SingleFilament> filament_list = new ArrayList<>();
        for (AbstractFilament fil : abstract_filament_list)
        {
            if (fil instanceof SingleFilament)
            {
                filament_list.add((SingleFilament) fil);
            }
            else if (fil instanceof FilamentChain)
            {
                filament_list.addAll(((FilamentChain) fil).getFilaments());
            }
            else
            {
                System.out.println("AbstractFilament of unknown type: " + fil.getClass().getCanonicalName());
            }
        }
        AbstractFilament.sortList(filament_list);
        return filament_list;
    }

    private static List<FilamentChain> toChain(List<SingleFilament> filament_list)
    {
        if (filament_list == null)
        {
            return null;
        }

        List<FilamentChain> filament_chain_list = new ArrayList<FilamentChain>();
        for (SingleFilament filament : filament_list)
        {
            filament_chain_list.add(new FilamentChain(filament));
        }
        AbstractFilament.sortList(filament_chain_list);
        return filament_chain_list;
    }

    /**
     * Internal class to store operations that modified the list of filament
     * chains. It is used for undo/redo of such operations.
     */
    private class ChainOperation
    {
        private int m_type;
        private FilamentChain m_old_filament_chain, m_new_filament_chain;
        private List<FilamentChain> m_filament_chain_list_op;

        private static final int ADD = 0, REMOVE = 1, MODIFY = 2;

        private ChainOperation(int type, FilamentChain filament_old, FilamentChain filament_new,
                               List<FilamentChain> list)
        {
            m_type = type; // No check for invalid type. TODO: Problem?
            m_old_filament_chain = filament_old;
            m_new_filament_chain = filament_new;
            m_filament_chain_list_op = list;
        }

        public int getType()
        {
            return m_type;
        }

        public FilamentChain getOldChain()
        {
            return m_old_filament_chain;
        }

        public FilamentChain getNewChain()
        {
            return m_new_filament_chain;
        }

        public List<FilamentChain> getChainList()
        {
            return m_filament_chain_list_op;
        }
    }
}
