package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileWriter;
import java.io.IOException;
import java.text.DecimalFormatSymbols;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.Scanner;
import java.util.Set;
import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.image.BufferedImage;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JOptionPane;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.event.ChangeListener;

import core.DataParameters.Any;
import core.DataParameters.Pre;
import core.DataParameters.Trace;
import core.Misc.Int3D;
import core.Misc.Pair;
import static core.Misc.*;
import core.WrappedSiZer;
import filters.PluginCellShape;
import ui.Config;
import ui.Dialogs;
import ui.FilamentSensor;
import ui.ImageFactory;
import ui.MainWindow;
import ui.PopUpOrientationFields;

public class DataImage extends DataProject
{
    private File m_image_file;

    private ImageProcessing m_image_processing;
    private DataParameters m_parameter_container;
    private DataFilaments m_filament_container;
    private int[][][] m_orientation_fields;
    private int[][] m_orientation_field_shapes;
    private Map<Integer, List<SingleFilament>> m_filaments_by_orientation_field;
    private boolean[][] m_mask;
    private CompressedBinaryImage m_cell_area;
    private long[] m_interior_data;

    private Tracer m_tracer;

    private String m_image_name, m_image_resolution;
    private static final String ls = System.lineSeparator();

    private PluginCellShape m_cell_data;

    public static final int ORIG = 0, PRE = 1, ELOG_COLOR = 2, ELOG_GREY = 3, DIFF = 4, BIN = 5,
                            ORI_MAP = 6,  HOUGH = 7, FINGERPRINT = 8, RED_GREEN = 9;

    // Main constructor. For reconstruction from save files.
    public DataImage(File img_file, PluginCellShape cell_data, DataParameters parameters,
                     DataFilaments filaments, List<List<Int1D>> orientation_fields,
                     boolean[][] mask)
              throws IOException
    {
        m_image_file = img_file;
        m_mask = mask;
        if (mask != null)
        {
            FilamentSensor.debugMessage("Mask given at creation of DataImage for " +
                                        img_file.getName());
        }

        m_parameter_container = (parameters == null ? getDefaultParameters() : parameters);

        if (img_file == null)
        {
            m_image_processing = null;
            m_cell_data = cell_data;
            m_image_name = "";
        }
        else
        {
            m_image_processing = new ImageProcessing(img_file);
            int min_range = m_parameter_container.getParameter(Pre.min_range);
            m_cell_data = (cell_data == null ? m_image_processing.getCellData(mask, min_range) : cell_data);
            m_image_name = img_file.getName();
            FilamentSensor.debugMessage("Processing with ImageProcessing was successful.");
        }

        m_filament_container = (filaments == null ? new DataFilaments() : filaments);

        if (orientation_fields != null && !orientation_fields.isEmpty())
        {
            m_orientation_fields = new int[orientation_fields.size()][][];
            for (int i = 0; i < orientation_fields.size(); i++)
            {
                List<Int1D> list_field = orientation_fields.get(i);
                if (list_field != null && !list_field.isEmpty())
                {
                    m_orientation_fields[i] = new int[list_field.size()][];
                    for (int j = 0; j < list_field.size(); j++)
                    {
                        m_orientation_fields[i][j] = list_field.get(j).the;
                    }
                }
                else
                {
                    m_orientation_fields[i] = new int[0][0];
                }
            }
            setOrientationFields(m_orientation_fields);
        }

        addAsProcessed();
    }

    // Constructor for new files.
    public DataImage(File img_file, DataParameters parameters, boolean[][] mask) throws IOException
    {
        this(img_file, null, parameters, null, null, mask);
    }

    protected void setParameters(DataParameters parameters)
    {
        m_parameter_container = parameters;
    }

    protected void setOrientationFields(int[][][] set) throws IOException
    {
        Pair<Int3D, Map<Integer, List<SingleFilament>>> pair = null;
        pair = OrientationFields.sortFilaments(set, m_filament_container.getFilaments());
        m_orientation_fields = pair.a.the;
        m_filaments_by_orientation_field = pair.b;
        determineOrientationFieldShapes();
        addAsProcessed();
    }

    protected void setMask(boolean[][] mask)
    {
        if (mask != null)
        {
            FilamentSensor.debugMessage("Mask given via DataImage.setMask() for " +
                                        m_image_file.getName());
        }
        m_mask = mask;
    }

    /**
     * Sets the ImageProcessing and LineSensor members to null in order to save memory.
     */
    protected void shutDown()
    {
        m_image_processing = null;
        m_tracer = null;
        m_mask = null;
        m_cell_data.shutdown();
    }

    protected void fullShutDown()
    {
        shutDown();
        m_cell_area = null;
    }

    /**
     * This method adds this DataImage object to the list of processed images,
     * that will be saved when saveProject is invoked. If a DataImage object for
     * the same image file already appears in the list of processed images, that
     * older object is discarded from the list, to avoid saving data for one
     * image twice.
     *
     * To uniquely identify the image to which a DataImage object belongs, the
     * content of the name field of the DataImage class is used.
     */
    void addAsProcessed()
    {
        for (int i = m_processed_images.size() - 1; i >= 0; i--)
        {
            DataImage tracked_change = m_processed_images.get(i);

            if (m_image_name.equals(tracked_change.getImageName()))
            {
                m_processed_images.remove(tracked_change);
            }
        }
        m_processed_images.add(this);
    }

    public DataParameters parameters()
    {
        if (m_force_defaults)
        {
            return getDefaultParameters();
        }
        
        if (m_parameter_container == null)
        {
            m_parameter_container = getDefaultParameters();
        }
        return m_parameter_container.completeFrom(getDefaultParameters());
    }

    private ImageProcessing imageProcessing() throws IOException
    {
        if (m_image_processing == null)
        {
            m_image_processing = new ImageProcessing(m_image_file);
        }
        return m_image_processing;
    }

    private Tracer tracer()
    {
        if (parameters().getParameter(Trace.curve) > 0)
        {
            if (m_tracer == null || !m_tracer.getClass().equals(CurveTracer.class))
            {
                m_tracer = new CurveTracer();
            }
        }
        else
        {
            if (m_tracer == null || !m_tracer.getClass().equals(LineSensor.class))
            {
                m_tracer = new LineSensor();
            }
        }
        return m_tracer;
    }

    private PluginCellShape cellData() throws IOException
    {
        if (m_cell_data == null)
        {
            m_cell_data = imageProcessing().getCellData(m_mask, parameters().getParameter(Pre.min_range));
        }
        return m_cell_data;
    }

    protected boolean[][] extendedMask() throws IOException
    {
        if (m_cell_data == null || m_cell_data.getExtendedMask() == null)
        {
            m_cell_data = imageProcessing().getCellData(m_mask, parameters().getParameter(Pre.min_range));
        }
        return m_cell_data.getExtendedMask();
    }

    protected boolean[][] areaMask(int padding) throws IOException
    {
        if (padding == 0)
        {
            // When run for the first time, create m_cell_area, which will survive shutDown().
            if (m_cell_area == null)
            {
                if (m_cell_data == null || m_cell_data.getPaddedMask(padding) == null)
                {
                    m_cell_data = imageProcessing().getCellData(m_mask, parameters().getParameter(Pre.min_range));
                }
                m_cell_area = new CompressedBinaryImage(m_cell_data.getPaddedMask(padding));
            }

            return m_cell_area.asArray();
        }
        if (m_cell_data == null || m_cell_data.getPaddedMask(padding) == null)
        {
            m_cell_data = imageProcessing().getCellData(m_mask, parameters().getParameter(Pre.min_range));
        }
        return m_cell_data.getPaddedMask(padding);
    }

    /**
     * Get the name of the image file corresponding to this data file.
     *
     * @return A string containing the name of the image file.
     */
    public String getImageName()
    {
        return m_image_name;
    }

    /**
     * Get the image file corresponding to this data file.
     *
     * @return The image file.
     */
    public File getFile()
    {
        return m_image_file;
    }

    public String getResolution()
    {
        if (m_image_processing != null)
        {
            m_image_resolution = m_image_processing.getResolution();
        }
        return m_image_resolution;
    }

    public PluginCellShape getCellData() throws IOException
    {
        return (PluginCellShape) cellData().clone();
    }

    public boolean isEmpty() throws IOException
    {
        boolean[][] binary_image = extendedMask();
        if (binary_image == null)
        {
            // Account for possibly loaded project which lacks area masks.
            m_cell_data = imageProcessing().getCellData(m_mask, parameters().getParameter(Pre.min_range));
            binary_image = extendedMask();
        }
        if (binary_image == null)
        {
            return true;
        }
        final int width = binary_image.length,
                  height = binary_image[0].length;

        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    return false;
                }
            }
        }
        return true;
    }

    public int[][][] getOrientationFields()
    {
        if (m_orientation_fields == null)
        {
            return new int[0][0][0];
        }
        return m_orientation_fields.clone();
    }

    public boolean usesDefaultParameters()
    {
        return parameters().equals(m_default_parameters);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////////////    Export functions for text files     /////////////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    /**
     * The import from CSV assumes orientations counterclockwise from
     * version 0.1.3 on. This implies "inverting the y-axis". When older
     * CSV files are imported, filament end points will be wrong. The
     * filament class has a method to remedy this.
     * 
     * FIXME: Do not import csv files with curved filaments.
     */
    public boolean importCSV()
    {
        File csv_file = Dialogs.load(false, "Choose csv file.");
        List<String> input = new ArrayList<String>();
        try
        {
            Scanner scanner = new Scanner(csv_file);
            while (!scanner.hasNext("%filaments"))
            {
                FilamentSensor.debugMessage("Skipping line.");
                scanner.nextLine();
            }
            scanner.nextLine();
            
            // Now, read filaments, skipping comments and stopping as soon as another key appears.
            while (scanner.hasNext())
            {
                FilamentSensor.debugMessage("Reading line.");
                String line = scanner.nextLine();
                if (line.startsWith("#"))
                {
                    continue;
                }
                else if (line.startsWith("%"))
                {
                    break;
                }
                input.add(line);
            }
            for (String line : input)
            {
                String dec = Character.toString((new DecimalFormatSymbols()).getDecimalSeparator());
                line = line.replaceAll(",", "_").replaceAll("\\.", dec);
                scanner = new Scanner(line);
                scanner.useDelimiter("_");
                if (scanner.hasNextDouble())
                {
                    double center_x = scanner.nextDouble();
                    double center_y = scanner.nextDouble();
                    double length = scanner.nextDouble();
                    double angle = scanner.nextDouble();
                    long width = Math.round(Const.MF * scanner.nextDouble());
                    int head_x = (int)Math.round(center_x + length * Math.cos(angle * Const.RAD) / 2.0);
                    int head_y = (int)Math.round(center_y - length * Math.sin(angle * Const.RAD) / 2.0);
                    int tail_x = (int)Math.round(center_x - length * Math.cos(angle * Const.RAD) / 2.0);
                    int tail_y = (int)Math.round(center_y + length * Math.sin(angle * Const.RAD) / 2.0);
                    Filament filament = new Filament(new Point(head_x, head_y),
                                                     new Point(tail_x, tail_y),
                                                     width, true);
                    m_filament_container.addFilamentChain(new FilamentChain(filament), false);
                }
                
            }
            scanner.close();
        }
        catch(FileNotFoundException e)
        {
            System.out.println("File not found." + e.getMessage());
        }

        return false;
    }

    public boolean importCSV_old()
    {
        File csv_file = Dialogs.load(false, "Choose csv file.");
        List<String> input = new ArrayList<String>();
        try
        {
            Scanner scanner = new Scanner(csv_file);
            while (scanner.hasNext("#center_x,center_y,length,angle,width"))
            {
                FilamentSensor.debugMessage("Skipping line.");
                scanner.nextLine();
            }
            while (scanner.hasNext())
            {
                FilamentSensor.debugMessage("Reading line.");
                input.add(scanner.nextLine());
            }
            for (String line : input)
            {
                String dec = Character.toString((new DecimalFormatSymbols()).getDecimalSeparator());
                line = line.replaceAll(",", "_").replaceAll("\\.", dec);
                scanner = new Scanner(line);
                scanner.useDelimiter("_");
                if (scanner.hasNextDouble())
                {
                    double center_x = scanner.nextDouble();
                    double center_y = scanner.nextDouble();
                    double length = scanner.nextDouble();
                    double angle = scanner.nextDouble();
                    long width = Math.round(Const.MF * scanner.nextDouble());
                    int head_x = (int)Math.round(center_x + length * Math.cos(angle * Const.RAD) / 2.0);
                    int head_y = (int)Math.round(center_y - length * Math.sin(angle * Const.RAD) / 2.0);
                    int tail_x = (int)Math.round(center_x - length * Math.cos(angle * Const.RAD) / 2.0);
                    int tail_y = (int)Math.round(center_y + length * Math.sin(angle * Const.RAD) / 2.0);
                    Filament filament = new Filament(new Point(head_x, head_y),
                                                     new Point(tail_x, tail_y),
                                                     width, true);
                    m_filament_container.addFilamentChain(new FilamentChain(filament), false);
                }
                
            }
            scanner.close();
        }
        catch(FileNotFoundException e)
        {
            System.out.println("File not found." + e.getMessage());
        }

        return false;
    }

    /**
     * Exports the list of filament chains as comma separated values.
     */
    public void exportCSV(File path, boolean chains, boolean dialog) throws IOException
    {
        if (m_interior_data == null)
        {
            m_interior_data = new long[]{-1,-1,-1};
        }
        String out = "#" + FilamentSensor.NAME + "-" + FilamentSensor.VERSION + ls +
                     "#cell_properties" + ls +
                     "%center_x" + ls + cellData().getCenter().x + ls +
                     "%center_y" + ls + cellData().getCenter().y + ls +
                     "%area" + ls + cellData().getArea() + ls +
                     "%aspect_ratio" + ls + cellData().getAspectRatio() / Const.MF + ls +
                     "%long_half_axis" + ls + cellData().getLongHalfAxis() / Const.MF + ls +
                     "%orientation" + ls + cellData().getOrientation() / Const.MF + ls +
                     "%order_parameter" + ls + cellData().getOrderParameter() / Const.MF + ls +
                     "%new_orientation" + ls + cellData().getNewOrientation() / Const.MF + ls +
                     "%new_order_parameter" + ls + cellData().getNewOrderParameter() / Const.MF + ls +
                     "%area_interior" + ls + m_interior_data[0] + ls +
                     "%mean_brightness_interior" + ls + m_interior_data[1] / Const.MF + ls +
                     "%area_boundary" + ls + (cellData().getArea() - m_interior_data[0]) + ls +
                     "%mean_brightness_boundary" + ls + m_interior_data[2] / Const.MF + ls +
                     "%excursion_ratio" + ls + cellData().getExcursion() / Const.MF + ls;
        if (m_filament_container.countChains() > 0)
        {
            out += "%filaments" + ls + "#center_x,center_y,length,angle,width,signed_curv,total_curv" + ls;
            for (SingleFilament fil : m_filament_container.getFilaments())
            {
                out += fil.getCenter().x / 10.0 + "," + fil.getCenter().y / 10.0 + "," +
                       fil.getLength() / Const.MF + ","  + fil.getOrientation() / Const.MF + "," +
                       fil.getWidth() / Const.MF + "," + fil.getSignedCurvature() / Const.MF + "," +
                       fil.getAbsoluteCurvature()/ Const.MF + ls;
            }
        }

        if (m_orientation_fields != null)
        {
            for (int i = 1; i < m_orientation_fields.length; i++)
            {
                List<SingleFilament> filaments = null;
                if (m_filaments_by_orientation_field != null &&
                    m_filaments_by_orientation_field.size() > i)
                {
                    filaments = m_filaments_by_orientation_field.get(i);
                }
                FilamentSensor.debugMessage("Field " + i + " data: " + m_orientation_field_shapes[i]);
                for (String s : processOrientationField(m_orientation_field_shapes[i],
                                                        m_orientation_fields[i], countBlocks(),
                                                        filaments))
                {
                    out += "%orientation_field_" + i + "_" + s;
                }
            }
        }

        export("Filament data to export", out, dialog, path, chains);
    }

    public void exportFingerprint(File path) throws IOException
    {
        int[][] fingerprint = tracer().makeFingerprint(cellData().getLongHalfAxis() / 5e5);
        File file = new File(path.getAbsolutePath() + File.separator +
                             m_image_file.getName().replaceAll("\\.", "_") + "_fingerprint.csv");
        writeFile(file, imageToString(fingerprint));
    }

    public void exportHistogram(File path, boolean dialog, String parameters, int type)
            throws IOException
    {
        String type_name = "";
        switch (type)
        {
        case ELOG_COLOR:
            type_name = "eLoG";
            break;
        case HOUGH:
            type_name = "Hough transformation";
            break;
        default:
            return;
        }

        if (imageProcessing().getImage(type) == null)
        {
            FilamentSensor.debugMessage(type_name + "s have not been calculated yet...");
            return;
        }

        String out = parameters + ls + "%orientation_histogram" + ls + "#orientation,pixels" + ls;
        out += imageProcessing().getHistogram(type);

        export(type_name + " histogram", out, dialog, path, false);
    }

    private void export(String description, String content, boolean dialog, File path, boolean chains)
    {
        if (dialog)
        {
            Object option[] = new Object[]
            { "OK", "Save as..." };
            JScrollPane message = new JScrollPane(new JTextArea(content));
            message.setMaximumSize(new Dimension(400, 300));
            message.setPreferredSize(new Dimension(400, 300));

            // Continue only if "Save as..." is clicked in the dialog.
            if (JOptionPane.showOptionDialog(MainWindow.frame(), message, description, JOptionPane.YES_NO_OPTION,
                    JOptionPane.INFORMATION_MESSAGE, null, option, option[0]) != JOptionPane.NO_OPTION)
            {
                return;
            }

            JFileChooser jfc = new JFileChooser();

            String folder = Config.get("LastFolders", "Export");

            if (folder != "")
            {
                jfc.setCurrentDirectory(new File(folder));
            }

            // Continue only if the approving option is clicked in the dialog.
            if (jfc.showSaveDialog(MainWindow.frame()) != JFileChooser.APPROVE_OPTION)
            {
                return;
            }

            if (!writeFile(jfc.getSelectedFile(), content))
            {
                return;
            }

            try
            {
                folder = jfc.getSelectedFile().getParentFile().getCanonicalPath();
                Config.set("LastFolders", "Export", folder);
            }
            catch (IOException e)
            {
            }
        }
        else
        {
            File file = new File(path.getAbsolutePath() + File.separator +
                                 m_image_file.getName().replaceAll("\\.", "_") +
                                 (chains ? "_chains.csv" : "_filaments.csv"));
            writeFile(file, content);
        }
    }

    private boolean writeFile(File file, String content)
    {
        try
        {
            // Default file name for batch processing.
            if (file == null)
            {
                // "[A-Za-z0-9]{3,4}$" should allow for 3 or 4 character file
                // name extensions.
                file = new File(m_image_file.getCanonicalPath().replaceAll("\\.[A-Za-z0-9]{3,4}$", "") + ".csv");
            }

            BufferedWriter bwr = new BufferedWriter(new FileWriter(file, false));
            bwr.write(content);
            bwr.close();

            FilamentSensor.debugMessage("Exported data to '" + file + "'.");
        }
        catch (IOException e)
        {
            System.out.println("Exporting to file failed: " + e);
            return false;
        }
        return true;
    }

    public boolean saveImage(File path, String identifier, int type, boolean filaments,
                             Color color, boolean overwrite) throws IOException
    {
        return saveImage(path, identifier, makeImage(type, filaments, color), overwrite);
    }

    public boolean saveImage(File path, String identifier, BufferedImage image, boolean overwrite)
    {
        if (image == null)
        {
            return false;
        }
        File img_file = null;
        if (path == null)
        {
            img_file = Dialogs.save(false, "Save image...");
        }
        else
        {
            img_file = new File(path.getAbsolutePath() + File.separator +
                                m_image_file.getName().replaceAll("\\.", "_") +
                                (identifier != null ? "_" + identifier : "") + ".png");
            Config.set("LastFolders", "Save", path.getAbsolutePath());
        }

        if (img_file == null)
        {
            return false;
        }

        try
        {
            FilamentSensor.debugMessage("Try to save image to '" + img_file.getAbsolutePath() + "'.");
            if (img_file.exists())
            {
                String msg = "File exists '" + img_file.getAbsolutePath() + "'.";
                if (overwrite && img_file.delete())
                {
                    FilamentSensor.debugMessage(msg + " Deleted!");
                }
                else
                {
                    FilamentSensor.debugMessage(msg + " Skipped!");
                    return false;
                }
            }
            return ImageIO.write(image, "png", img_file);
        }
        catch (IOException e)
        {
            System.out.println(e.getMessage());
            return false;
        }
    }

    public boolean makeIllustrations(File path) throws IOException
    {
        BufferedImage main = makeImage(ORIG);
        BufferedImage other = main;
        for (Integer key : m_filaments_by_orientation_field.keySet())
        {
            if (key <= 0)
            {
                other = addFilaments(other, m_filaments_by_orientation_field.get(key), Color.cyan);
            }
            else if (key == 1)
            {
                main = addFilaments(main, m_filaments_by_orientation_field.get(key), Color.orange);
            }
            else
            {
                other = addFilaments(other, m_filaments_by_orientation_field.get(key), Color.orange);
            }
        }
        addAsProcessed();
        return saveImage(path, "main", main, true) && saveImage(path, "other", other, true);
    }
    

    ////////////////////////////////////////////////////////////////////////////////////////////////
    ////////////////////  Exposing the public methods of m_image_processing  ///////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    public void updatePreprocessing() throws IOException
    {
        Map<Any, Integer> pre_params = parameters().getParameters(DataParameters.PRE);
        boolean changed = hasChanged();

        boolean[][] area_mask = ((pre_params.get(Pre.lg_with_mask) == 1) && !isEmpty() ?
                                  extendedMask() : null);
        changed = imageProcessing().updatePreprocessing(pre_params, area_mask, changed);
        setChanged(changed);
    }

    public String calculateELoGs() throws IOException
    {
        Map<Any, Integer> elog_params = parameters().getParameters(DataParameters.ELOG);
        String out = "#Parameters:" + ls;
        for (Any key : elog_params.keySet())
        {
            out += "%" + key + ls + (key.name() == "orientations" || key.name() == "threshold" ?
                                (int)elog_params.get(key) : elog_params.get(key) / 10.0) + ls;
        }

        if (imageProcessing().calculateELoGs(elog_params))
        {
            setChanged(true);
        }
        return out;
    }

    public String calculateHough() throws IOException
    {
        Map<Any, Integer> hough_params = parameters().getParameters(DataParameters.HOUGH);
        String out = "#Parameters:" + ls;
        for (Any key : hough_params.keySet())
        {
            out += "%" + key + ls + hough_params.get(key) + ls;
        }

        updateBinarization();
        if (imageProcessing().calculateHough(hough_params))
        {
            setChanged(true);
        }
        return out;
    }

    public void updateBinarization() throws IOException
    {
        Map<Any, Integer> bin_params = parameters().getParameters(DataParameters.BIN);
        boolean elog_null = true;

        if (imageProcessing().updateBinarization(bin_params, elog_null,
                                                 (isEmpty() ? null : areaMask(0))));
        {
            setChanged(true);
        }
    }

    public int scoreImage() throws IOException
    {
        return imageProcessing().scoreImage(extendedMask(), 3, parameters().getParameter(Pre.min_range));
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    //////////////////  Exposing the public methods of m_parameter_container  //////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    public int getParameter(Any key)
    {
        return parameters().getParameter(key);
    }

    public int getBlackWhite(boolean get_white) throws IOException
    {
        if (get_white)
        {
            int white = parameters().getParameter(DataParameters.Pre.white);
            if (white < 0)
            {
                return imageProcessing().getBlackWhite()[1];
            }
            return white;
        }
        int black = parameters().getParameter(DataParameters.Pre.black);
        if (black < 0)
        {
            return imageProcessing().getBlackWhite()[0];
        }
        return black;
    }

    public int getElogThreshold() throws IOException
    {
        int threshold = parameters().getParameter(DataParameters.Elog.threshold);
        if (threshold < 0)
        {
            return imageProcessing().getElogThreshold();
        }
        return threshold;
    }

    /**
     * Does not use the method parameters(), so it is not affected by m_force_defaults.
     * @param key
     * @param number
     */
    public void setParameter(Any key, int number)
    {
        if (m_parameter_container == null)
        {
            m_parameter_container = getDefaultParameters();
        }
        m_parameter_container.completeFrom(getDefaultParameters());
        m_parameter_container.setParameter(key, number);
    }

    public Set<Any> keySet(int type)
    {
        return parameters().keySet(type);
    }

    ////////////////////////////////////////////////////////////////////////////////////////////////
    ///////////////////  Exposing the public methods of m_filament_container  //////////////////////
    ////////////////////////////////////////////////////////////////////////////////////////////////

    public void updateTraceParams()
    {
        Map<Any, Integer> trace_params = parameters().getParameters(DataParameters.TRACE);

        m_filament_container.updateTraceParams(trace_params);
    }

    public void cleanTracedFilaments()
    {
        m_filament_container.cleanTracedFilaments(true);
    }

    public void keepDiscardAllFilaments()
    {
        m_filament_container.keepDiscardAllFilaments();
    }

    public void remedyOldAngles()
    {
        m_filament_container.remedyOldAngles();
    }

    public Pair<Int2D, Integer> calculateWidthMap(double tolerance)
    {
        Tracer tracer = tracer();
        int max = tracer.calcWidthMap(m_image_processing.getBinaryImage(), tolerance);
        return new Pair<>(new Int2D(tracer.getWidthMap()), max);
    }

    public void scanFilaments() throws IOException
    {
        Tracer sensor = tracer();
        boolean[][] binary_image = imageProcessing().getBinaryImage();
        Map<Any, Integer> trace_params = parameters().getParameters(DataParameters.TRACE);

        boolean[][] interior = m_filament_container.scanFilaments(sensor, binary_image,
                                                                  areaMask(0), trace_params);
        if (trace_params.get(Trace.split) * trace_params.get(Trace.curve) > 0)
        {
            m_filament_container.splitToLinear();
        }
        m_interior_data = imageProcessing().areaAndMeanBrightness(interior, areaMask(0));
        cellData().calculateOrderParameter(sensor.getOrientationField());
        addAsProcessed();
        //new PopUpWiZer(calculateWiZer());
    }

    public BufferedImage[] calculateWiZer()
    {
        Map<Any, Integer> wizer_params = parameters().getParameters(DataParameters.WIZER);
        WrappedSiZer sizerer = new WrappedSiZer(m_filament_container.getFilaments());
        return sizerer.calculateWiZer(900, 600, 0.05, wizer_params);
    }

    public void calculateOrientationFields(boolean show_images) throws IOException
    {
        int[][] orientation_field = tracer().getOrientationField();
        OrientationFields calculator = new OrientationFields(cellData().getArea(), extendedMask(),
                                                             orientation_field,
                                                             m_filament_container.getFilaments());
        Pair<Int3D, Map<Integer, List<SingleFilament>>> pair = calculator.label(parameters().getParameters(DataParameters.ORI));
        if (pair == null)
        {
            return;
        }

        m_orientation_fields = pair.a.the;
        m_filaments_by_orientation_field = pair.b;
        for (Integer key : m_filaments_by_orientation_field.keySet())
        {
            FilamentSensor.debugMessage("Field " + key + " has " +
                                        m_filaments_by_orientation_field.get(key).size() + " filaments.");
        }

        determineOrientationFieldShapes();

        List<BufferedImage> images = new ArrayList<>();

        if (show_images)
        {
            for (int i = 1; i < m_orientation_fields.length; i++)
            {
                images.add(ImageFactory.makeColorImage(calculator.makeImage(m_orientation_fields[i]), true));
            }
            new PopUpOrientationFields(images);
            return;
        }
        else
        {
            
        }

        addAsProcessed();
    }

    public void setChanged(boolean changed)
    {
        m_filament_container.setChanged(changed);
    }

    public boolean hasChanged()
    {
        return m_filament_container.hasChanged();
    }

    public void addChangeListener(ChangeListener cl)
    {
        m_filament_container.addChangeListener(cl);
    }

    public List<SingleFilament> getSelectedFilaments()
    {
        return m_filament_container.getSelectedFilaments();
    }

    public List<AbstractFilament> getSelection()
    {
        return m_filament_container.getSelection();
    }

    public void clearSelection()
    {
        m_filament_container.clearSelection();
    }

    public void deleteSelected()
    {
        m_filament_container.deleteSelected();
    }

    public void setSelection(AbstractFilament selected)
    {
        m_filament_container.setSelection(selected);
    }

    public void setSelection(List<? extends AbstractFilament> list)
    {
        m_filament_container.setSelection(list);
    }

    public int countChains()
    {
        return m_filament_container.countChains();
    }

    public int countFilaments()
    {
        return m_filament_container.countFilaments();
    }

    public int countRows()
    {
        return m_filament_container.countRows();
    }

    public int countOrientationFields()
    {
        if (m_filaments_by_orientation_field == null)
        {
            return 0;
        }
        return m_filaments_by_orientation_field.size();
    }

    public AbstractFilament getEntryInRow(int row)
    {
        return m_filament_container.getEntryInRow(row);
    }

    public AbstractFilament changeEntryInRow(int row)
    {
        return m_filament_container.changeEntryInRow(row);
    }

    public int getRowOfEntry(AbstractFilament fil)
    {
        return m_filament_container.getRowOfEntry(fil);
    }

    public AbstractFilament wrap(AbstractFilament fil)
    {
        return m_filament_container.wrap(fil);
    }

    public List<SingleFilament> getFilaments()
    {
        return m_filament_container.getFilaments();
    }

    public List<SingleFilament> getFilamentsByOrientationField(int n)
    {
        if (m_filaments_by_orientation_field == null)
        {
            return null;
        }
        return m_filaments_by_orientation_field.get(n);
    }

    public FilamentChain getFilamentChain(int index)
    {
        return m_filament_container.getFilamentChain(index);
    }

    public List<FilamentChain> getFilamentChains()
    {
        return m_filament_container.getFilamentChains();
    }

    public void addFilamentChain(FilamentChain chain, boolean undoable)
    {
        m_filament_container.addFilamentChain(chain, undoable);
    }

    public boolean modifyAbstractFilament(AbstractFilament before, AbstractFilament after)
    {
        return m_filament_container.modifyAbstractFilament(before, after);
    }

    public void undoChain()
    {
        m_filament_container.undoChain();
    }

    public void redoChain()
    {
        m_filament_container.redoChain();
    }

    public BufferedImage makeImage(int type, boolean filaments, Color color) throws IOException
    {
        BufferedImage image;

        switch (type)
        {
        case ORIG:
        case PRE:
        case ELOG_COLOR:
        case ELOG_GREY:
        case BIN:
        case HOUGH:
            image = imageProcessing().getImage(type);
            break;
        case ORI_MAP:
            image = ImageFactory.makeColorImage(tracer().getOrientationField(), false);
            break;
        case FINGERPRINT:
            double axis = cellData().getLongHalfAxis() / 5e5;
            image = ImageFactory.makeDeepGreyImage(tracer().makeFingerprint(axis));
            break;
        case RED_GREEN:
            image = imageProcessing().getRedGreenImage();
            break;
        default:
            return null;
        }

        if (image == null)
        {
            return null;
        }

        if (filaments && (type == ORIG || type == PRE || type == BIN || type == RED_GREEN))
        {
            return addFilaments(image, m_filament_container.getFilaments(), color);
        }
        return image;
    }

    public BufferedImage makeImage(int type) throws IOException
    {
        switch (type)
        {
        case ORIG:
        case PRE:
        case ELOG_COLOR:
        case ELOG_GREY:
        case DIFF:
        case BIN:
        case HOUGH:
            return imageProcessing().getImage(type);
        }
        return null;
    }

    public BufferedImage redGreenImage(boolean[][] mask) throws IOException
    {
        if (mask == null)
        { 
            return null;
        }
        return ImageFactory.makeRedGreenImage(imageProcessing().getOriginalImage(), mask);
    }

    public BufferedImage excursionImage() throws IOException
    {
        return ImageFactory.makeThreeColorImage(cellData().getExcursionImage());
    }

    public List<BufferedImage> orientationFieldImages(Color color) throws IOException
    {
        FilamentSensor.debugMessage("Start making orientation field images.");
        List<BufferedImage> images = new ArrayList<>();

        BufferedImage tmp = makeImage(ORIG);
        for (Integer key : m_filaments_by_orientation_field.keySet())
        {
            FilamentSensor.debugMessage("Make image " + key);
            if (key <= 0)
            {
                images.add(0, addFilaments(tmp, m_filaments_by_orientation_field.get(key), color));
                continue;
            }
            images.add(addFilaments(tmp, m_filaments_by_orientation_field.get(key), color));
        }
        addAsProcessed();
        return images;
    }

    private static String imageToString (int[][] image)
    {
        String out = Arrays.deepToString(image);
        out = out.replaceAll("\\s+","").replace("],[", ls).replace("[", "").replace("]", "");
        return out;
    }

    private static BufferedImage addFilaments(BufferedImage image, List<SingleFilament> filaments,
                                              Color color)
    {
        BufferedImage output = new BufferedImage(image.getWidth(), image.getHeight(),
                                                 BufferedImage.TYPE_4BYTE_ABGR);
        Graphics2D graphics2d = output.createGraphics();
        graphics2d.drawImage(image, null, 0, 0);
        graphics2d.setColor(color);

        // draw filaments. TODO: Refactor to use Chains?
        for (SingleFilament filament : filaments)
        {
            graphics2d.setStroke(new BasicStroke((float) (filament.getWidth() / Const.MF),
                                                 BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
            List<Point> points = filament.getPoints();
            for (int i = 0; i < points.size() - 1; i++)
            {
                graphics2d.drawLine(points.get(i).x, points.get(i).y,
                                    points.get(i+1).x, points.get(i+1).y);
            }
        }

        graphics2d.dispose();
        return output;
    }

    private void determineOrientationFieldShapes() throws IOException
    {
        m_orientation_field_shapes = new int[m_orientation_fields.length][];
        boolean[][] mask = areaMask(0);
        for (int i = 1; i < m_orientation_fields.length; i++)
        {
            boolean[][] shape = new boolean[mask.length][mask[0].length];
            int width = 0;
            for (SingleFilament filament : m_filaments_by_orientation_field.get(i))
            {
                filament.drawToImage(shape);
                int this_width = (int)((filament.getWidth() - 1) / Const.M) + 1; // Ceiling
                if (this_width > width)
                {
                    width = this_width;
                }
            }
            // TODO: Include mask in morphological function to speed up.
            shape = Calc.dilateAndErode(shape, Calc.circleMask(100 + width), Calc.circleMask(100));
            for (int x = 0; x < shape.length; x++)
            {
                for (int y = 0; y < shape[0].length; y++)
                {
                    shape[x][y] = shape[x][y] && mask[x][y];
                }
            }
            m_orientation_field_shapes[i] = shapeMoments(shape);
            FilamentSensor.debugMessage("Field " + i + " data: " + m_orientation_field_shapes[i]);
        }
    }

    private static List<String> processOrientationField(int[] shape_data, int[][] field,
                                                        int n_blocks, List<SingleFilament> filaments)
    {
        // Moments
        double m_o = 0, m_oo = 0, m2_o = 0, m2_oo = 0;
        double n = field.length;
        for (int[] p : field)
        {
            m_o += p[2];
            m_oo += p[2] * p[2];
            int alt_ori = (p[2] + 90) % 180;
            m2_o += alt_ori;
            m2_oo += alt_ori * alt_ori;
        }
        m_o /= n;
        m2_o /= n;
        m_oo = Math.sqrt(m_oo / n - m_o * m_o);
        m2_oo = Math.sqrt(m2_oo / n - m2_o * m2_o);
        List<String> out = new ArrayList<>();
        out.add("center_x" + ls + shape_data[0] + ls);
        out.add("center_y" + ls + shape_data[1] + ls);
        out.add("area" + ls + (shape_data[2] / Const.MF) + ls);
        out.add("aspect_ratio" + ls + (shape_data[3] / Const.MF) + ls);
        out.add("long_half_axis" + ls + (shape_data[4] / Const.MF) + ls);
        out.add("alignment" + ls + (shape_data[5] / Const.MF) + ls);
        if (m2_oo < m_oo)
        {
            m_o = (m2_o + 90.) % 180.;
            m_oo = m2_oo;
        }
        out.add("orientation_mean" + ls + (Math.round(m_o * Const.MF) / Const.MF) + ls);
        out.add("orientation_sigma" + ls + (Math.round(m_oo * Const.MF) / Const.MF) + ls);
        
        int n_filaments = -1;
        double total_mass = -1;
        String filaments_text = "";
        if (filaments != null)
        {
            n_filaments = filaments.size();
            for (SingleFilament f : filaments)
            {
                total_mass += f.getMass();
                filaments_text += f.getCenter().x / 10.0 + "," + f.getCenter().y / 10.0 + "," +
                                  f.getLength() / Const.MF + "," + f.getOrientation() / Const.MF +
                                  "," + f.getWidth() / Const.MF + ls;
            }
        }
        out.add("number_of_filaments" + ls + n_filaments + ls);
        out.add("mass" + ls + (Math.round(total_mass * Const.MF) / Const.MF) + ls);
        out.add("filaments" + ls + filaments_text);
        return out;
    }

    private int[] shapeMoments(boolean[][] shape) throws IOException
    {
        final int width = shape.length,
                  height = shape[0].length;
        double sum = 0, sum_x = 0, sum_y = 0, sum_xx = 0, sum_xy = 0, sum_yy = 0;
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (shape[x][y])
                {
                    sum++;
                    sum_x += x;
                    sum_y += y;
                    sum_xx += x * x;
                    sum_xy += x * y;
                    sum_yy += y * y;
                }
            }
        }
        if (sum == 0)
        {
            FilamentSensor.debugError("No shape found!");
            return new int[]{-1, -1, -1, -1, -1, -1};
        }
        int mean_x = (int)Math.round(sum_x / sum);
        int mean_y = (int)Math.round(sum_y / sum);
        int area = (int)Math.round(sum * Const.MF / cellData().getArea());
        double cov_xx = sum_xx / sum - sum_x * sum_x / (sum * sum);
        double cov_xy = sum_xy / sum - sum_x * sum_y / (sum * sum);
        double cov_yy = sum_yy / sum - sum_y * sum_y / (sum * sum);
        double ew_hi = (cov_xx + cov_yy) / 2.0 +
                       Math.sqrt((cov_xx - cov_yy) * (cov_xx - cov_yy) / 4.0 + cov_xy * cov_xy);
        double ew_lo = (cov_xx + cov_yy) / 2.0 -
                       Math.sqrt((cov_xx - cov_yy) * (cov_xx - cov_yy) / 4.0 + cov_xy * cov_xy);
        int aspect_ratio = (int)Math.round(Math.sqrt(ew_hi / ew_lo) * Const.MF);
        int long_half_axis = (int)Math.round(Math.sqrt(ew_hi) * Const.MF);
        int phi = (int)Math.round((180 - Math.atan2(ew_hi - cov_xx, cov_xy) / Const.RAD) * Const.MF);
        return new int[]{mean_x, mean_y, area, aspect_ratio, long_half_axis, phi};
    }

    private int countBlocks()
    {
        List<Point> points = new ArrayList<>();
        for (int[][] field : m_orientation_fields)
        {
            for (int[] p : field)
            {
                Point tmp = new Point(p[0], p[1]);
                if (!points.contains(tmp))
                {
                    points.add(tmp);
                }
            }
        }
        return points.size();
    }

    private static class CompressedBinaryImage
    {
        int width, height;
        int[][] shape;

        CompressedBinaryImage (boolean[][] bin_img)
        {
            width = bin_img.length;
            height = bin_img[0].length;
            shape = new int[width][];
            for (int x = 0; x < width; x++)
            {
                List<Integer> list = new ArrayList<>();
                boolean current = false;
                for (int y = 0; y < height; y++)
                {
                    if (bin_img[x][y] != current)
                    {
                        list.add(y);
                        current = !current;
                    }
                }
                shape[x] = new int[list.size()];
                for (int i = 0; i < list.size(); i++)
                {
                    shape[x][i] = list.get(i);
                }
            }
        }

        private boolean[][] asArray ()
        {
            boolean[][] bin_img = new boolean[width][height];
            for (int x = 0; x < width; x++)
            {
                int n = shape[x].length;
                if (n == 1)
                {
                    continue;
                }
                        
                boolean current = false;
                for (int i = 0; i < n; i++)
                {
                    if (!current)
                    {
                        for (int y = shape[x][i]; y < (i+1 >= n ? height : shape[x][i+1]); y++)
                        {
                            bin_img[x][y] = true;
                        }
                    }
                    current = !current;
                }
            }
            return bin_img;
        }
    }
}
