package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2013-2015 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.image.BufferedImage;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JCheckBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.ScrollPaneConstants;
import javax.swing.SpinnerNumberModel;
import javax.swing.SwingConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

import core.DataParameters.Any;
import core.DataParameters.Bin;
import core.DataParameters.Ori;
import core.DataParameters.Pre;
import core.DataParameters.Trace;
import ui.Config;
import ui.Dialogs;
import ui.FilamentSensor;
import ui.IOImageData;
import ui.IOProjectData;
import ui.ImageFactory;
import ui.MainWindow;

public abstract class DataProject
{
    protected static List<Files> m_files;
    protected static List<DataImage> m_processed_images;
    protected static boolean m_force_defaults;
    protected static DataParameters m_default_parameters;
    protected static Map<Integer, File> m_file_dictionary;

    public static boolean getForce()
    {
        return m_force_defaults;
    }

    public static void setForce(boolean force)
    {
        m_force_defaults = force;
    }

    public static DataParameters getDefaultParameters()
    {
        return (DataParameters) m_default_parameters.clone();
    }

    public static void setDefaultParameters(DataParameters params)
    {
        m_default_parameters = ((DataParameters) params.clone()).completeFrom(new DataParameters());
    }

    private static Color colorDialog(String message)
    {
        HashMap<String, Color> map = new HashMap<String, Color>();
        map.put("Orange", Color.orange);
        map.put("Dark Orange", new Color(255, 140, 0));
        map.put("Cyan", Color.cyan);
        map.put("Blue", Color.blue);

        Object[] options = map.keySet().toArray();
        String color_name = (String) JOptionPane.showInputDialog(MainWindow.frame(), message, "Choose Color",
                                    JOptionPane.QUESTION_MESSAGE, null, options, options[0]);
        return map.get(color_name);
    }

    /**
     * This method traces filaments for all image files in m_files.
     */
    public static void batchProcessing(DataImage caller)
    {
        // Save dialog for csv files and filament images dialog to decide
        // whether orientation field images should be saved.
        File csv_path = Dialogs.save(true, "Folder to save csv.");
        if (csv_path == null) { return; }
        Config.set("LastFolders", "Save", csv_path.getAbsolutePath());
        Color color = colorDialog("Choose the color for filaments in the output images.");
        File image_path = Dialogs.save(true, "Folder to save images with filaments.");
        if (image_path == null) { return; }
        boolean[] flags = flagsDialog(flagTable(), new int[]{0,1,2,3,4,5});

        // Batch
        long time = -System.nanoTime();
        List<DataImage> batched = new ArrayList<>();
        boolean[][] previous_area = null;
        Collections.sort(m_files);
        for (Files pair : m_files)
        {
            DataImage data = null;

            // Prefer data from the current session, default to data from XML files.
            for (DataImage data_changed : m_processed_images)
            {
                if (pair.image().getName().equals(data_changed.getImageName()))
                {
                    data = data_changed;
                }
            }
            data = batchStep(data, pair, flags, image_path, color, previous_area);
            if(data == null)
            {
                continue;
            }
            if (flags[3])
            {
                try
                {
                    previous_area = data.areaMask(5);
                    //new PopUpDebug(data.areaMask(0), false, pair.image().getName());
                }
                catch (IOException e)
                {
                    FilamentSensor.debugError("Corrupt DataImage on mask creation attempt.");
                }
            }

            if (data.getImageName().equals(caller.getImageName()))
            {
                FilamentSensor.debugMessage("Identified " + data.getImageName() + " as caller.");
                caller = data;
            }
            else
            {
                FilamentSensor.debugMessage("Shutting down " + data.getImageName());
                data.shutDown();
            }
            batched.add(data);
        }

        batchSave(batched, flags, csv_path, image_path, color);
        time = Math.round((System.nanoTime() + time) * (double)1e-9);
        String msg = String.format("Processing time: %d:%02d:%02d", time/3600, (time/60)%60, time%60);
        JOptionPane.showMessageDialog(MainWindow.frame(), msg, "Batch processing done!",
                                      JOptionPane.INFORMATION_MESSAGE, null);
    }

    public static void metaBatch()
    {
        Color color = colorDialog("Choose the color for filaments in the output images.");
        boolean[] flags = flagsDialog(flagTable(), new int[]{0,1,2,4});

        MetaBatchData meta = setupMetaBatch();
        List<File[]> projects = meta.project_table;
        DataParameters parameters = meta.parameters;
        boolean update = meta.update;

        long time = -System.nanoTime();
        for (File[] info : projects)
        {
            IOProjectData project = (update ? new IOProjectData(info[0], info[1], info[2]) :
                                     new IOProjectData(info[1]));
            if (parameters != null)
            {
                m_force_defaults = false;
                m_default_parameters = parameters;
            }
            batchRun(info[3], info[4], parameters, color, flags);
            if (!project.saveProject(info[0]))
            {
                JOptionPane.showMessageDialog(MainWindow.frame(), "This project could not be saved: " + info[0],
                                              "Project not saved!", JOptionPane.WARNING_MESSAGE, null);
            }
            System.out.println(info[0].getName() + " done!");
        }
        time = Math.round((System.nanoTime() + time) * (double)1e-9);
        String msg = String.format("Processing time: %d:%02d:%02d", time/3600, (time/60)%60, time%60);
        JOptionPane.showMessageDialog(MainWindow.frame(), msg, "Batch processing done!",
                                      JOptionPane.INFORMATION_MESSAGE, null);
    }

    public static void areaBatch()
    {
        List<File[]> projects = new ArrayList<>();

        while (true)
        {
            File image_folder = Dialogs.load(true, "Determine input image folder");
            if (image_folder != null && image_folder.exists() && image_folder.canRead() &&
                image_folder.isDirectory())
            {
                File image_save = Dialogs.save(true, "Folder to save area images.");
                File csv_save = Dialogs.save(true, "Folder to save csv.");
                projects.add(new File[]{image_folder, image_save, csv_save});
            }
            else
            {
                JOptionPane.showMessageDialog(MainWindow.frame(), "This project could not be added.",
                                              "Project not added!", JOptionPane.WARNING_MESSAGE, null);
            }

            String overview = "Common path, Input image folder, Output image folder, CSV Folder" + Const.ls;
            for (File[] info : projects)
            {
                for (String s : extractCommonAncestor(info))
                {
                    overview += s + ", ";
                }
                overview = overview.substring(0, overview.length()-2) + Const.ls;
            }
            overview += Const.ls + "Would you like to add another project to the batch list?";

            if (JOptionPane.showOptionDialog(MainWindow.frame(), overview, "Add another project?",
                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null) ==
                    JOptionPane.NO_OPTION)
            {
                break;
            }
        }
        String overview = "Common path, Input image folder, Output image folder" + Const.ls;
        for (File[] info : projects)
        {
            for (String s : extractCommonAncestor(info))
            {
                overview += s + ", ";
            }
            overview = overview.substring(0, overview.length()-2) + Const.ls;
        }
        overview += Const.ls + "Do you want to start batch processing?";
        if (JOptionPane.showOptionDialog(MainWindow.frame(), overview, "Batch Plan",
                JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null) ==
                JOptionPane.NO_OPTION)
        {
            return;
        }

        long time = -System.nanoTime();
        for (File[] info : projects)
        {
            if (info[0] == null || !info[0].exists() || !info[0].canRead() ||
                !info[0].isDirectory())
            {
                FilamentSensor.debugError("Invalid folder: " + info[0].getAbsolutePath() + "! Skipping...");
                continue;
            }
            FilamentSensor.debugMessage("Image folder: " + info[0].getAbsolutePath());
            new IOProjectData(info[0]);
            Collections.sort(m_files);
            for (Files pair : m_files)
            {
                DataImage data = null;
                try
                {
                    data = new DataImage(pair.image(), null, null);
                    BufferedImage image = ImageFactory.makeBinaryImage(data.areaMask(0), false);
                    data.saveImage(info[1], "area", image, true);
                    data.exportCSV(info[2], false, false);
                }
                catch (IOException e)
                {
                    JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + pair.image(),
                                                  "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
                }
                data.fullShutDown();
            }
        }
        time = Math.round((System.nanoTime() + time) * (double)1e-9);
        String msg = String.format("Processing time: %d:%02d:%02d", time/3600, (time/60)%60, time%60);
        JOptionPane.showMessageDialog(MainWindow.frame(), msg, "Batch processing done!",
                                      JOptionPane.INFORMATION_MESSAGE, null);
    }

    public static void redGreenBatch()
    {
        List<File[]> projects = new ArrayList<>();

        while (true)
        {
            File image_folder = Dialogs.load(true, "Determine input image folder");
            if (image_folder != null && image_folder.exists() && image_folder.canRead() &&
                image_folder.isDirectory())
            {
                File image_save = Dialogs.save(true, "Folder to save area images.");
                projects.add(new File[]{image_folder, image_save});
            }
            else
            {
                JOptionPane.showMessageDialog(MainWindow.frame(), "This project could not be added.",
                                              "Project not added!", JOptionPane.WARNING_MESSAGE, null);
            }

            String overview = "Common path, Input image folder, Output image folder" + Const.ls;
            for (File[] info : projects)
            {
                for (String s : extractCommonAncestor(info))
                {
                    overview += s + ", ";
                }
                overview = overview.substring(0, overview.length()-2) + Const.ls;
            }
            overview += Const.ls + "Would you like to add another project to the batch list?";

            if (JOptionPane.showOptionDialog(MainWindow.frame(), overview, "Add another project?",
                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null) ==
                    JOptionPane.NO_OPTION)
            {
                break;
            }
        }
        String overview = "Common path, Input image folder, Output image folder" + Const.ls;
        for (File[] info : projects)
        {
            for (String s : extractCommonAncestor(info))
            {
                overview += s + ", ";
            }
            overview = overview.substring(0, overview.length()-2) + Const.ls;
        }
        overview += Const.ls + "Do you want to start batch processing?";
        if (JOptionPane.showOptionDialog(MainWindow.frame(), overview, "Batch Plan",
                JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null) ==
                JOptionPane.NO_OPTION)
        {
            return;
        }

        long time = -System.nanoTime();
        for (File[] info : projects)
        {
            if (info[0] == null)
            {
                FilamentSensor.debugError("Invalid folder! Skipping...");
                continue;
            }
            new IOProjectData(info[0]);
            Collections.sort(m_files);
            for (Files pair : m_files)
            {
                DataImage data = null;
                try
                {
                    data = new DataImage(pair.image(), null, null);
                    data.saveImage(info[1], "area", data.redGreenImage(data.areaMask(0)), true);
                }
                catch (IOException e)
                {
                    JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + pair.image(),
                                                  "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
                }
                data.fullShutDown();
            }
        }
        time = Math.round((System.nanoTime() + time) * (double)1e-9);
        String msg = String.format("Processing time: %d:%02d:%02d", time/3600, (time/60)%60, time%60);
        JOptionPane.showMessageDialog(MainWindow.frame(), msg, "Batch processing done!",
                                      JOptionPane.INFORMATION_MESSAGE, null);
    }

    private static void batchRun(File image_path, File csv_path, DataParameters parameters,
                                 Color color, boolean[] flags)
    {
        if (csv_path == null || image_path == null)
        {
            return;
        }

        // Batch
        List<DataImage> batched = new ArrayList<>();
        boolean[][] previous_area = null;
        Collections.sort(m_files);
        for (Files pair : m_files)
        {
            DataImage data = batchStep(null, pair, flags, image_path, color, previous_area);

            if(data == null)
            {
                continue;
            }
            if (flags[3])
            {
                try
                {
                    previous_area = data.areaMask(5);
                    //new PopUpDebug(data.areaMask(0), false, pair.image().getName());
                }
                catch (IOException e)
                {
                    FilamentSensor.debugError("Corrupt DataImage on mask creation attempt.");
                }
            }

            FilamentSensor.debugMessage("Shutting down " + data.getImageName());
            data.shutDown();
            batched.add(data);
        }

        batchSave(batched, flags, csv_path, image_path, color);
    }
    
    private static DataImage batchStep(DataImage data, Files pair, boolean[] flags, File image_path,
                                       Color color, boolean[][] mask)
    {
        try
        {
            if (data == null) // This is always the case in meta-batch.
            {
                if (pair.data() == null)
                {
                    data = new DataImage(pair.image(), getDefaultParameters(), mask);
                }
                else
                {
                    IOImageData file = new IOImageData(pair.data());
                    data = file.read(pair.image().getParent(), getDefaultParameters(), getForce());
                    data.setMask(mask);
                }
            }

            if (data.isEmpty())
            {
                FilamentSensor.debugMessage(data.getImageName() + " is empty.");
                data.saveImage(image_path, "filaments", DataImage.ORIG, false, color, true);
                data.fullShutDown();
                return null;
            }
            data.updatePreprocessing();
            data.updateBinarization();
            data.scanFilaments();
            if (flags[0])
            {
                data.calculateOrientationFields(false);
            }
            if (flags[4])
            {
                data.exportFingerprint(image_path);
            }
            data.saveImage(image_path, "filaments", DataImage.ORIG, true, color, true);
            if (data.getParameter(Trace.no_boundary) == 1)
            {
                data.saveImage(image_path, "interior", DataImage.RED_GREEN, true, color, true);
            }
            
        }
        catch (IOException e)
        {
            JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + pair.image(),
                                          "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
            return null;
        }
        return data;
    }

    private static void batchSave(List<DataImage> batched, boolean[] flags, File csv_path,
                                   File image_path, Color color)
    {
        // Post processing
        if (flags[0] && flags[1])
        {
            orientationFieldPostProcessing(batched);
        }

        for (DataImage data : batched)
        {
            List<BufferedImage> images = null;
            try
            {
                images = (flags[0] && flags[2] ? data.orientationFieldImages(color) :
                                                 new ArrayList<BufferedImage>());
                data.exportCSV(csv_path, false, false);
                FilamentSensor.debugMessage("Shutting down " + data.getImageName());
                data.fullShutDown();
            }
            catch (IOException e)
            {
                JOptionPane.showMessageDialog(MainWindow.frame(), "Images or csv could not be saved: " + data.getImageName(),
                                              "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
                continue;
            }
            if (!images.isEmpty())
            {
                data.saveImage(image_path, "noise", images.remove(0), true);
                int i = 1;
                for (BufferedImage image : images)
                {
                    data.saveImage(image_path, "orientation_field" + i++, image, true);
                }
            }
        }
    }

    private static DataParameters parameterChoiceDialog()
    {
        DataParameters parameters = new DataParameters();
        List<ParameterField> table = new ArrayList<>();

        table.add(new ParameterField("Preprocessing", null, null));
        table.add(new ParameterField("Minimal required brightness range", Pre.min_range, parameters));
        table.add(new ParameterField("Manual contrast?", Pre.contrast, parameters));
        table.add(new ParameterField("Manual Black value", Pre.black, parameters));
        table.add(new ParameterField("Manual White value", Pre.white, parameters));
        table.add(new ParameterField("10x Gauss Sigma", Pre.fsigma, parameters));
        table.add(new ParameterField("Laplace Neighborhood: 0 = 4-neighborhood, 1 = 8-neighborhood   ", Pre.lpmask, parameters));
        table.add(new ParameterField("10x Laplace Factor", Pre.lpfac, parameters));
        table.add(new ParameterField("10x Line Gaussian Sigma", Pre.line_sigma, parameters));
        table.add(new ParameterField("Binarization", null, null));
        table.add(new ParameterField("Minimum Brightness", Bin.minmean, parameters));
        table.add(new ParameterField("10x Binarization Sigma", Bin.sigma, parameters));
        table.add(new ParameterField("10x Noise Filter Threshold", Bin.rod_significance, parameters));
        table.add(new ParameterField("Restrict to area mask?", Bin.restrict, parameters));
        table.add(new ParameterField("Tracing", null, null));
        table.add(new ParameterField("Curved Filaments?", Trace.curve, parameters));
        table.add(new ParameterField("Minimum Line Length", Trace.minlen, parameters));
        table.add(new ParameterField("Minimum Crossing Angle", Trace.minangle, parameters));
        table.add(new ParameterField("Length of linear pieces", Trace.step, parameters));
        table.add(new ParameterField("Pixel Width Tolerance", Trace.tolerance, parameters));
        table.add(new ParameterField("Orientation Fields", null, null));
        table.add(new ParameterField("10x Orientation Field Minimum Area", Ori.min_area, parameters));
        table.add(new ParameterField("Orientation Field Minimum Filaments", Ori.min_filaments, parameters));
        
        Object[] params = new Object[table.size()];
        for (int i = 0; i < table.size(); i++)
        {
            ParameterField field = table.get(i);
            JPanel panel = new JPanel();
            panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
            panel.setMaximumSize(new Dimension(400, 28));
            if (field.m_spinbox != null)
            {
                panel.add(new JLabel(field.m_label, SwingConstants.LEFT));
                panel.add(Box.createHorizontalGlue());
                panel.add(field.m_spinbox);
            }
            else
            {
                JLabel label = new JLabel(field.m_label, SwingConstants.CENTER);
                Font orig_font = label.getFont();
                label.setFont(orig_font.deriveFont(Font.BOLD, panel.getFontMetrics(orig_font).getHeight()+2));
                panel.add(Box.createHorizontalGlue());
                panel.add(label);
                panel.add(Box.createHorizontalGlue());
            }
            params[i] = panel;
        }
        JOptionPane.showMessageDialog(MainWindow.frame(), params, "Please choose parameters.",
                                      JOptionPane.INFORMATION_MESSAGE, null);
        for (int i = 0; i < table.size(); i++)
        {
            ParameterField field = table.get(i);
            if (field.m_spinbox != null)
            {
                parameters.setParameter(field.m_parameter, (Integer)field.m_spinbox.getValue());
            }
        }
        return parameters;
    }

    private static boolean[] flagsDialog(String[] table, int[] to_use)
    {
        Object[] params = new Object[to_use.length];
        List<JCheckBox> checkboxes = new ArrayList<>();
        for (int i = 0; i < to_use.length; i++)
        {
            JPanel panel = new JPanel();
            panel.setLayout(new BoxLayout(panel, BoxLayout.X_AXIS));
            panel.setMaximumSize(new Dimension(400, 28));
            JCheckBox box = new JCheckBox(table[to_use[i]]);
            box.setSelected(true);
            checkboxes.add(box);
            panel.add(box);
            panel.add(Box.createHorizontalGlue());

            params[i] = panel;
        }
        JOptionPane.showMessageDialog(MainWindow.frame(), params, "Please choose batch settings.",
                                      JOptionPane.INFORMATION_MESSAGE, null);
        boolean[] flags = new boolean[table.length];
        for (int i = 0; i < checkboxes.size(); i++)
        {
            flags[to_use[i]] = checkboxes.get(i).isSelected();
        }
        return flags;
    }

    private static String[] flagTable()
    {
        return new String[]{"Determine orientation fields?",
                            "Post-Process orientation fields as time series?",
                            "Save orientation field images?",
                            "Only small changes in cell shape at each step?",
                            "Calculate fingerprints?",
                            "Save Excursion Images?"};
    }

    public static void batchElog(DataImage caller)
    {
        // Save dialogs for csv and filament images.
        File csv_path = Dialogs.save(true, "Folder to save histogram csv.");
        File image_path = Dialogs.save(true, "Folder to save eLoG images.");

        if (csv_path == null || image_path == null)
        {
            return;
        }

        Config.set("LastFolders", "Save", csv_path.getAbsolutePath());

        // Batch
        Collections.sort(m_files);
        for (Files pair : m_files)
        {
            DataImage data = null;

            // Prefer data from the current session, default to data from XML files.
            for (DataImage data_changed : m_processed_images)
            {
                if (pair.image().getName().equals(data_changed.getImageName()))
                {
                    data = data_changed;
                }
            }
            try
            {
                if (data == null)
                {
                    if (pair.data() == null)
                    {
                        data = new DataImage(pair.image(), getDefaultParameters(), null);
                    }
                    else
                    {
                        IOImageData file = new IOImageData(pair.data());
                        data = file.read(pair.image().getParent(), getDefaultParameters(), getForce());
                    }
                }
    
                String parameters = data.calculateELoGs();
                data.exportHistogram(csv_path, false, parameters, DataImage.ELOG_COLOR);
                data.saveImage(image_path, "eLoG", DataImage.ELOG_COLOR, false, null, true);
            }
            catch (IOException e)
            {
                JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + pair.image(),
                                              "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
                continue;
            }
            

            if (data.getImageName().equals(caller.getImageName()))
            {
                FilamentSensor.debugMessage("Identified " + data.getImageName() + " as caller.");
                caller = data;
            }
            else
            {
                FilamentSensor.debugMessage("Shutting down " + data.getImageName());
                data.fullShutDown();
            }
        }
    }

    public static void batchHough(DataImage caller)
    {
        // Save dialogs for csv and filament images.
        File csv_path = Dialogs.save(true, "Folder to save histogram csv.");
        File image_path = Dialogs.save(true, "Folder to save Hough images.");

        if (csv_path == null || image_path == null)
        {
            return;
        }

        Config.set("LastFolders", "Save", csv_path.getAbsolutePath());

        // Batch
        Collections.sort(m_files);
        for (Files pair : m_files)
        {
            DataImage data = null;

            // Prefer data from the current session, default to data from XML files.
            for (DataImage data_changed : m_processed_images)
            {
                if (pair.image().getName().equals(data_changed.getImageName()))
                {
                    data = data_changed;
                }
            }
            try
            {
                if (data == null)
                {
                    if (pair.data() == null)
                    {
                        data = new DataImage(pair.image(), getDefaultParameters(), null);
                    }
                    else
                    {
                        IOImageData file = new IOImageData(pair.data());
                        data = file.read(pair.image().getParent(), getDefaultParameters(), getForce());
                    }
                }
    
                data.updatePreprocessing();
                data.updateBinarization();
                String parameters = data.calculateHough();
                data.exportHistogram(csv_path, false, parameters, DataImage.HOUGH);
                data.saveImage(image_path, "Hough", DataImage.HOUGH, false, null, true);
            }
            catch (IOException e)
            {
                JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + pair.image(),
                                              "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
                continue;
            }

            if (data.getImageName().equals(caller.getImageName()))
            {
                FilamentSensor.debugMessage("Identified " + data.getImageName() + " as caller.");
                caller = data;
            }
            else
            {
                FilamentSensor.debugMessage("Shutting down " + data.getImageName());
                data.fullShutDown();
            }
        }
    }

    public static void updateProject()
    {
        IOProjectData project = new IOProjectData(true);
        File csv_path = Dialogs.save(true, "Folder to save csv.");
        File img_save_path = Dialogs.save(true, "Folder to save images.");
        boolean[] flags = flagsDialog(flagTable(), new int[]{0,1,2,4,5});
        Color color = (flags[0] && flags[2] ? colorDialog("Choose the color for filaments in the output images.")
                                            : null);
        if (!project.getImagePath().exists() || csv_path == null || !csv_path.exists())
        {
            return;
        }
        String image_path = project.getImagePath().getAbsolutePath();
        List<DataImage> data_list = new ArrayList<>();
        for (int n = 0; n < project.getLength(); n++)
        {
            File data_file = project.getDataFile(n);
            if (data_file != null)
            {
                IOImageData xml_file = new IOImageData(data_file);
                try
                {
                    DataImage data = xml_file.read(image_path, DataProject.getDefaultParameters(), false);
                    if (flags[0])
                    {
                        data.calculateOrientationFields(false);
                    }
                    if (flags[5])
                    {
                        data.saveImage(img_save_path, "excursions", data.excursionImage(), true);
                    }
                    data.shutDown();
                    data_list.add(data);
                }
                catch (IOException ex)
                {
                    JOptionPane.showMessageDialog(MainWindow.frame(), "No image could be loaded for file: " + data_file,
                                                  "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
                }
                FilamentSensor.verboseMessage(data_file.getName() + " updated.");
            }
        }

        batchSave(data_list, flags, csv_path, img_save_path, color);

        project.overwrite();
        JOptionPane.showMessageDialog(MainWindow.frame(), "Upgrade complete. Returning to FilamentSensor.",
                                      "Project updated!", JOptionPane.INFORMATION_MESSAGE, null);
    }

    public static void projectIllustrations()
    {
        IOProjectData project = new IOProjectData(true);
        File image_folder = Dialogs.save(true, "Folder to save images.");
        
        if (!project.getImagePath().exists() || image_folder == null || !image_folder.exists())
        {
            return;
        }
        String image_path = project.getImagePath().getAbsolutePath();
        List<DataImage> data_list = new ArrayList<>();
        for (int n = 0; n < project.getLength(); n++)
        {
            File data = project.getDataFile(n);
            if (data != null)
            {
                IOImageData xml_file = new IOImageData(data);
                try
                {
                    data_list.add(xml_file.read(image_path, DataProject.getDefaultParameters(), false));
                    data_list.get(data_list.size()-1).makeIllustrations(image_folder);
                    data_list.get(data_list.size()-1).fullShutDown();
                }
                catch (IOException ex)
                {
                    JOptionPane.showMessageDialog(MainWindow.frame(), "No image could be loaded for file: " + data,
                                                  "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
                }
                FilamentSensor.verboseMessage(data.getName() + " updated.");
            }
        }
        JOptionPane.showMessageDialog(MainWindow.frame(), "Illustrations complete. Returning to FilamentSensor.",
                                      "Illustrations done!", JOptionPane.INFORMATION_MESSAGE, null);
    }

    private static void orientationFieldPostProcessing(List<DataImage> data_list)
    {
        FilamentSensor.debugMessage("Start orientation field post-processing.");
        boolean changed = true;
        OrientationFieldPostProcessing post = new OrientationFieldPostProcessing();
        while (changed)
        {
            changed = false;
            for (int i = 1; i < data_list.size(); i++)
            {
                int[][][] set1 = data_list.get(i-1).getOrientationFields(),
                          set2 = data_list.get(i).getOrientationFields();
                if (post.sort(set1, set2))
                {
                    try
                    {
                        data_list.get(i-1).setOrientationFields(post.get(false));
                        data_list.get(i).setOrientationFields(post.get(true));
                        changed = true;
                    }
                    catch (IOException ex)
                    {
                        JOptionPane.showMessageDialog(MainWindow.frame(), "Image loading error in orientation field post processing, images " + (i-1) + " and " + i + ". Skipping!",
                                                      "Images skipped!", JOptionPane.WARNING_MESSAGE, null);
                    }
                }
            }
            for (int i = data_list.size() - 1; i > 0; i--)
            {
                int[][][] set1 = data_list.get(i-1).getOrientationFields(),
                          set2 = data_list.get(i).getOrientationFields();
                if (post.sort(set1, set2))
                {
                    try
                    {
                        data_list.get(i-1).setOrientationFields(post.get(false));
                        data_list.get(i).setOrientationFields(post.get(true));
                        changed = true;
                    }
                    catch (IOException ex)
                    {
                        JOptionPane.showMessageDialog(MainWindow.frame(), "Image loading error in orientation field post processing, images " + (i-1) + " and " + i + ". Skipping!",
                                                      "Images skipped!", JOptionPane.WARNING_MESSAGE, null);
                    }
                }
            }
            if (changed)
            {
                FilamentSensor.debugMessage("Orientation fields changed. Re-running.");
            }
        }
        FilamentSensor.debugMessage("Orientation field post-processing finished.");
    }

    private static String[] extractCommonAncestor(File[] files)
    {
        if (files == null)
        {
            return new String[]{""};
        }

        List<String> common_path = null; new ArrayList<>();
        File common_folder = null;

        // Split paths into lists of strings and determine common path.
        for (File file : files)
        {
            File f = file;
            if (f == null)
            {
                return new String[]{""};
            }
            if (common_folder == null)
            {
                common_folder = f;
            }
            List<String> path = new ArrayList<>();
            String name = f.getName();
            while (!name.isEmpty())
            {
                path.add(0, name);
                f = f.getParentFile();
                name = f.getName();
            }
            if (common_path == null)
            {
                common_path = path;
            }
            else
            {
                if (path.size() < common_path.size())
                {
                    common_path = common_path.subList(0, path.size());
                    common_folder = file;
                }
                for (int i = common_path.size() - 1; i >= 0; i--)
                {
                    if (path.get(i).equals(common_path.get(i)))
                    {
                        break;
                    }
                    common_path.remove(i);
                    common_folder = common_folder.getParentFile();
                }
            }
        }

        String[] strings = new String[files.length + 1];
        strings[0] = common_folder.getAbsolutePath();
        for (int i = 0; i < files.length; i++)
        {
            strings[i+1] = files[i].getAbsolutePath().substring(strings[0].length());
            if (strings[i+1].equals(""))
            {
                strings[i+1] = ".";
            }
        }
        return strings;
    }

    private static File[] getLocations()
    {
        File image_folder = Dialogs.load(true, "Image folder."),
             project_file = Dialogs.save(false, "Project file name.");
        return new File[]{project_file, image_folder, project_file.getParentFile(),
                          Dialogs.save(true, "Folder to save images with filaments."),
                          Dialogs.save(true, "Folder to save csv.")};
    }

    private static File[] defaultLocations()
    {
        File image_folder = Dialogs.load(true, "Image folder.");
        String root = image_folder.getAbsolutePath() + File.separator;
        File project_file = new File(root + "xml" + File.separator +
                                     image_folder.getName() + ".xml"),
             xml_folder = project_file.getParentFile(),
             img_folder = new File(root + "img"),
             csv_folder = new File(root + "csv");
        xml_folder.mkdirs();
        img_folder.mkdirs();
        csv_folder.mkdirs();
        return new File[]{project_file, image_folder, xml_folder, img_folder, csv_folder};
    }

    private static boolean saveMetaBatch (boolean update, List<File[]> project_table,
                                          DataParameters parameters)
    {
        File file = Dialogs.save(false, "Save batch plan.");
        
        DocumentBuilder document_builder = null;
        try
        {
            DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
            document_builder = docFactory.newDocumentBuilder();
        }
        catch (ParserConfigurationException e)
        {
            System.err.println("FilamentSensor DataProject.saveMetaBatch(): " +
                               "Failed to create DocumentBuilder: " + e.getMessage());
            return false;
        }
        Document document = document_builder.newDocument();
        Element root = document.createElement("FilamentSensorBatchPlan");
        document.appendChild(root);

        // Version information
        Element element = document.createElement("program");
        root.appendChild(element);
        element.setAttribute("name", FilamentSensor.NAME);
        element.setAttribute("version", FilamentSensor.VERSION);

        // Paths of projects.
        element = document.createElement("projects");
        root.appendChild(element);
        element.setAttribute("update", "" + (update ? 1 : 0));
        int i = 0;
        for (File[] files : project_table)
        {
            Element this_project = document.createElement("project" + i++);
            element.appendChild(this_project);
            int j = 0;
            for (File f : files)
            {
                this_project.setAttribute("file" + j++, f.getAbsolutePath());
            }
        }
        if (parameters != null)
        {
            element = document.createElement("parameters");
            root.appendChild(element);
            for (int type : DataParameters.ALL)
            {
                Element parameter_type = document.createElement(DataParameters.getName(type));
                element.appendChild(parameter_type);
                for (Any key : parameters.keySet(type))
                {
                    parameter_type.setAttribute(key.name(), "" + parameters.getParameter(key));
                }
            }
        }
        
        try
        {
            // write the created document object to file
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            BufferedWriter writer = new BufferedWriter(new FileWriter(file));
            StreamResult result = new StreamResult(writer);
            DOMSource source = new DOMSource(document);
            transformer.transform(source, result);
            writer.close();
        }
        catch (Exception e) // IOException or
                            // TransformerConfigurationException
        {
            FilamentSensor.debugError("Error while saving '" + file + "': " + e.getMessage());
            return false;
        }
        return true;
    }

    private static MetaBatchData loadMetaBatch ()
    {
        File file = Dialogs.load(false, "Load batch plan.");
        
        Document document = null;
        try
        {
            DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
            DocumentBuilder document_builder = docFactory.newDocumentBuilder();
            document = document_builder.parse(file);
        }
        catch (ParserConfigurationException | SAXException | IOException e)
        {
            System.err.println("FilamentSensor DataProject.loadMetaBatch(): " +
                               "Failed to create DocumentBuilder: " + e.getMessage());
            return null;
        }
        
        Element root = document.getDocumentElement();
        if (!root.getTagName().equals("FilamentSensorBatchPlan"))
        {
            System.err.println("The specified file is valid XML, but does not seem to be a valid FilamentSensor batch plan...");
            return null;
        }

        Element program = (Element)root.getElementsByTagName("program").item(0);
        if (!program.getAttribute("version").equals(FilamentSensor.VERSION))
        {
            JOptionPane.showMessageDialog(MainWindow.frame(), "Batch plan is from old FilamentSensor version: " + program.getAttribute("version"),
                                          "Old version!", JOptionPane.WARNING_MESSAGE, null);
        }

        Element projects = (Element)root.getElementsByTagName("projects").item(0);
        boolean update = (getInt(projects, "update").equals(1));

        List<File[]> project_table = new ArrayList<>();
        for (Node pr = projects.getFirstChild(); pr != null; pr = pr.getNextSibling())
        {
            if (!pr.getNodeName().contains("project"))
            {
                continue;
            }
            Element info = (Element)pr;
            int n_files = 0;
            for (int i = 0; true; i++)
            {
                if (!info.hasAttribute("file" + i))
                {
                    n_files = i;
                    break;
                }
            }
            File[] files = new File[n_files];
            for (int i = 0; i < n_files; i++)
            {
                files[i] = new File(info.getAttribute("file" + i));
            }
            project_table.add(files);
        }

        Element parameters = (Element)root.getElementsByTagName("parameters").item(0);

        return new MetaBatchData(update, project_table, DataParameters.fromXML(parameters));
    }

    private static MetaBatchData setupMetaBatch ()
    {
        if (JOptionPane.showOptionDialog(MainWindow.frame(), "Load previously saved batch plan?",
                "Load Batch Plan?", JOptionPane.YES_NO_OPTION,
                JOptionPane.QUESTION_MESSAGE, null, null, null) == JOptionPane.YES_OPTION)
        {
            MetaBatchData meta = loadMetaBatch();
            String overview = "<html><body>Common path, Project file, Input image folder, XML folder, " +
                    "Output image folder, CSV Folder<br>";
            for (File[] info : meta.project_table)
            {
                for (String s : extractCommonAncestor(info))
                {
                    overview += s + ", ";
                }
                overview = overview.substring(0, overview.length()-2) + "<br>";
            }
            overview += "<br>Do you want to start batch processing?";
            JScrollPane scroller = new JScrollPane(new JLabel(overview + "</body></html>"),
                                                   ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                                                   ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
            scroller.setBorder(BorderFactory.createEmptyBorder());

            if (JOptionPane.showOptionDialog(MainWindow.frame(), scroller, "Batch Plan",
                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null) ==
                    JOptionPane.NO_OPTION)
            {
                return null;
            }
            return meta;
        }
        List<File[]> projects = new ArrayList<>();
        DataParameters parameters = null;

        boolean update = (JOptionPane.showOptionDialog(MainWindow.frame(),
                              "Would you like to process new images or update existing projects?",
                              "Update or new?", JOptionPane.DEFAULT_OPTION,
                              JOptionPane.QUESTION_MESSAGE, null,
                              new String[] {"New", "Update"}, null) == 1);
        boolean default_folders = false;
        if (!update)
        {
            parameters = parameterChoiceDialog();
            default_folders = (JOptionPane.showOptionDialog(MainWindow.frame(),
                                   "Would you like to use default folder and file names?",
                                   "Default folders?", JOptionPane.YES_NO_OPTION,
                                   JOptionPane.QUESTION_MESSAGE, null, null, null) == JOptionPane.YES_OPTION);
        }
        else if (JOptionPane.showOptionDialog(MainWindow.frame(), "Set parameters to use for all images " +
                "of all projects?", "Set parameters?", JOptionPane.YES_NO_OPTION,
                JOptionPane.QUESTION_MESSAGE, null, null, null) == JOptionPane.YES_OPTION)
        {
            parameters = parameterChoiceDialog();
        }

        while (true)
        {
            if (update)
            {
                IOProjectData project = new IOProjectData(true);
                if (project.valid())
                {
                    projects.add(project.getLocations());
                }
                else
                {
                    JOptionPane.showMessageDialog(MainWindow.frame(), "This project could not be added.",
                                                  "Project not added!", JOptionPane.WARNING_MESSAGE, null);
                }
            }
            else
            {
                projects.add((default_folders ? defaultLocations() : getLocations()));
            }

            String overview = "<html><body>Common path, Project file, Input image folder, " +
                              "XML folder, Output image folder, CSV Folder<br>";
            for (File[] info : projects)
            {
                for (String s : extractCommonAncestor(info))
                {
                    overview += s + ", ";
                }
                overview = overview.substring(0, overview.length()-2) + "<br>";
            }
            overview += "<br>Would you like to add another project to the batch list?";
            JScrollPane scroller = new JScrollPane(new JLabel(overview + "</body></html>"),
                                                   ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                                                   ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
            scroller.setBorder(BorderFactory.createEmptyBorder());

            if (JOptionPane.showOptionDialog(MainWindow.frame(), scroller, "Add another project?",
                    JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null) ==
                    JOptionPane.NO_OPTION)
            {
                break;
            }
        }
        String overview = "<html><body>Common path, Project file, Input image folder, XML folder, " +
                          "Output image folder, CSV Folder<br>";
        for (File[] info : projects)
        {
            for (String s : extractCommonAncestor(info))
            {
                overview += s + ", ";
            }
            overview = overview.substring(0, overview.length()-2) + "<br>";
        }
        overview += "<br>Do you want to start batch processing?";
        JScrollPane scroller = new JScrollPane(new JLabel(overview + "</body></html>"),
                                               ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                                               ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        scroller.setBorder(BorderFactory.createEmptyBorder());

        if (JOptionPane.showOptionDialog(MainWindow.frame(), scroller, "Batch Plan",
                JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE, null, null, null) ==
                JOptionPane.NO_OPTION)
        {
            return null;
        }
        if (JOptionPane.showOptionDialog(MainWindow.frame(), "Would you like to save this batch plan for later use?",
                "Save batch plan?", JOptionPane.YES_NO_OPTION,
                JOptionPane.QUESTION_MESSAGE, null, null, null) == JOptionPane.YES_OPTION)
        {
            saveMetaBatch(update, projects, parameters);
        }
        return new MetaBatchData(update, projects, parameters);
    }

    private static Integer getInt(Element element, String attribute)
    {
        if (element.hasAttribute(attribute))
        {
            return Integer.parseInt(element.getAttribute(attribute));
        }
        return null;
    }

    /**
     * This is a simple pairing class to enforce one-to-one correspondence of
     * image files with their respective XML data files.
     */
    public static class Files implements Comparable<Files>
    {
        private File image;
        private File data;

        public Files(File image, File data)
        {
            this.image = image;
            this.data = data;
        }

        public File image()
        {
            return image;
        }

        public File data()
        {
            return data;
        }

        public void setData(File data_file)
        {
            data = data_file;
        }

        @Override
        public int compareTo(Files that)
        {
            return image().compareTo(that.image());
        }
    }

    private static class ParameterField
    {
        private String m_label;
        private Any m_parameter;
        private JSpinner m_spinbox;

        private ParameterField(String label, Any parameter, DataParameters parameters)
        {
            m_label = label;
            m_parameter = parameter;
            if (parameter == null)
            {
                m_spinbox = null;
            }
            else
            {
                m_spinbox = new JSpinner(new SpinnerNumberModel(parameters.getParameter(parameter),
                                         -1, 10000, 1));
                m_spinbox.setMaximumSize(new Dimension(50, 28));
            }
        }
    }

    private static class MetaBatchData
    {
        private boolean update;
        private List<File[]> project_table;
        private DataParameters parameters;

        private MetaBatchData (boolean upd, List<File[]> table, DataParameters params)
        {
            update = upd;
            project_table = table;
            parameters = params;
        }
    }

}
