package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Point;
import java.util.ArrayList;
import java.util.List;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class Filament extends SingleFilament
{
    private Point m_head, m_tail;
    private boolean m_keep;

    private Point m_center; // center is scaled by a factor of 10.
    // length, angle and width are longs that are scaled by 1,000,000
    // to avoid the problems of floating point arithmetics.
    private long m_length;
    private long m_orientation; // 0 =< angle < 180,000,000
    private long m_width;

    public Filament(Point a, Point b, long width, boolean keep)
    {
        m_head = a;
        m_tail = b;
        m_center = calcCenter(a, b);
        m_length = Calc.distanceM(a, b);
        m_orientation = Calc.orientation(a, b);
        this.m_width = width;
        this.m_keep = keep;
    }

    @Override
    public ClusterableFilament makeClusterable(int time)
    {
        return new ClusterableFilament(this, time);
    }

    // Getters

    public List<Point> getPoints()
    {
        List<Point> points = new ArrayList<>();
        points.add(new Point(m_head));
        points.add(new Point(m_tail));
        return points;
    }

    @Override
    public List<Filament> splitIntoLinearPieces()
    {
        List<Filament> out = new ArrayList<>();
        out.add(this);
        return out;
    }

    public Point getHead()
    {
        return new Point(m_head);
    }

    public Point getTail()
    {
        return new Point(m_tail);
    }

    @Override
    public Point getCenter()
    {
        return new Point(m_center);
    }

    @Override
    public boolean getKeep()
    {
        return m_keep;
    }

    @Override
    public long getLength()
    {
        return m_length;
    }

    public double getMass()
    {
        return m_length * m_width / (Const.MF * Const.MF);
    }

    @Override
    public long getOrientation()
    {
        return m_orientation;
    }

    @Override
    public long getWidth()
    {
        return m_width;
    }

    @Override
    public void setKeep(boolean keep)
    {
        this.m_keep = keep;
    }

    @Override
    public void setWidth(long width)
    {
        this.m_width = width;
    }

    @Override
    public void invert()
    {
        Point new_tail = new Point(m_head);
        m_head = new Point(m_tail);
        m_tail = new Point(new_tail);
    }

    @Override
    public void remedyOldAngles()
    {
        int tmp = m_head.y;
        m_head.y = m_tail.y;
        m_tail.y = tmp;
    }

    @Override
    public Element toXML(Document document, int chain_number, int filament_number)
    {
        Element element = document.createElement("filament" + chain_number + "_" + filament_number);

        element.setAttribute("start_x", "" + m_head.x);
        element.setAttribute("start_y", "" + m_head.y);
        element.setAttribute("end_x", "" + m_tail.x);
        element.setAttribute("end_y", "" + m_tail.y);
        element.setAttribute("center_x", "" + m_center.x);
        element.setAttribute("center_y", "" + m_center.y);
        element.setAttribute("length", "" + m_length);
        element.setAttribute("angle", "" + m_orientation);
        element.setAttribute("width", "" + m_width);
        element.setAttribute("keep", m_keep ? "yes" : "no");

        return element;
    }

    // Note that the coordinates of the center are scaled by a factor of 10
    // in comparison to the coordinates of the corner points.
    private static Point calcCenter(Point a, Point b)
    {
        return new Point((a.x + b.x) * 5, (a.y + b.y) * 5);
    }

    @Override
    public Object clone()
    {
        Filament dup = new Filament(new Point(m_head.x, m_head.y), new Point(m_tail.x, m_tail.y), m_width, m_keep);
        return dup;
    }

    @Override
    public boolean equals(Object object)
    {
        if (object == null || !(object.getClass().equals(this.getClass())))
        {
            return false;
        }

        Filament fil = (Filament) object;
        return (m_keep == fil.m_keep && m_head.x == fil.m_head.x && m_head.y == fil.m_head.y && m_tail.x == fil.m_tail.x
                && m_tail.y == fil.m_tail.y && m_width == fil.m_width && m_center.x == fil.m_center.x && m_center.y == fil.m_center.y
                && m_length == fil.m_length && m_orientation % (180 * Const.M) == fil.m_orientation % (180 * Const.M));
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + ((m_head == null) ? 0 : m_head.hashCode());
        result = prime * result + (int) (m_orientation ^ (m_orientation >>> 32));
        result = prime * result + ((m_tail == null) ? 0 : m_tail.hashCode());
        result = prime * result + ((m_center == null) ? 0 : m_center.hashCode());
        result = prime * result + (m_keep ? 1231 : 1237);
        result = prime * result + (int) (m_length ^ (m_length >>> 32));
        result = prime * result + (int) (m_width ^ (m_width >>> 32));
        return result;
    }

    public double[] distances(Filament that)
    {
        double[] out = new double[2];
        double tmp = Math.abs(m_orientation - that.m_orientation) / Const.MF;
        out[0] = (tmp > 90. ? 180. - tmp : tmp);
        Point p1 = m_head, q1 = m_tail, p2 = that.m_head, q2 = that.m_tail;
        int z1_x = p1.x - q1.x, z1_y = p1.y - q1.y,
            z2_x = p2.x - q2.x, z2_y = p2.y - q2.y,
            h_x  = p1.x - p2.x, h_y  = p1.y - p2.y;
        double t2 = (z1_x * h_y - z1_y * h_x) / (double)(z1_x * z2_y - z1_y * z2_x);
        if (t2 >= 0 && t2 <= 1)
        {
            double t1 = (h_x + z2_x * t2) / (double)z1_x;
            if (t1 >= 0 && t1 <= 1)
            {
                out[1] = 0;
                return out;
            }
        }

        double dist_square = pointLineDistanceSquare(p1, q1, p2);
        tmp = pointLineDistanceSquare(p1, q1, q2);
        dist_square = (dist_square > tmp ? tmp : dist_square);
        tmp = pointLineDistanceSquare(p2, q2, p1);
        dist_square = (dist_square > tmp ? tmp : dist_square);
        tmp = pointLineDistanceSquare(p2, q2, q1);
        out[1] = (dist_square > tmp ? tmp : dist_square);
        return out;
    }

//    private double pointLineDistanceSquare(Point p1, Point q1, Point p)
//    {
//        int z_x = p1.x - q1.x, z_y = p1.y - q1.y, h_x  = p1.x - p.x, h_y  = p1.y - p.y;
//        double t = (z_x * h_x + z_y * h_y) / (double)(z_x * z_x + z_y * z_y);
//        if (t > 1)
//        {
//            return p1.distanceSq(p);
//        }
//        if (t < 0)
//        {
//            return q1.distanceSq(p);
//        }
//        double d_x = q1.x + t * z_x - p.x, d_y = q1.y + t * z_y - p.y;
//        return d_x * d_x + d_y * d_y;
//    }
}
