package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.util.ArrayList;
import java.util.List;
import java.awt.Point;
import java.lang.Math;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;

import ui.FilamentSensor;

public class FilamentChain extends AbstractFilament
{

    private List<SingleFilament> m_filament_list;

    private boolean m_keep;
    private Point m_center;
    // m_angle, m_length and m_width are longs that are scaled by 1,000,000
    // to avoid the problems of floating point arithmetics.
    private long m_orientation;
    private long m_length;
    private long m_width;

    public FilamentChain(List<SingleFilament> filament_list)
    {
        m_filament_list = (filament_list == null ? new ArrayList<SingleFilament>() : filament_list);

        m_keep = false;
        for (SingleFilament filament : m_filament_list)
        {
            if (filament.getKeep())
            {
                m_keep = true;
                break;
            }
        }

        chainAverages();
    }

    public FilamentChain(SingleFilament filament)
    {
        if (filament == null)
        {
            m_filament_list = new ArrayList<>();
            m_keep = false;
            chainAverages();
            return;
        }
        m_filament_list = new ArrayList<>();
        // This is ok, because m_filament_list has only one filament afterwards.
        if (!add(filament))
        {
            chainAverages();
            m_keep = false;
        }
        else
        {
            m_keep = filament.getKeep();
        }
    }

    public FilamentChain()
    {
        m_filament_list = new ArrayList<>();
        chainAverages();
        m_keep = true;
    }

    public FilamentChain(Node chain)
    {
        m_filament_list = new ArrayList<>();
        m_keep = false;

        for (Node filament = chain.getFirstChild(); filament != null; filament = filament.getNextSibling())
        {
            try
            {
                if (filament.getNodeType() != Node.ELEMENT_NODE || !filament.hasAttributes())
                {
                    continue;
                }

                CurvedFilament fil = new CurvedFilament((Element) filament);
                m_keep = m_keep || fil.getKeep();
                m_filament_list.add(fil);
            }
            catch (Exception e)
            {
                FilamentSensor.debugError("Error parsing Filament '" + filament + "': " + e.getMessage());
            }
        }

        chainAverages();
    }

    // The following methods allow direct manipulation of the filament chain.
    // They exist mostly for testing. If they are ever called by running code
    // a sanity check for the filament chain should be performed.

    boolean add(SingleFilament filament)
    {
        if (filament != null && m_filament_list.add(filament))
        {
            chainAverages();
            return true;
        }
        return false;
    }

    boolean removeFilament(SingleFilament filament)
    {
        if (m_filament_list.remove(filament))
        {
            chainAverages();
            return true;
        }
        return false;
    }

    // Getter methods.

    @Override
    public Point getCenter()
    {
        return m_center;
    }

    @Override
    public long getOrientation()
    {
        return m_orientation;
    }

    @Override
    public boolean getKeep()
    {
        return m_keep;
    }

    public boolean getKeepAny()
    {
        for (SingleFilament filament : m_filament_list)
        {
            if (filament.getKeep())
            {
                return true;
            }
        }
        return m_keep;
    }

    @Override
    public long getLength()
    {
        return m_length;
    }

    @Override
    public long getWidth()
    {
        return m_width;
    }

    List<SingleFilament> getFilaments()
    {
        FilamentChain chain = (FilamentChain) clone();
        return chain.m_filament_list;
    }

    public SingleFilament getFilament(int position)
    {
        if (m_filament_list == null || position >= m_filament_list.size())
        {
            return null;
        }
        return (SingleFilament) (m_filament_list.get(position).clone());
    }

    SingleFilament changeFilament(int position)
    {
        if (m_filament_list == null || position >= m_filament_list.size())
        {
            return null;
        }
        return m_filament_list.get(position);
    }

    public int getListSize()
    {
        if (m_filament_list == null)
        {
            return -1;
        }

        return m_filament_list.size();
    }

    boolean contains(SingleFilament filament)
    {
        if (m_filament_list == null)
        {
            return false;
        }

        return m_filament_list.contains(filament);
    }

    @Override
    public void setKeep(boolean keep)
    {
        m_keep = keep;

        if (getListSize() == 1)
        {
            m_filament_list.get(0).setKeep(keep);
        }
    }

    public void setKeepAll(boolean keep)
    {
        m_keep = keep;

        for (SingleFilament filament : m_filament_list)
        {
            filament.setKeep(keep);
        }
    }

    @Override
    public void setWidth(long width)
    {
        if (getListSize() == 1)
        {
            m_filament_list.get(0).setWidth(width);
            width();
        }
        else
        {
            System.err.println("FilamentSensor FilamentChain.setWidth: " +
            		           "Width can only be changed for single filaments!");
        }
    }

    @Override
    public void invert()
    {
        if (m_filament_list == null || m_filament_list.isEmpty())
        {
            return;
        }

        List<SingleFilament> new_list = new ArrayList<>();

        for (int i = m_filament_list.size() - 1; i >= 0; i--)
        {
            SingleFilament filament = m_filament_list.get(i);
            filament.invert();
            new_list.add(filament);
        }

        m_filament_list = new_list;
    }

    @Override
    public void remedyOldAngles()
    {
        if (m_filament_list == null || m_filament_list.isEmpty())
        {
            return;
        }

        for (SingleFilament filament : m_filament_list)
        {
            filament.remedyOldAngles();
        }
    }

    @Override
    public Element toXML(Document document, int chain_number, int dump)
    {
        Element chain_element = document.createElement("chain" + chain_number);

        chain_element.setAttribute("center_x", "" + m_center.x);
        chain_element.setAttribute("center_y", "" + m_center.y);
        chain_element.setAttribute("length", "" + m_length);
        chain_element.setAttribute("angle", "" + m_orientation);
        chain_element.setAttribute("width", "" + m_width);

        for (int j = 0; j < getListSize(); j++)
        {
            chain_element.appendChild(getFilament(j).toXML(document, chain_number, j));
        }
        return chain_element;
    }

    // Methods for calculation of chain properties.

    private void chainAverages()
    {
        if (m_filament_list == null || m_filament_list.isEmpty())
        {
            m_center = new Point(0, 0);
            m_orientation = 0;
            m_length = 0;
            m_width = 0;
            return;
        }

        // Speed up treatment of trivial filament chains.
        // Necessary due to frequent conversions.
        if (m_filament_list.size() == 1)
        {
            m_center = m_filament_list.get(0).getCenter();
            m_orientation = m_filament_list.get(0).getOrientation();
            m_length = m_filament_list.get(0).getLength();
            m_width = m_filament_list.get(0).getWidth();
            return;
        }

        m_center = center();
        m_orientation = angle();
        m_length = length();
        m_width = width();
    }

    // The center is calculated as a length*width-weighted average.
    private Point center()
    {
        double x_numerator = 0;
        double y_numerator = 0;
        double denominator = m_filament_list.size();

        for (int i = 0; i < m_filament_list.size(); i++)
        {
            denominator += m_filament_list.get(i).getLength() * m_filament_list.get(i).getWidth();
            x_numerator += m_filament_list.get(i).getLength() * m_filament_list.get(i).getWidth()
                           * m_filament_list.get(i).getCenter().x;
            y_numerator += m_filament_list.get(i).getLength() * m_filament_list.get(i).getWidth()
                           * m_filament_list.get(i).getCenter().y;
        }
        return new Point((int) Math.floor(x_numerator / denominator), (int) Math.floor(y_numerator / denominator));
        // Use floor instead of round, because center is scaled 10-fold:
        // Math.round( ((int)Math.round(1234,5))/10f ) = Math.round( 1235/10f )
        // = 124

    }

    // The angle is calculated as a length*width-weighted average.
    // The implementation is probably very inefficient.
    private long angle()
    {
        long unchanged_angle = 0;
        long changed_angle = 0;
        long[] angles = new long[m_filament_list.size()];
        for (int i = 0; i < m_filament_list.size(); i++)
        {
            angles[i] = m_filament_list.get(i).getOrientation();
            for (int j = 0; j < i; j++)
            {
                // Sum of angle differences
                unchanged_angle += Math.abs(angles[i] - angles[j]);
                // Sum of differences between angles rotated by 90°
                changed_angle += Math.abs(((angles[i] + 90 * Const.M) % (180 * Const.M)) -
                                          ((angles[j] + 90 * Const.M) % (180 * Const.M)));
            }
        }

        // The condition will be true, if orientations accumulate around 0° = 180°.
        if (changed_angle < unchanged_angle)
        {
            double numerator = 0;
            double denominator = m_filament_list.size();
            for (int i = 0; i < m_filament_list.size(); i++)
            {
                denominator += (double) m_filament_list.get(i).getLength() * (double) m_filament_list.get(i).getWidth();
                numerator += (double) m_filament_list.get(i).getLength() * (double) m_filament_list.get(i).getWidth()
                             * (((m_filament_list.get(i).getOrientation() + 90 * Const.M) % (180 * Const.M)) - 90 * Const.M);
            }
            // +180*M because in java, modulo can be negative.
            return Math.round((numerator / denominator + 180 * Const.M) % (180 * Const.M));
        }
        else
        {
            double numerator = 0;
            double denominator = m_filament_list.size();
            for (int i = 0; i < m_filament_list.size(); i++)
            {
                denominator += (double) m_filament_list.get(i).getLength() * (double) m_filament_list.get(i).getWidth();
                numerator += (double) m_filament_list.get(i).getLength() * (double) m_filament_list.get(i).getWidth()
                             * m_filament_list.get(i).getOrientation();
            }
            return Math.round(numerator / denominator);
        }
    }

    /**
     * The length calculation for a filament chain requires the filaments
     * to be ordered such that neighboring elements of the list identify
     * neighboring filaments.
     */
    private long length()
    {
        sortFilamentList();
        long length = 0;
        for (int i = 0; i < m_filament_list.size() - 1; i++)
        {
            length += Calc.distanceM(m_filament_list.get(i).getCenter(), m_filament_list.get(i + 1).getCenter()) / 10;
        }
        length += Math.round(0.5 * m_filament_list.get(0).getLength() + 0.5
                             * m_filament_list.get(m_filament_list.size() - 1).getLength());

        return length;
    }

    // The width of the chain is calculated as the length-weighted average of
    // filament widths.
    private long width()
    {
        double numerator = 0;
        double denominator = m_filament_list.size();
        for (int i = 0; i < m_filament_list.size(); i++)
        {
            denominator += m_filament_list.get(i).getLength();
            numerator += (double) m_filament_list.get(i).getLength() * (double) m_filament_list.get(i).getWidth();
        }
        return Math.round(numerator / denominator);
    }

    /**
     * This method sorts the list such that neighboring elements of the list
     * identify neighboring filaments. Returns true, if the implied sanity
     * check succeeds.
     */
    private boolean sortFilamentList()
    {
        if (m_filament_list.size() == 0)
        {
            return false;
        }

        return true; // TODO: Should probably call a method in
                     // CalculateFilamentChains.
    }

    @Override
    public Object clone()
    {
        List<SingleFilament> list = new ArrayList<>();
        for (int i = 0; i < m_filament_list.size(); i++)
        {
            list.add((SingleFilament) m_filament_list.get(i).clone());
        }
        FilamentChain dup = new FilamentChain(list);
        return dup;
    }

    @Override
    public boolean equals(Object object)
    {
        if (object == null || !(object.getClass().equals(this.getClass())))
        {
            return false;
        }

        FilamentChain fil = (FilamentChain) object;
        return (m_center.x == fil.m_center.x && m_center.y == fil.m_center.y && m_length == fil.m_length
                && m_width == fil.m_width && m_orientation % (180 * Const.M) == fil.m_orientation % (180 * Const.M)
                && m_filament_list.containsAll(fil.m_filament_list) && fil.m_filament_list.containsAll(m_filament_list));
    }

    @Override
    public int hashCode()
    {
        final int prime = 31;
        int result = 1;
        result = prime * result + (int) (m_orientation ^ (m_orientation >>> 32));
        result = prime * result + ((m_center == null) ? 0 : m_center.hashCode());
        result = prime * result + ((m_filament_list == null) ? 0 : m_filament_list.hashCode());
        result = prime * result + (int) (m_length ^ (m_length >>> 32));
        result = prime * result + (int) (m_width ^ (m_width >>> 32));
        return result;
    }

}
