package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import core.DataParameters.Any;
import core.DataParameters.Pre;
import filters.FilterContrast;
import filters.FilterCrossCorrelation;
import filters.FilterGauss;
import filters.FilterLaplace;
import filters.FilterLineGauss;
import filters.FilterPreprocessing;
import ui.FilamentSensor;

public class FilterBench
{

    private List<FilterPreprocessing> m_filter_bench;

    public FilterBench(String bench, boolean[][] cell_area)
    {
        m_filter_bench = new ArrayList<>();
        String[] filters = bench.split("/");
        for (String filter : filters)
        {
            String[] parameters = filter.split(";");
            switch (parameters[0])
            {
            case "C":
                if (parameters.length != 3)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                    break;
                }
                try
                {
                    m_filter_bench.add(new FilterContrast(Integer.parseInt(parameters[1]),
                                                          Integer.parseInt(parameters[2])));
                }
                catch (NumberFormatException e)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                }
                break;
            case "D":
                if (parameters.length != 2)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                    break;
                }
                try
                {
                    m_filter_bench.add(new FilterLineGauss(Integer.parseInt(parameters[1]) / 10.0,
                                                           cell_area));
                }
                catch (NumberFormatException e)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                }
                break;
            case "G":
                if (parameters.length != 2)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                    break;
                }
                try
                {
                    m_filter_bench.add(new FilterGauss(Integer.parseInt(parameters[1]) / 10.0));
                }
                catch (NumberFormatException e)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                }
                break;
            case "L":
                if (parameters.length != 3)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                    break;
                }
                try
                {
                    m_filter_bench.add(new FilterLaplace(Integer.parseInt(parameters[1]),
                                                         Integer.parseInt(parameters[2]) / 10.0));
                }
                catch (NumberFormatException e)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                }
                break;
            case "X":
                if (parameters.length != 3)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                    break;
                }
                try
                {
                    double[][][] masks = makeMasks(1, Integer.parseInt(parameters[1]));
                    m_filter_bench.add(new FilterCrossCorrelation(masks, Integer.parseInt(parameters[2]) / 100.0));
                }
                catch (NumberFormatException e)
                {
                    FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                }
                break;
            default:
                FilamentSensor.debugError("Invalid filter definition " + filter + ". Moving on...");
                break;
            }
        }
    }

    public FilterBench(Map<Any, Integer> pre_params, boolean[][] cell_area,
                       int orig_range, int black, int white)
    {
        m_filter_bench = new ArrayList<>();
        if (orig_range < pre_params.get(Pre.min_range))
        {
            return;
        }
        if (pre_params.get(Pre.contrast) != 0)
        {
            black = pre_params.get(Pre.black);
            white = pre_params.get(Pre.white);
        }

        m_filter_bench.add(new FilterContrast(black, white));

        if (pre_params.get(Pre.rm_sng_px) != 0)
        {
            FilamentSensor.debugError("Removing single pixels is deprecated.");
        }

        FilterGauss gauss = new FilterGauss(pre_params.get(Pre.fsigma) / 10.0);
        FilterLaplace laplace = new FilterLaplace(pre_params.get(Pre.lpmask),
                                                  pre_params.get(Pre.lpfac) / 10.0);
        if (pre_params.get(Pre.order) == 0)
        {
            if (pre_params.get(Pre.do_gauss) != 0)
            {
                m_filter_bench.add(gauss);
            }
            if (pre_params.get(Pre.do_laplace) != 0)
            {
                m_filter_bench.add(laplace);
            }
        }
        else
        {
            if (pre_params.get(Pre.do_laplace) != 0)
            {
                m_filter_bench.add(laplace);
            }
            if (pre_params.get(Pre.do_gauss) != 0)
            {
                m_filter_bench.add(gauss);
            }
        }

        if (pre_params.get(Pre.cross_corr) != 0)
        {
            double[][][] masks = makeMasks(1, pre_params.get(Pre.corr_mask_size));
            m_filter_bench.add(new FilterCrossCorrelation(masks, pre_params.get(Pre.corr_zero) / 100.0));
        }

        if (pre_params.get(Pre.do_line_gauss) != 0)
        {
            m_filter_bench.add(new FilterLineGauss(pre_params.get(Pre.line_sigma) / 10.0, cell_area));
        }
    }

    int[][] apply(int[][] image)
    {
        FilamentSensor.debugMessage(toXML());
        for (FilterPreprocessing filter : m_filter_bench)
        {
            image = filter.apply(image);
        }
        return image;
    }

    void clear()
    {
        m_filter_bench.clear();
    }

    public String toXML()
    {
        String out = "";
        if (m_filter_bench.size() == 0)
        {
            return out;
        }
        for (FilterPreprocessing filter : m_filter_bench)
        {
            out += filter.toXML() + "#";
        }
        return out = out.substring(0, out.length()-1);
    }

    private static double[][][] makeMasks(int widths, int size)
    {
        if (size <= 0 || widths <= 0)
        {
            return null;
        }

        int range = size + widths;
        double[][][] temp = new double[4 * size * widths][2 * range + 1][2 * range + 1];
        double sigma = 0.5;
        double[][] gauss = Calc.gaussianMask(widths, sigma);

        for (int w = 0; w < widths; w++)
        {
            for (int dir = 0; dir < size; dir++)
            {
                double slope = (double)dir / (double)size;

                for (int x = - range; x <= range; x++)
                {
                    temp[dir + 4 * w * size][range + x][range + (int)Math.round(slope * x)] = 1;
                }
                temp[dir + 4 * w * size] = convolve(convolve(temp[dir + 4 * w * size], Const.CIRCLE[w+1]), gauss);
            }

            for (int dir=size; dir<3*size; dir++)
            {
                double epols = ( 2 * (double)size - dir ) / size;
                for (int y = - range; y <= range; y++)
                {
                    temp[dir + 4 * w * size][range + (int)Math.round(epols * y)][range + y] = 1;
                }
                temp[dir + 4 * w * size] = convolve(convolve(temp[dir + 4 * w * size], Const.CIRCLE[w+1]), gauss);
            }

            for (int dir=3*size; dir<4*size; dir++)
            {
                double slope = (dir - 4 * (double)size) / size;
                for (int x = - range; x <= range; x++)
                {
                    temp[dir + 4 * w * size][range + x][range + (int)Math.round(slope * x)] = 1;
                }
                temp[dir + 4 * w * size] = convolve(convolve(temp[dir + 4 * w * size], Const.CIRCLE[w+1]), gauss);
            }
        }
        double[][][] out = new double[4 * size * widths][2 * size + 1][2 * size + 1];
        for (int n = 0; n < 4 * size * widths; n++)
        {
            for (int x = 0; x <= 2 * size; x++)
            {
                for (int y = 0; y <= 2 * size; y++)
                {
                    out[n][x][y] = temp[n][x+widths][y+widths];
                }
            }
        }
        return out;
    }

    private static double[][] convolve(double[][] image, double[][] mask)
    {
        if (image.length <= 0 || image[0].length <= 0 || mask.length <= 0 || mask[0].length <= 0 ||
            image.length < mask.length || image[0].length < mask[0].length ||
            mask.length % 2 == 0 || mask[0].length % 2 == 0)
        {
            return null;
        }

        int width = image.length;
        int height = image[0].length;
        int range_horizontal = ( mask.length - 1 ) / 2;
        int range_vertical = ( mask[0].length - 1 ) / 2;

        double out[][] = new double[width][height];

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < range_vertical; j++)
            {
                out[i][j] = image[i][j];
                out[i][height-j-1] = image[i][height-j-1];
            }
        }
        for (int j = 0; j < height; j++)
        {
            for (int i = 0; i < range_horizontal; i++)
            {
                out[i][j] = image[i][j];
                out[height-i-1][j] = image[height-i-1][j];
            }
        }
        for (int i = range_horizontal; i < width - range_horizontal; i++)
        {
            for (int j = range_vertical; j < height - range_vertical; j++)
            {
                for (int a = -range_horizontal; a <= range_horizontal; a++)
                {
                    for (int b = -range_vertical; b <= range_vertical; b++)
                    {
                        out[i][j] += mask[a + range_horizontal][b + range_vertical] * image[i + a][j + b];
                    }
                }
            }
        }
        return out;
    }
}
