package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.Map;

import core.DataParameters.Any;
import core.DataParameters.Bin;
import core.DataParameters.Elog;
import core.DataParameters.Hough;
import filters.FilterBinarization;
import filters.FilterELoG;
import filters.FilterHistogramMethods;
import filters.PluginCellShape;
import filters.PluginScoreImage;
import ui.FilamentSensor;
import ui.ImageFactory;
import ui.ImageFileChooser;

public final class ImageProcessing
{
    private int[][] m_int_image_initial,
                    m_int_image_preprocessed,
                    m_int_image_filter_difference;
    private boolean[][] m_binary_image,
                        m_binary_image_hough;
    
    private int[] m_elog_histogram,
                  m_hough_histogram;

    private int m_width, m_height, m_auto_black, m_auto_white, m_orig_range,
                m_auto_elog_threshold;

    private BufferedImage m_buffered_image_red_green;

    private int m_preprocessing_parameters_hash,
                m_binarization_parameters_hash,
                m_elog_parameters_hash,
                m_hough_parameters_hash;

    private boolean m_update_elogs,
                    m_update_hough,
                    m_update_binarization;

    private FilterELoG m_filter_elog;

    ImageProcessing(File img_file) throws IOException
    {
        m_int_image_initial = ImageFileChooser.readGreyImage(img_file);
        m_width = m_int_image_initial.length;
        m_height = m_int_image_initial[0].length;
        getRange();

        FilamentSensor.debugMessage("Image in original size was successfully read in by ImageProcessing!");

        m_update_elogs = false;
        m_update_hough = false;
        m_update_binarization = false;

        m_preprocessing_parameters_hash = 0;
        m_binarization_parameters_hash = 0;

        m_int_image_preprocessed = null;
        m_int_image_filter_difference = null;
        m_binary_image_hough = null;
        m_binary_image = null;
        m_elog_histogram = null;
        m_hough_histogram = null;

        if (m_width <= 0 || m_height <= 0)
        {
            return;
        }

        m_binary_image = new boolean[m_width][m_height];
        m_int_image_preprocessed = new int[m_width][m_height];
        FilterHistogramMethods contrast = new FilterHistogramMethods(m_int_image_initial,
                                                                     m_auto_black, m_auto_white);
        m_auto_black = contrast.black();
        m_auto_white = contrast.white();
        m_auto_elog_threshold = (new FilterHistogramMethods(m_int_image_initial, true,
                                                            FilterHistogramMethods.LI)).threshold();
    }

    String getResolution()
    {
        return m_width + "x" + m_height;
    }

    int[][] getOriginalImage()
    {
        return m_int_image_initial;
    }

    PluginCellShape getCellData(boolean[][] mask, int min_range)
    {
        return PluginCellShape.getCellData(m_int_image_initial, mask, m_orig_range, min_range);
    }

    boolean[][] getBinaryImage()
    {
        return m_binary_image;
    }

    BufferedImage getRedGreenImage()
    {
        return m_buffered_image_red_green;
    }
    
    int[] getBlackWhite()
    {
        return new int[]{ m_auto_black, m_auto_white };
    }

    int getElogThreshold()
    {
        return m_auto_elog_threshold;
    }

    String getHistogram(int type)
    {
        int[] histogram = null;
        switch (type)
        {
        case DataImage.ELOG_COLOR:
            histogram = m_elog_histogram;
            break;
        case DataImage.HOUGH:
            histogram = m_hough_histogram;
            break;
        default:
            return null;
        }
        String out = "";
        for (int i = 0; i < histogram.length; i++)
        {
            out += (i * 180 / (double)histogram.length) + "," + histogram[i] + "\n";
        }
        return out;
    }

    boolean updatePreprocessing(Map<Any, Integer> pre_params, boolean[][] cell_area, boolean changed)
    {
        FilamentSensor.debugMessage("Update preprocessing.");

        if (pre_params.hashCode() != m_preprocessing_parameters_hash)
        {
            m_int_image_filter_difference = new int[m_width][m_height];
            m_int_image_preprocessed = new int[m_width][m_height];

            FilterBench bench = new FilterBench(pre_params, cell_area, m_orig_range, m_auto_black,
                                                m_auto_white);
            m_int_image_preprocessed = bench.apply(m_int_image_initial);

            for (int x = 0; x < m_width; x++)
            {
                for (int y = 0; y < m_height; y++)
                {
                    m_int_image_filter_difference[x][y] = m_int_image_preprocessed[x][y] -
                                                          m_int_image_initial[x][y];
                }
            }

            // invalidate binarization
            m_binarization_parameters_hash = 0;
            changed = true;
        }

        m_preprocessing_parameters_hash = pre_params.hashCode();
        m_update_elogs = true;
        m_update_binarization = true;
        return changed;
    }

    boolean updateBinarization(Map<Any, Integer> bin_params, boolean elog_null, boolean[][] mask)
    {
        FilamentSensor.debugMessage("Update binarization.");

        if (bin_params.hashCode() == m_binarization_parameters_hash && !m_update_binarization)
        {
            return false;
        }

        switch (bin_params.get(Bin.method))
        {
        case 0:
            m_binary_image = FilterBinarization.area(m_int_image_preprocessed,
                                                     bin_params.get(Bin.minmean),
                                                     bin_params.get(Bin.sigma) / 10.0,
                                                     bin_params.get(Bin.area_significance) / 10.0);
            break;
        case 1:
            m_binary_image = FilterBinarization.rod(m_int_image_preprocessed,
                                                    bin_params.get(Bin.minmean),
                                                    bin_params.get(Bin.sigma) / 10.0,
                                                    bin_params.get(Bin.rod_significance) / 10.0);
            break;
        default:
            return false;
        }

        if (bin_params.get(Bin.restrict) != 0 && mask != null)
        {
            final int width = m_binary_image.length,
                      height = m_binary_image[0].length,
                      mask_width = mask.length,
                      mask_height = mask[0].length;
            if (mask_width == width && mask_height == height)
            {
                for (int x = 0; x < width; x++)
                {
                    for (int y = 0; y < height; y++)
                    {
                        m_binary_image[x][y] = m_binary_image[x][y] || !mask[x][y];
                    }
                }
            }
        }
        if (bin_params.get(Bin.thicken) != 0)
        {
            thickenLines(m_binary_image);
        }
        else
        {
            cleanPixelsBin(m_binary_image);
        }

        m_binarization_parameters_hash = bin_params.hashCode();
        m_update_binarization = false;
        m_update_hough = true;
        return true;
    }

    boolean calculateELoGs(Map<Any, Integer> elog_params)
    {
        if ( (elog_params.hashCode() == m_elog_parameters_hash && !m_update_elogs) || m_int_image_initial == null)
        {
            return false;
        }

        int threshold = elog_params.get(Elog.threshold);
        threshold = (threshold < 0 ? m_auto_elog_threshold : threshold);
        m_filter_elog.calculateELoGs(m_int_image_initial, elog_params.get(Elog.tolerance),
                                     elog_params.get(Elog.contrast), threshold,
                                     elog_params.get(Elog.orientations),
                                     elog_params.get(Elog.sigma_x)/10.0,
                                     elog_params.get(Elog.sigma_y)/10.0);

        return true;
    }

    boolean calculateHough(Map<Any, Integer> hough_params)
    {
        if (hough_params.hashCode() == m_hough_parameters_hash && !m_update_hough)
        {
            return false;
        }

        int[][] hough_image = null;

        switch (hough_params.get(Hough.image))
        {
        case 2:
            hough_image = houghTransform(m_binary_image, hough_params.get(Hough.orientations));
            break;
        default:
            return false;
        }

        m_binary_image_hough = findMaxima(hough_image);
        makeHoughHistogram(hough_image, m_binary_image_hough);
        
        m_hough_parameters_hash = hough_params.hashCode();
        m_update_hough = false;
        return true;
    }

    int scoreImage(boolean[][] mask, int sigma, int min_range)
    {
        return PluginScoreImage.scoreImage(m_int_image_initial, mask, sigma, min_range);
    }

    long[] areaAndMeanBrightness (boolean[][] mask, boolean[][] cell)
    {
        long[] out = new long[3];
        long count = 0;
        for (int x = 0; x < m_width; x++)
        {
            for (int y = 0; y < m_height; y++)
            {
                if (cell[x][y])
                {
                    if (mask[x][y])
                    {
                        out[0]++;
                        out[1] += m_int_image_initial[x][y];
                    }
                    else
                    {
                        count++;
                        out[2] += m_int_image_initial[x][y];
                    }
                }
            }
        }
        out[1] = (out[1] * Const.M) / Math.max(1, out[0]);
        out[2] = (out[2] * Const.M) / Math.max(1, count);
        if (count > 0)
        {
            m_buffered_image_red_green = ImageFactory.makeTwoLevelRedGreenImage(m_int_image_initial, mask, cell);
        }
        return out;
    }

    private static int[][] houghTransform(boolean[][] bool_image, int orientations)
    {
        int width = bool_image.length;
        int half_width = width / 2;
        int height = bool_image[0].length;
        int half_height = height / 2;
        int half_diagonal = (int)Math.ceil(Math.sqrt(half_width * half_width +
                                                     half_height * half_height));
        double step = Math.PI / (double) orientations;
        int[][] hough_image = new int[2 * half_diagonal + 1][orientations + 20];

        for (int x = 1 - half_width; x < half_width; x++)
        {
            for (int y = 1 - half_height; y < half_height; y++)
            {
                if (!bool_image[x + half_width - 1][y + half_height - 1])
                {
                    for (int i = - 10; i < orientations + 10; i++)
                    {
                        // Orientation inversion using +sin.
                        int dist = (int) Math.round(Math.sin(i * step) * x + Math.cos(i * step) * y);
                        hough_image[half_diagonal + dist][i+10]++;
                    }
                }
            }
        }
        return hough_image;
    }

    private static boolean[][] findMaxima(int[][] image)
    {
        int width = image.length;
        int height = image[0].length;
        int[][] tmp_image = new int[width][height];
        boolean[][] maxima = new boolean[width][height];
        int radius = 1;
        for (int x = radius; x < width - radius; x++)
        {
            for (int y = radius; y < height - radius; y++)
            {
                for (int i = -radius; i <= radius; i++)
                {
                    for (int j = -radius; j <= radius; j++)
                    {
                        tmp_image[x][y] += image[x+i][y+j];
                    }
                }
            }
        }
        int range = (2*radius+1);
        for (int x = range; x < width - range; x++)
        {
            for (int y = range; y < height - range; y++)
            {
                maxima[x][y] = true;
                for (int i = -range; i <= range; i+= range)
                {
                    for (int j = -range; j <= range; j+= range)
                    {
                        if (i == 0 && j == 0)
                        {
                            continue;
                        }
                        maxima[x][y] = maxima[x][y] && (tmp_image[x][y] >= tmp_image[x+i][y+j]) && tmp_image[x][y] != 0;
                    }
                }
            }
        }
        
        int[][] max_image = new int[width][height];
        int counter = 1;

        for (int t = 0; counter > 0; t ++)
        {
            counter = 0;
            for (int x = range; x < width - range; x++)
            {
                for (int y = range; y < height - range; y++)
                {
                    maxima[x][y] = true;
                    for (int i = -range; i <= range; i+= range)
                    {
                        for (int j = -range; j <= range; j+= range)
                        {
                            if (i == 0 && j == 0)
                            {
                                continue;
                            }
                            maxima[x][y] = maxima[x][y] && tmp_image[x][y] != 0 &&
                                           (tmp_image[x][y] >= tmp_image[x+i][y+j] + t * range * range);
                        }
                    }
                    if (maxima[x][y])
                    {
                        max_image[x][y]++;
                        counter++;
                    }
                }
            }
        }
        maxima = (new FilterHistogramMethods(max_image, true,
                                             FilterHistogramMethods.OTSU)).binaryImage();

        return maxima;
    }

    private void makeHoughHistogram(int[][] image, boolean[][] mask)
    {
        int width = image.length;
        int height = image[0].length;
        m_hough_histogram = new int[height - 20];

        for (int x = 0; x < width; x++)
        {
            for (int y = 10; y < height - 10; y++)
            {
                if (mask[x][y])
                {
                    m_hough_histogram[y - 10] += image[x][y];
                }
            }
        }
    }

    private void getRange()
    {
        int max = 0;
        int min = Integer.MAX_VALUE;
        for (int x = 0; x < m_width; x++)
        {
            for (int y = 0; y < m_height; y++)
            {
                int pixel = m_int_image_initial[x][y];

                if (pixel > max)
                {
                    max = pixel;
                }
                if (pixel < min)
                {
                    min = pixel;
                }
            }
        }

        m_auto_black = min;
        m_auto_white = max;
        m_orig_range = max - min;
    }

    /**
     * Remove single, standalone pixels that are probably noise.
     */
    private static void cleanPixelsBin(boolean[][] image)
    {
        if (image == null)
        {
            return;
        }

        int width = image.length;
        int height = image[0].length;

        for (int i = 1; i < width - 1; i++)
        {
            for (int j = 1; j < height - 1; j++)
            {
                if (image[i][j] && !image[i - 1][j] && !image[i][j - 1]
                        && !image[i + 1][j] && !image[i][j + 1])
                {
                    image[i][j] = false;
                }
            }
        }
    }



    private static void thickenLines(boolean[][] binary_image)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0)
        {
            return;
        }

        int width = binary_image.length;
        int height = binary_image[0].length;
        boolean[][] tmp = new boolean[width][height];

        for (int x = 1; x < width-1; x++)
        {
            for (int y = 1; y < height-1; y++)
            {
                if (!binary_image[x][y])
                {
                    tmp[x][y-1] = tmp[x-1][y] = tmp[x][y] = tmp[x+1][y] = tmp[x][y+1] = true;
                }
            }
        }

        for (int x = 1; x < width-1; x++)
        {
            for (int y = 1; y < height-1; y++)
            {
                if (tmp[x][y])
                {
                    binary_image[x][y] = false;
                }
            }
        }

        for (int x = 1; x < width-1; x++)
        {
            for (int y = 1; y < height-1; y++)
            {
                if (!tmp[x][y])
                {
                    binary_image[x][y-1] = binary_image[x-1][y] = binary_image[x][y]
                            = binary_image[x+1][y] = binary_image[x][y+1] = true;
                }
            }
        }
    }

    BufferedImage getImage(int type)
    {
        switch (type)
        {
        case DataImage.ORIG:
            return ImageFactory.makeGreyImage(m_int_image_initial, false);
        case DataImage.PRE:
            return ImageFactory.makeGreyImage(m_int_image_preprocessed, false);
        case DataImage.ELOG_COLOR:
            if (m_filter_elog == null) { return null; }
            return m_filter_elog.imageColor();
        case DataImage.ELOG_GREY:
            if (m_filter_elog == null) { return null; }
            return m_filter_elog.imageGrey();
        case DataImage.DIFF:
            return ImageFactory.makeGreyImage(m_int_image_filter_difference, true);
        case DataImage.BIN:
            return ImageFactory.makeBinaryImage(m_binary_image, true);
        case DataImage.HOUGH:
            return ImageFactory.makeBinaryImage(m_binary_image_hough, false);
        default:
            return null;
        }
    }

}
