package core;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2014-2016 Benjamin Eltzner
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

public class SLINKClusterer implements HierarchicalClusterer
{

    private List<Node> m_tree_nodes,
                       m_rogue_nodes;
    private Node m_root;

    public SLINKClusterer (List<Clusterable> data)
    {
        m_root = null;
        m_rogue_nodes = new ArrayList<>();
        m_tree_nodes = new ArrayList<>();
        for (Clusterable item : data)
        {
            m_tree_nodes.add(new Node(item));
        }
    }

    public void cluster()
    {
        List<Node> leaves = new ArrayList<>();
        for (Node node : m_tree_nodes)
        {
            if (node.isLeaf())
            {
                leaves.add(node);
            }
        }
        for (Node node : m_rogue_nodes)
        {
            if (node.isLeaf())
            {
                leaves.add(node);
            }
        }
        Collections.sort(leaves);
        m_rogue_nodes = new ArrayList<>();
        m_tree_nodes = leaves;

        int links = leaves.size();
        long[] levels = new long[links],
               auxiliary = new long[links];
        int[]  finals = new int[links];

        // SLINK algorithm.
        for (int n = 0; n < links; n++)
        {
            finals[n] = n;
            levels[n] = Long.MAX_VALUE;
            for (int i = 0; i < n; i++)
            {
                auxiliary[i] = leaves.get(i).distanceSquare(leaves.get(n));
            }
            for (int i = 0; i < n; i++)
            {
                if (levels[i] < auxiliary[i])
                {
                    auxiliary[finals[i]] = Math.min(auxiliary[finals[i]], auxiliary[i]);
                }
                else
                {
                    auxiliary[finals[i]] = Math.min(auxiliary[finals[i]], levels[i]);
                    levels[i] = auxiliary[i];
                    finals[i] = n;
                }
            }
            for (int i = 0; i < n; i++)
            {
                if (levels[finals[i]] <= levels[i])
                {
                    finals[i] = n;
                }
            }
        }

        // Prettifying tree.
        Node[] roots = leaves.toArray(new Node[leaves.size()]);
        Level[] sorted_levels = new Level[links-1];
        for (int i = 0; i < links-1; i++)
        {
            sorted_levels[i] = new Level(i, finals[i], levels[i]);
        }
        Arrays.sort(sorted_levels);

        for(Level l : sorted_levels)
        {
            Node new_node = new Node(roots[l.m_lo], roots[l.m_hi], l.m_score);
            roots[l.m_lo] = new_node;
            roots[l.m_hi] = new_node;
            m_tree_nodes.add(new_node);
        }
        m_root = (Node)root();
    }

    public void trim()
    {
        if (m_tree_nodes == null || m_tree_nodes.isEmpty())
        {
            System.err.println("No tree to trim!");
            return;
        }

        final int a = 100;
        boolean miles_to_go = true;
        while (miles_to_go)
        {
            boolean go_on = true;
            for (int i = m_tree_nodes.size() - 1; i >-1 && go_on; i--)
            {
                Node test_node = m_tree_nodes.get(i);
                if (test_node.isLeaf())
                {
                    if (i == 0)
                    {
                        miles_to_go = false;
                    }
                    continue;
                }

                Node child1 = test_node.child1(),
                     child2 = test_node.child2(),
                     parent = test_node.parent();
                if (test_node.m_level > 1.01 * Math.max(child1.m_level, child2.m_level))
                {
                    int size1 = child1.size(),
                        size2 = child2.size();
                    if (a * size1 < Math.min(a * 10, size2))
                    {
                        child1.orphan();
                        child2.newParent();
                        if (parent != null)
                        {
                            parent.newChild(test_node, child2);
                        }
                        List<Node> branch = child1.branch();
                        m_rogue_nodes.addAll(branch);
                        m_tree_nodes.removeAll(branch);
                        m_tree_nodes.remove(test_node);
                        go_on = false;
                    }
                    else if (a * size2 < Math.min(a * 10, size1))
                    {
                        child1.newParent();
                        child2.orphan();
                        if (parent != null)
                        {
                            parent.newChild(test_node, child1);
                        }
                        List<Node> branch = child2.branch();
                        m_rogue_nodes.addAll(branch);
                        m_tree_nodes.removeAll(branch);
                        m_tree_nodes.remove(test_node);
                        go_on = false;
                    }
                }
                if (i == 0)
                {
                    miles_to_go = false;
                }
            }
        }
    }

    public List<List<Clusterable>> greatestClusters(int total_min, long max_dist)
    {
        List<List<Clusterable>> clusters = new ArrayList<>();
        List<Clusterable> noise = new ArrayList<>();
        while (true)
        {
            cluster();
            List<Node> sieve = m_root.branch_cut(max_dist * max_dist);
            System.out.println("Sizes of trimmed branches:");
            List<Clusterable> trimmed = new ArrayList<>();
            for (Node n : sieve)
            {
                if (n.size() < total_min)
                {
                    System.out.print(n.size() + ", ");
                    if (n.isChildOf(m_root))
                    {
                        m_root = m_root.sibling(n);
                    }
                    trimmed.addAll(n.trim());
                }
            }
            noise.addAll(trimmed);
            /*if (clusters.isEmpty())
            {
                noise.addAll(trimmed);
            }
            else
            {
                clusters.get(clusters.size() - 1).addAll(trimmed);
            }*/
            System.out.println();
            m_tree_nodes = m_root.leaves();
            cluster();
            int min_size = (int)Math.sqrt(getLeaves().size() + total_min * total_min);
            Node cluster = m_root.largestChunk(min_size, 0);
            if (cluster == null)
            {
                clusters.add(m_root.getData());
                clusters.add(noise);
                return clusters;
            }
            if (cluster.isChildOf(m_root))
            {
                m_root = m_root.sibling(cluster);
            }
            clusters.add(cluster.trim());
            m_tree_nodes = m_root.leaves();
        }
    }

    public List<List<Clusterable>> denoise(int total_min, long max_dist)
    {
        List<Clusterable> data = new ArrayList<>();
        List<Clusterable> noise = new ArrayList<>();
        cluster();
        List<Node> sieve = m_root.branch_cut(max_dist * max_dist);
        for (Node n : sieve)
        {
            if (n.size() < total_min)
            {
                noise.addAll(n.getData());
            }
            else
            {
                data.addAll(n.getData());
            }
        }
        List<List<Clusterable>> out = new ArrayList<>();
        out.add(data);
        out.add(noise);
        return out;
    }

    @Override
    public AbstractNode<?> root()
    {
        for (Node n : m_tree_nodes)
        {
            if (n.parent() == null)
            {
                m_root = n;
            }
        }
        return m_root;
    }

    @Override
    public <T extends AbstractNode<?>> List<T> leaves()
    {
        // TODO Auto-generated method stub
        return null;
    }

    @Override
    public Document toXML()
    {
        Document doc = null;
        try
        {
            // Initialize documentBuilder
            doc = DocumentBuilderFactory.newInstance().newDocumentBuilder().newDocument();
        }
        catch (ParserConfigurationException e)
        {
            System.out.println("Fatal: failed to create DocumentBuilder! " + e.getMessage());
            return null;
        }
        return walk((Node)root(), null, doc, "");
    }

    private List<Node> getLeaves()
    {
        List<Node> leaves = new ArrayList<>();
        for (Node n : m_tree_nodes)
        {
            if (n.isLeaf())
            {
                leaves.add(n);
            }
        }
        return leaves;
    }

    private static Document walk(Node node, Element parent, Document doc, String label)
    {
        if (node == null)
        {
            return null;
        }
        Element element = node.element(doc, label);
        if (parent == null)
        {
            doc.appendChild(element);
        }
        else
        {
            parent.appendChild(element);
        }
        walk(node.child1(), element, doc, label + "0");
        walk(node.child2(), element, doc, label + "1");
        return doc;
    }

    private static class Level implements Comparable<Level>
    {
        private int m_lo, m_hi;
        private long m_score;

        private Level(int lo, int hi, long score)
        {
            m_lo = lo;
            m_hi = hi;
            m_score = score;
        }

        @Override
        public int compareTo(Level that)
        {
            if (m_score > that.m_score)
            {
                return 1;
            }
            if (m_score < that.m_score)
            {
                return -1;
            }
            return 0;
        }
    }

    public static class Node extends AbstractNode<Node> implements Comparable<Node>
    {
        private Clusterable m_point;
        private long m_level;
        private int m_size;

        private Node(Node child1, Node child2, long level)
        {
            // Make sure child1 is >= child2
            boolean swap = child1.m_size < child2.m_size;
            setChild1(swap ? child2 : child1);
            setChild2(swap ? child1 : child2);
            child1.setParent(this);
            child2.setParent(this);
            m_point = null;
            m_level = level;
            m_size = child1.m_size + child2.m_size;
        }

        private Node(Clusterable point)
        {
            m_point = point;
            m_level = 0;
            m_size = 1;
        }

        public boolean isLeaf()
        {
            return m_point != null;
        }

        public int label()
        {
            if (m_point == null)
            {
                return -1;
            }
            return m_point.label();
        }

        public List<Clusterable> getData()
        {
            List<Clusterable> data = new ArrayList<>();
            if (m_point != null)
            {
                data.add(m_point);
            }
            else
            {
                data.addAll(child1().getData());
                data.addAll(child2().getData());
            }
            return data;
        }

        private List<Clusterable> trim()
        {
            parent().disappear(this);
            return getData();
        }

        private boolean isChildOf(Node node)
        {
            return parent().equals(node);
        }

        private List<Node> branch_cut (long threshold)
        {
            List<Node> out = new ArrayList<>();
            if (m_level <= threshold)
            {
                out.add(this);
                return out;
            }
            out.addAll(child1().branch_cut(threshold));
            out.addAll(child2().branch_cut(threshold));
            return out;
        }
            

        private Node largestChunk(int min_size, int max_size)
        {
            if (isLeaf() || child1().size() < min_size || m_size < max_size)
            {
                return null;
            }
            int size2 = child2().size();
            Node largest = null;
            if (size2 > min_size)
            {
                if (size2 > max_size)
                {
                    max_size = size2;
                }
                largest = child2();
            }
            Node next = child1().largestChunk(min_size, max_size);
            if (next == null)
            {
                if (largest == null)
                {
                    return null;
                }
                next = child1();
            }
            return (largest == null || next.size() >= max_size ? next : largest);
        }

        private long distanceSquare(Node that)
        {
            if (m_point == null || that.m_point == null)
            {
                return -1;
            }
            return m_point.distanceSquare(that.m_point);
        }

        private int size()
        {
            return m_size;
        }

        private List<Node> branch()
        {
            List<Node> branch = new ArrayList<>();
            branch.add(this);
            if (m_point == null)
            {
                branch.addAll(child1().branch());
                branch.addAll(child2().branch());
            }
            return branch;
        }

        private List<Node> leaves()
        {
            List<Node> leaves = new ArrayList<>();
            if (m_point == null)
            {
                leaves.addAll(child1().leaves());
                leaves.addAll(child2().leaves());
            }
            else
            {
                orphan();
                leaves.add(this);
            }
            return leaves;
        }

        private Element element(Document doc, String identifier)
        {
            Element this_element = doc.createElement("cluster" + identifier);
            this_element.setAttribute("size", "" + size());
            this_element.setAttribute("level", "" + m_level);
            if (isLeaf())
            {
                this_element.setAttribute("point", "" + m_point.label());
            }
            return this_element;
        }

        @Override
        public int compareTo(Node that)
        {
            if (label() > that.label())
            {
                return 1;
            }
            if (label() < that.label())
            {
                return -1;
            }
            return 0;
        }
    }
}
