package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.Point;

import core.Calc;
import ui.FilamentSensor;

public class FilterBinarization
{

    public FilterBinarization()
    {
    }

    public static boolean[][] area(int[][] image, int minimum_mean,
                                                double sigma, double significance)
    {
        significance *= significance;
        final int mean_range = (int) (Math.ceil(2 * sigma)),
                  gauss_range = (int) (Math.ceil(3 * sigma)),
                  step = (2 * mean_range + 1) * (2 * mean_range + 1) * minimum_mean;

        final double[][] gauss = Calc.gaussianMask(gauss_range, sigma);

        final int width = image.length,
                  height = image[0].length;
        int[][] squares = new int[width][height];
        boolean[][] out = new boolean[width][height];

        double[] weights = new double[256];
        for (int i = 0; i < weights.length; i++)
        {
            if (i < 30)
            {
                weights[i] = (double)i * i;
            }
            else
            {
                weights[i] = (4. * (-i + 36 * Math.sqrt(i) - 100)) / 9.;
                weights[i] *= weights[i];
            }
        }

        for (int i = 0; i < gauss_range; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[i][j] = true;
                out[width-1-i][j] = true;
            }
        }
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < gauss_range; j++)
            {
                out[i][j] = true;
                out[i][height-1-j] = true;
            }
        }

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                squares[i][j] = image[i][j] * image[i][j];
            }
        }

        for (int x = gauss_range; x < width-gauss_range; x++)
        {
            int bright = 0;
            for (int y = gauss_range; y < height-gauss_range; y++)
            {
                // Neighborhood mean thresholding
                if (y == gauss_range)
                {
                    bright = 0;
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        for(int b = -mean_range; b <= mean_range; b++)
                        {
                            bright += image[x+a][y+b];
                        }
                    }
                }
                else
                {
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        bright += (image[x+a][y+mean_range] -
                                   image[x+a][y-mean_range-1]);
                    }
                }

                if(bright < step)
                {
                    out[x][y] = true;
                    continue;
                }
                
                // Gaussian adaptive means
                double gauss_mean = 0.0;
                double sum_gauss = 0.0;
                
                for (int a = -gauss_range; a <= gauss_range; a++)
                {
                    for (int b = -gauss_range; b <= gauss_range; b++)
                    {
                        int i = x + a;
                        int j = y + b;
                
                        gauss_mean += (gauss[a + gauss_range][b + gauss_range] * image[i][j]);
                        sum_gauss += gauss[a + gauss_range][b + gauss_range];
                    }
                }
                
                gauss_mean /= sum_gauss;
                if (image[x][y] <= gauss_mean)
                {
                    out[x][y] = true;
                    continue;
                }

                // Statistical area filter
                double gauss_squares = 0;
            
                for (int a = -gauss_range; a <= gauss_range; a++)
                {
                    for (int b = -gauss_range; b <= gauss_range; b++)
                    {
                        int i = x + a;
                        int j = y + b;
            
                        gauss_squares += (squares[i][j] * gauss[a + gauss_range][b + gauss_range]);
                    }
                }
                out[x][y] = (10000.0 * gauss_squares / sum_gauss <=
                             10000.0 * gauss_mean * gauss_mean +
                             weights[(int) gauss_mean] * significance);
            }
        }
        return out;
    }


    public static boolean[][] rod(int[][] image, int minimum_mean,
                                              double sigma, double significance)
    {
        final int mean_range = (int) (Math.ceil(2 * sigma)),
                  gauss_range = (int) (Math.ceil(3 * sigma)),
                  step = (2 * mean_range + 1) * (2 * mean_range + 1) * minimum_mean,
                  width = image.length,
                  height = image[0].length;

        final double[][] gauss = Calc.gaussianMask(gauss_range, sigma);
        final Point[][] lines = latticeInit(gauss_range);

        double[] weights = new double[256];
        for (int i = 0; i < weights.length; i++)
        {
            if (i < 30)
            {
                weights[i] = 1. / (double)i;
            }
            else
            {
                weights[i] = 9. / (4. * (-i + 36 * Math.sqrt(i) - 100));
            }
        }

        boolean[][] out = new boolean[width][height];

        for (int i = 0; i < gauss_range; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[i][j] = true;
                out[width-1-i][j] = true;
            }
        }
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < gauss_range; j++)
            {
                out[i][j] = true;
                out[i][height-1-j] = true;
            }
        }

        for (int x = gauss_range; x < width-gauss_range; x++)
        {
            int bright = 0;
            for (int y = gauss_range; y < height-gauss_range; y++)
            {
                // Neighborhood mean thresholding
                if (y == gauss_range)
                {
                    bright = 0;
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        for(int b = -mean_range; b <= mean_range; b++)
                        {
                            bright += image[x+a][y+b];
                        }
                    }
                }
                else
                {
                    for(int a = -mean_range; a <= mean_range; a++)
                    {
                        bright += (image[x+a][y+mean_range] -
                                   image[x+a][y-mean_range-1]);
                    }
                }

                if(bright < step)
                {
                    out[x][y] = true;
                    continue;
                }

                // Gaussian adaptive means
                double sum = 0.0;
                double sum_gauss = 0.0;

                for (int a = -gauss_range; a <= gauss_range; a++)
                {
                    for (int b = -gauss_range; b <= gauss_range; b++)
                    {
                        int i = x + a;
                        int j = y + b;

                        sum += (gauss[a + gauss_range][b + gauss_range] * image[i][j]);
                        sum_gauss += gauss[a + gauss_range][b + gauss_range];
                    }
                }

                sum /= sum_gauss;
                if (image[x][y] <= sum)
                {
                    out[x][y] = true;
                    continue;
                }

                // Statistical rod filter
                double mean = 0;
                double standard_deviation = 0;

                for (int ori=0; ori<4*gauss_range; ori++)
                {
                    double rod_mean = image[x][y];

                    for (int dist=1; dist<=gauss_range; dist++)
                    {
                        int dx = lines[ori][dist].x,
                            dy = lines[ori][dist].y;

                        rod_mean += (image[x+dx][y+dy] + image[x-dx][y-dy]);
                    }

                    rod_mean /= (double)(2 * gauss_range + 1);
                    standard_deviation += rod_mean * rod_mean;
                    mean += rod_mean;
                }
                mean /= (double) (4 * gauss_range);
                standard_deviation = Math.sqrt(standard_deviation / (double) (4 * gauss_range) -
                                               mean * mean);
                
                out[x][y] = (100.0 * standard_deviation * weights[(int) mean] < significance);
            }
        }
        return out;
    }

    private static Point[][] latticeInit(int radius)
    {
        Point lines[][] = new Point[4*radius][radius+1];

        FilamentSensor.debugMessage("Creating " + 4*radius + " lines.");

        for (int direction = 0; direction < radius; direction++)
        {
            double slope = (double)direction / (double)radius;

            for(int x=0; x<=radius; x++)
            {
                lines[direction][x] = new Point(x, (int)Math.round(slope * x));
            }
        }

        for (int direction=radius; direction<3*radius; direction++)
        {
            double epols = ( 2 * (double)radius - direction ) / radius;
            for(int y=0; y<=radius; y++)
            {
                lines[direction][y] = new Point((int)Math.round(epols * y), y);
            }
        }

        for (int direction=3*radius; direction<4*radius; direction++)
        {
            double slope = (direction - 4 * (double)radius) / radius;
            for(int x=0; x<=radius; x++)
            {
                lines[direction][x] = new Point(x, (int)Math.round(slope * x));
            }
        }

        return lines;
    }

}
