package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import core.Const;

public class FilterCrossCorrelation implements FilterPreprocessing
{

    private double[][][] m_masks;
    private double m_zero_crossing;
    public static final String TYPE = "X";

    public FilterCrossCorrelation(double[][][] masks, double zero_crossing)
    {
        m_masks = masks;
        m_zero_crossing = zero_crossing;
    }

    public int[][] apply(int[][] image)
    {
        if (image == null || m_masks == null ||image.length <= 0 || image[0].length <= 0 ||
            m_masks.length <= 0 || m_masks[0].length <= 0 || m_masks[0][0].length <= 0 ||
            image.length < m_masks[0].length || image[0].length < m_masks[0][0].length ||
            m_masks[0].length % 2 == 0 || m_masks[0][0].length % 2 == 0)
        {
            return image;
        }

        final int width = image.length,
                  height = image[0].length,
                  n_masks = m_masks.length,
                  range_horizontal = ( m_masks[0].length - 1 ) / 2,
                  range_vertical = ( m_masks[0][0].length - 1 ) / 2;
        final double area = (2 * range_horizontal + 1) * (2 * range_vertical + 1);

        // Precalculate prefactors from masks to save time
        double[] factor1 = new double[n_masks], factor2 = new double[n_masks];
        for (int m = 0; m < n_masks; m++)
        {
            double sum = 0.0, square_sum = 0.0;
            for (int a = 0; a <= 2 * range_horizontal; a++)
            {
                for (int b = 0; b <= 2 * range_vertical; b++)
                {
                    sum += m_masks[m][a][b];
                    square_sum += m_masks[m][a][b] * m_masks[m][a][b];
                }
            }
            factor1[m] = area / Math.sqrt(area * square_sum - sum * sum);
            factor2[m] = sum / Math.sqrt(area * square_sum - sum * sum);
        }

        // Precalculate matrix of squares to save time
        double[][] squares = new double[width][height];

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                squares[i][j] = image[i][j] * image[i][j];
            }
        }

        int[][] out = new int[width][height];

        // Edges of image remain unchanged
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < range_vertical; j++)
            {
                out[i][j] = image[i][j];
                out[i][height-j-1] = image[i][height-j-1];
            }
        }
        for (int j = 0; j < height; j++)
        {
            for (int i = 0; i < range_horizontal; i++)
            {
                out[i][j] = image[i][j];
                out[width-i-1][j] = image[width-i-1][j];
            }
        }

        // Calculate the maximum cross correlation for each point.
        double total_min = Double.MAX_VALUE;
        double total_max = Double.MIN_VALUE;
        double first_sum = 0.0;
        double first_square_sum = 0.0;
        for (int a = 0; a <= 2 * range_horizontal; a++)
        {
            for (int b = 0; b <= 2 * range_vertical; b++)
            {
                first_sum += image[a][b];
                first_square_sum += squares[a][b];
            }
        }

        for (int i = range_horizontal; i < width - range_horizontal; i++)
        {
            if (i > range_horizontal)
            {
                for (int b = 0; b <= 2 * range_vertical; b++)
                {
                    first_sum += (image[i+range_horizontal][b] -
                                  image[i-range_horizontal-1][b]);
                    first_square_sum += (squares[i+range_horizontal][b] -
                                         squares[i-range_horizontal-1][b]);
                }
            }
            double window_sum = first_sum;
            double window_square_sum = first_square_sum;

            for (int j = range_vertical; j < height - range_vertical; j++)
            {
                if (j > range_vertical)
                {
                    for (int a = - range_horizontal; a <= range_horizontal; a++)
                    {
                        window_sum += (image[i+a][j+range_vertical] -
                                       image[i+a][j-range_vertical-1]);
                        window_square_sum += (squares[i+a][j+range_vertical] -
                                              squares[i+a][j-range_vertical-1]);
                    }
                }

                double max = Double.MIN_VALUE;
                if (window_square_sum == 0.0)
                {
                    out[i][j] = 0;
                    continue;
                }

                double factor = 1 / Math.sqrt(area * window_square_sum - (double)window_sum * window_sum);
                for (int m = 0; m < n_masks; m++)
                {
                    double cross_sum = 0.0;
                    for (int a = - range_horizontal; a <=  range_horizontal; a++)
                    {
                        for (int b = - range_vertical; b <= range_vertical; b++)
                        {
                            cross_sum += image[i+a][j+b] *
                                         m_masks[m][range_horizontal + a][range_vertical + b];
                        }
                    }
                    double correlation = factor * (factor1[m] * cross_sum - factor2[m] * window_sum);
                    if (correlation > max)
                    {
                        max = correlation;
                    }
                }
                if (max < total_min)
                {
                    total_min = max;
                }
                if (max > total_max)
                {
                    total_max = max;
                }

                out[i][j] = (int)(Const.M * max);
            }
        }

        total_min += m_zero_crossing * (total_max + total_min);
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[i][j] = (int)(255 * (out[i][j] / Const.MF - total_min) / (total_max - total_min));
                if (out[i][j] < 0)
                {
                    out[i][j] = 0;
                }
            }
        }
        return out;
    }

    @Override
    public String type()
    {
        return TYPE;
    }

    @Override
    public String toXML()
    {
        return TYPE + ";" + (m_masks.length / 4) + ";" + Math.round(100 * m_zero_crossing);
    }
}
