package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.image.BufferedImage;

import core.Calc;
import ui.FilamentSensor;
import ui.ImageFactory;

public class FilterELoG
{

    private int[][] m_int_image_elog, m_int_image_elog_alt;
    private int[] m_elog_histogram;
    private BufferedImage m_buffered_image_elog;

    public FilterELoG()
    {
    }

    public void calculateELoGs(int[][] image, int tolerance, int contrast, int threshold,
                               int orientations, double sigma_x, double sigma_y)
    {
        double[][][] gaussians = gaussians(orientations, sigma_x, sigma_y),
                     masks = new double[gaussians.length][][];

        // Invert order to correct angle definitions.
        for (int ori = 0; ori < gaussians.length; ori++)
        {
            masks[ori] = convolve(gaussians[gaussians.length - 1 - ori],
                                  new double[][]{{0,-1,0},
                                                 {-1,4,-1},
                                                 {0,-1,0}});
        }

        FilamentSensor.debugMessage("Processing eLoGs...");

        double total_max = Double.MIN_VALUE,
               total_min = Double.MAX_VALUE;
        final int width = image.length,
                  height = image[0].length;
        double[][] tmp = new double[width][height];
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                if (image[i][j] > threshold)
                {
                    ELoGData elog = new ELoGData(image, i, j, masks, false);
                    m_int_image_elog[i][j] = elog.max_ori;
                    tmp[i][j] = elog.max;

                    if(elog.max > total_max)
                    {
                        total_max = elog.max;
                    }
                    if(elog.max < total_min)
                    {
                        total_min = elog.max;
                    }
                }
                else
                {
                    m_int_image_elog[i][j] = -1;
                }
            }

            if (i % (width / 10) == 0)
            {
                FilamentSensor.debugMessage((100 * i / width) + "%");
            }
        }

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                if (image[i][j] > threshold)
                {
                    m_int_image_elog_alt[i][j] = Math.min(255, (int)( contrast * 25.5 *
                                                                      (tmp[i][j] - total_min) /
                                                                      (total_max - total_min) ));
                }
                else
                {
                    m_int_image_elog_alt[i][j] = 0;
                }
            }
        }
        
        boolean[][] elog_image_otsu = (new FilterHistogramMethods(m_int_image_elog_alt, true, 1,
                                                                  FilterHistogramMethods.OTSU)).binaryImage();

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                if (!elog_image_otsu[i][j])
                {
                    m_int_image_elog[i][j] = -1;
                }
                else if (m_int_image_elog[i][j] != -1)
                {
                    m_elog_histogram[m_int_image_elog[i][j]]++;
                }
            }
        }
        m_buffered_image_elog = ImageFactory.makeColorImage(m_int_image_elog, orientations, false);

        FilamentSensor.debugMessage("100% done.");
    }

    public BufferedImage imageGrey()
    {
        return ImageFactory.makeGreyImage(m_int_image_elog_alt, false);
    }

    public BufferedImage imageColor()
    {
        return m_buffered_image_elog;
    }

    private static double[][][] gaussians(int orientations, double sigma_x, double sigma_y)
    {
        int range = (int)Math.ceil(3*Math.max(sigma_x, sigma_y));
        double phi_step = 180.0 / orientations;

        double[] phi = new double[orientations];
        double[][][] gaussians = new double[orientations][][];

        FilamentSensor.debugMessage("Processing eLoGs...");

        for (int ori = 0; ori < orientations; ori++)
        {
            phi[ori] = Math.toRadians(ori * phi_step);
            gaussians[ori] = Calc.gaussianMask(range, phi[ori], sigma_x, sigma_y);
        }

        return gaussians;
    }

    private static double[][] convolve(double[][] image, double[][] mask)
    {
        if (image.length <= 0 || image[0].length <= 0 || mask.length <= 0 || mask[0].length <= 0 ||
            image.length < mask.length || image[0].length < mask[0].length ||
            mask.length % 2 == 0 || mask[0].length % 2 == 0)
        {
            return null;
        }

        int width = image.length;
        int height = image[0].length;
        int range_horizontal = ( mask.length - 1 ) / 2;
        int range_vertical = ( mask[0].length - 1 ) / 2;

        double out[][] = new double[width][height];

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < range_vertical; j++)
            {
                out[i][j] = image[i][j];
                out[i][height-j-1] = image[i][height-j-1];
            }
        }
        for (int j = 0; j < height; j++)
        {
            for (int i = 0; i < range_horizontal; i++)
            {
                out[i][j] = image[i][j];
                out[height-i-1][j] = image[height-i-1][j];
            }
        }
        for (int i = range_horizontal; i < width - range_horizontal; i++)
        {
            for (int j = range_vertical; j < height - range_vertical; j++)
            {
                for (int a = -range_horizontal; a <= range_horizontal; a++)
                {
                    for (int b = -range_vertical; b <= range_vertical; b++)
                    {
                        out[i][j] += mask[a + range_horizontal][b + range_vertical] * image[i + a][j + b];
                    }
                }
            }
        }
        return out;
    }

    /**
     * Simple wrapper class for return values of eLoG.
     */
    private class ELoGData
    {
        private int max_ori = -1;
        private double min = Double.MAX_VALUE, max = 0;

        private ELoGData(int[][] image, int x, int y, double[][][] masks, boolean calc_min)
        {
            int orientations = masks.length;
            final int width = image.length,
                      height = image[0].length;

            if (orientations == 0)
            {
                return;
            }
            
            int range = (masks[0].length - 1) / 2;

            double response[] = new double[orientations];

            for (int ori = 0; ori < orientations; ori++)
            {
                response[ori] = 0.0;

                for (int dx = -range; dx <= range; dx++)
                {
                    for (int dy = -range; dy <= range; dy++)
                    {
                        int a = x + dx;
                        int b = y + dy;

                        if (a >= 0 && a < width && b >= 0 && b < height)
                        {
                            response[ori] += image[a][b] * masks[ori][range+dx][range+dy];
                        }
                    }
                }

                if (response[ori] > max)
                {
                    max = response[ori];
                    max_ori = ori;
                }
            }

            if(calc_min)
            {
                for(int z=0; z<=orientations/3; z++)
                {
                    int test_ori = (max_ori + orientations / 3 + z) % orientations;
                    if(response[test_ori] < min)
                    {
                        min = response[test_ori];
                    }
                }
            }
        }
    }
}
