package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import core.Calc;

public class FilterGauss implements FilterPreprocessing
{

    private double m_sigma;
    public static final String TYPE = "G";

    public FilterGauss(double sigma)
    {
        m_sigma = sigma;
    }

    /**
     * Applies a gaussian filter to m_image_preprocessed. The size of the
     * convolution mask is chosen such that a maximally bright boundary
     * contributes at most one unit of brightness.
     */
    public final int[][] apply(int[][] image)
    {
        if (m_sigma == 0.0)
        {
            return image;
        }
        // Find range such that the boundary is not too bright.
        int range = 2;
        while (8. * 255. * range * Math.exp(-range * range/(2.0 * m_sigma * m_sigma)) /
               (2.0 * Math.PI * m_sigma * m_sigma) > 1)
        {
            range++;
        }

        double[][] mask = Calc.gaussianMask(range, m_sigma);
        final int width = image.length,
                  height = image[0].length;
        int out[][] = new int[width][height];

        // Left and right boundary
        for (int i = 0; i < width; i++)
        {
            if (i == range)
            {
                i = width - range;
            }
            for (int j = 0; j < height; j++)
            {
                double sum = 0.0;
                double bias = 0.0;

                for (int a = -range; a <= range; a++)
                {
                    for (int b = -range; b <= range; b++)
                    {
                        if (i+a >= 0 && i+a < width && j+b >=0 && j+b < height)
                        {
                            sum += mask[a + range][b + range] * image[i + a][j + b];
                            bias += mask[a + range][b + range];
                        }
                    }
                }
                out[i][j] = (int) (sum / bias);
            }
        }
        // Upper and lower boundary
        for (int j = 0; j < height; j++)
        {
            if (j == range)
            {
                j = height - range;
            }
            for (int i = 0; i < width; i++)
            {
                double sum = 0.0;
                double bias = 0.0;

                for (int a = -range; a <= range; a++)
                {
                    for (int b = -range; b <= range; b++)
                    {
                        if (i+a >= 0 && i+a < width && j+b >=0 && j+b < height)
                        {
                            sum += mask[a + range][b + range] * image[i + a][j + b];
                            bias += mask[a + range][b + range];
                        }
                    }
                }

                out[i][j] = (int) (sum / bias);
            }
        }
        // Interior
        for (int i = range; i < width - range; i++)
        {
            for (int j = range; j < height - range; j++)
            {
                double sum = 0.0;
                double bias = 0.0;

                for (int a = -range; a <= range; a++)
                {
                    for (int b = -range; b <= range; b++)
                    {
                        sum += mask[a + range][b + range] * image[i + a][j + b];
                        bias += mask[a + range][b + range];
                    }
                }

                out[i][j] = (int) (sum / bias);
            }
        }
        return out;
    }

    @Override
    public String type()
    {
        return TYPE;
    }

    @Override
    public String toXML()
    {
        return TYPE + ";" + Math.round(10 * m_sigma);
    }
}
