package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import core.Const;
import ui.FilamentSensor;

public class FilterHistogramMethods
{

    private boolean[][] m_binary_image;
    private int[][] m_truncated_image;
    private int[] m_brightness_histogram;
    private int m_width, m_height, m_threshold, m_black, m_white;
    public static final int OTSU = 0, LI = 1, TRIANGLE = 2;

    /**
     * This is the constructor to obtain the Otsu binarized image.
     */
    public FilterHistogramMethods(int[][] int_image, boolean make_images, int lower_bound, int type)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            m_threshold = - 1;
            return;
        }

        m_width = int_image.length;
        m_height = int_image[0].length;

        m_brightness_histogram = makeHistogram(int_image, 0, 255);

        int number_of_pixels = m_width * m_height;
        if (lower_bound > 0)
        {
            number_of_pixels = 0;
            for (int i = lower_bound; i < 256; i++)
            {
                number_of_pixels += m_brightness_histogram[i];
            }
            m_brightness_histogram[0] = number_of_pixels;
            for (int i = 1; i < lower_bound; i++)
            {
                m_brightness_histogram[i] = 0;
            }
            number_of_pixels *= 2;
        }
        
        m_threshold = calcThreshold(number_of_pixels, 0, 255, type);

        if (make_images)
        {
            makeImages(int_image);
        }
    }
    
    public FilterHistogramMethods(int[][] int_image, boolean make_images, int type)
    {
        this(int_image, make_images, 0, type);
    }

    /**
     * This constructor is for contrast enhancement.
     * MAY CHANGE INPUT int_image!
     */
    public FilterHistogramMethods(int[][] int_image, int min, int max)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            m_threshold = - 1;
            return;
        }

        m_width = int_image.length;
        m_height = int_image[0].length;

        m_brightness_histogram = makeHistogram(int_image, min, max);

        // Get Otsu m_threshold as upper bound for m_black.
        int black_threshold = calcThreshold(m_width * m_height, 0, 255, OTSU);
        int new_number = 0;

        for (int i = black_threshold; i < 256; i++)
        {
            new_number += m_brightness_histogram[i];
        }

        // Get Otsu m_threshold for the part of the histogram above
        // black_threshold as a lower bound for m_white.
        int white_threshold = black_threshold;
        int n_over = new_number;
        while(new_number / n_over < 100 && white_threshold < 254) // This could be adjustable.
        {
            int tmp = white_threshold;
            white_threshold = calcThreshold(n_over, white_threshold, 255, OTSU);
            if (tmp == white_threshold)
            {
                break;
            }

            n_over = 0;
            for (int i = white_threshold; i < 256; i++)
            {
                n_over += m_brightness_histogram[i];
            }
        }

        calcBlackWhite(new_number);
        m_black = (m_black > black_threshold ? black_threshold :
                    (m_black < black_threshold / 2 ? black_threshold / 2 : m_black));
        m_white = (m_white < white_threshold ? white_threshold : m_white);
        m_white = (m_white < m_black + 128 ? m_black + 128 : m_white);
        m_white = (m_white > 255 ? 255 : m_white);

        FilamentSensor.debugMessage("After adjustment: m_black = " + m_black + ". m_white = " + m_white + ".");
    }

    public static int[][] enhanceContrast(int[][] image, int black, int white)
    {
        double bw_scale = 255 / ((double) (white - black));
        int width = image.length;
        int height = image[0].length;
        int[][] out = new int[width][height];

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                int pixel = image[i][j];
                pixel = (int) Math.round(bw_scale * (pixel - black));
                pixel = (pixel < 0 ? 0 : pixel);
                pixel = (pixel > 255 ? 255 : pixel);

                out[i][j] = pixel;
            }
        }
        return out;
    }

    /**
     * MAY CHANGE INPUT int_image!
     */
    private int[] makeHistogram(int[][] int_image, int min, int max)
    {
        int[] histogram = new int[256];
        double scale = 255 / ((double) (max - min));

        for (int i = 0; i < m_width; i++)
        {
            for (int j = 0; j < m_height; j++)
            {
                if(min != 0 || max != 255)
                {
                    int_image[i][j] = (int_image[i][j] < min ? 0 :
                                      (int_image[i][j] > max ? 255 :
                                       (int) Math.round(scale * (int_image[i][j] - min))));
                }
                histogram[int_image[i][j]]++;
            }
        }
        return histogram;
    }

    private int calcThreshold(int number_of_pixels, int min, int max, int type)
    {
        long sum_below = 0, sum_above = 0;
        double goodness_max = 0.0;

        long n_below = 0;
        long n_above = number_of_pixels;
        int thresh = 0;

        for (int i = min; i < max + 1; i++)
        {
            sum_above += (i + 1) * m_brightness_histogram[i];
        }

        switch (type)
        {
        case OTSU:
            FilamentSensor.debugMessage("Calculating Otsu m_threshold...");
            for (int t = min; t < max; t++)
            {
                n_below += m_brightness_histogram[t];
                n_above -= m_brightness_histogram[t];

                sum_below += (t + 1) * m_brightness_histogram[t];  // including t
                sum_above -= (t + 1) * m_brightness_histogram[t];  // excluding t

                double goodness = ((n_below * n_above) <= 0 ? 0 :
                                      (double) (n_below * sum_above - n_above * sum_below) *
                                      (double) (n_below * sum_above - n_above * sum_below) /
                                      (double) (n_below * n_above));

                if (goodness > goodness_max)
                {
                    goodness_max = goodness;
                    thresh = t;
                }
            }
            break;
        case LI:
            FilamentSensor.debugMessage("Calculating Li m_threshold...");
            double[] mean_below = new double[max - min], mean_above = new double[max - min];
            for (int t = min; t < max; t++)
            {
                n_below += m_brightness_histogram[t];
                n_above -= m_brightness_histogram[t];

                sum_below += (t + 1) * m_brightness_histogram[t];  // including t
                sum_above -= (t + 1) * m_brightness_histogram[t];  // excluding t

                mean_below[t - min] = (n_below == 0 ? 0 : sum_below / (double) n_below);
                mean_above[t - min] = (n_above == 0 ? mean_above[t - 1 - min] : sum_above / (double) n_above);
            }
            
            int t_old = -1, t_new = 1;
            while (t_old != t_new)
            {
                t_old = t_new;
                t_new = (Math.log(mean_below[t_old]) <= 0 ?
                         (int) Math.round(mean_above[t_old] / Math.log(mean_above[t_old])) :
                         (int) Math.round((mean_below[t_old] - mean_above[t_old]) /
                                          (Math.log(mean_below[t_old]) - Math.log(mean_above[t_old]))));
            }
            thresh = t_new - 1;
            break;
        case TRIANGLE:
            FilamentSensor.debugMessage("Calculating triangle m_threshold...");
            int max_value = 0, max_pos = 0, min_pos = 0;

            for (int i = m_brightness_histogram.length - 1; i > 0; i--)
            {
                if (m_brightness_histogram[i] > 0)
                {
                    min_pos = i;
                    break;
                }
            }
            if (min_pos < m_brightness_histogram.length - 1)
            {
                min_pos++;
            }

            for (int i = 0; i <= min_pos; i++)
            {
                if (m_brightness_histogram[i] > max_value)
                {
                    max_pos = i;
                    max_value = m_brightness_histogram[i];
                }
            }

            if (min_pos == max_pos)
            {
                return min_pos;
            }

            // use up-right pointing normal to the connecting line of max and min
            int normal_x = m_brightness_histogram[max_pos],
                normal_y = min_pos - max_pos,
                min_intercept = Integer.MAX_VALUE;
            thresh = min_pos;
            for (int i = max_pos; i < min_pos; i++)
            {
                int intercept = normal_x * i + normal_y * m_brightness_histogram[i];
                // Lowest scalar product with normal gives the line parallel to
                // the min-max-connection which is furthest below it.
                if (intercept < min_intercept)
                {
                    thresh = i;
                    min_intercept = intercept;
                }
            }
            thresh--;
            break;
        default:
            thresh = 0;
        }

        FilamentSensor.debugMessage("Threshold found: " + thresh);

        return thresh;
    }

    private void makeImages(int[][] int_image)
    {
        m_binary_image = new boolean[m_width][m_height];
        m_truncated_image = new int[m_width][m_height];

        for (int i = 0; i < m_width; i++)
        {
            for (int j = 0; j < m_height; j++)
            {
                m_binary_image[i][j] = int_image[i][j] > m_threshold;
                m_truncated_image[i][j] = (int_image[i][j] > m_threshold ? int_image[i][j] : 0);
            }
        }

        // clean single pixels
        for (int i = 1; i < m_width - 1; i++)
        {
            for (int j = 1; j < m_height - 1; j++)
            {
                if (m_binary_image[i][j] && !m_binary_image[i - 1][j] &&
                   !m_binary_image[i][j - 1] && !m_binary_image[i + 1][j] &&
                   !m_binary_image[i][j + 1])
                {
                    m_binary_image[i][j] = false;
                    m_truncated_image[i][j] = 0;
                }
            }
        }
    }

    private void calcBlackWhite(int bright_pixels)
    {
        m_black = 0;
        m_white = 255;
        int step = bright_pixels / Const.CONTRAST_FACTOR;
        step = (step > 10 ? 10 : step);

        for (int i = 1; i < 255; i++)
        {
            if(FilamentSensor.VERBOSE())
            {
                if ((i - 1) % 10 == 0)
                {
                    System.out.print("Histogram " + i + "-" + (i+9) + ": ");
                }
                System.out.print(m_brightness_histogram[i] + ", ");
                if (i % 10 == 0 || i == 254)
                {
                    System.out.println("");
                }
            }

            // Compensate for bright disturbances.
            if (m_brightness_histogram[i] > m_brightness_histogram[i + 1] + step && m_brightness_histogram[i + 1] != 0)
            {
                m_white = i + 1;
            }

            // Compensate for bright background.
            if (m_brightness_histogram[254 - i] > m_brightness_histogram[255 - i] + 400 * step && m_brightness_histogram[255 - i] != 0)
            {
                m_black = 254 - i;
            }
        }

        FilamentSensor.debugMessage("Initial try: m_black = " + m_black + ". m_white = " + m_white + ". step = " + step);
    }

    public boolean[][] binaryImage()
    {
        return m_binary_image;
    }

    public int threshold()
    {
        return m_threshold;
    }

    public int black()
    {
        return m_black;
    }

    public int white()
    {
        return m_white;
    }
}
