package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import ui.FilamentSensor;

public class FilterLaplace implements FilterPreprocessing
{

    private double[] m_factors;
    public static final String TYPE = "L";
    
    public FilterLaplace(int type, double factor)
    {
        switch (type)
        {
        case 0:
            m_factors = new double[]{1, factor, 0};
            break;
        case 1:
            m_factors = new double[]{1, 0, factor};
            break;
        default:
            FilamentSensor.debugError("Unknown laplace mask-type: " + type);
            return;
        }
    }

    public FilterLaplace(double factor1, double factor2, double factor3)
    {
        m_factors = new double[]{factor1, factor2, factor3};
    }

    /**
     * Applies a laplace filter to m_image_preprocessed. The convolution mask is chosen from a
     * predefined list and multiplied by lpfac.
     */
    public int[][] apply(int[][] image)
    {
        if (image == null || image.length <= 0 || image[0].length <= 0 || m_factors == null)
        {
            return null;
        }

        double a = -m_factors[2], b = -m_factors[2] - m_factors[1],
               c = m_factors[0] + 4 * m_factors[1] + 8 * m_factors[2];
        double[][] mask = new double[][]{
            {a, b, a},
            {b, c, b},
            {a, b, a}};

        int width = image.length;
        int height = image[0].length;
        int tmp[][] = new int[width][height];

        for (int i = 0; i < width; i++)
        {
            tmp[i][0] = image[i][0];
            tmp[i][height-1] = image[i][height-1];
        }
        for (int i = 0; i < height; i++)
        {
            tmp[0][i] = image[0][i];
            tmp[width-1][i] = image[width-1][i];
        }

        for (int x = 1; x < width - 1; x++)
        {
            for (int y = 1; y < height - 1; y++)
            {
                int sum = 0;

                for (int i = -1; i <= 1; i++)
                {
                    for (int j = -1; j <= 1; j++)
                    {
                        sum += mask[i + 1][j + 1] * image[x + i][y + j];
                    }
                }
                tmp[x][y] = Math.min(Math.max((int)sum, 0), 255);
            }
        }

        return tmp;
    }

    @Override
    public String type()
    {
        return TYPE;
    }

    @Override
    public String toXML()
    {
        return (TYPE + ";" + Math.round(10 * m_factors[0]) + ";" +
                Math.round(10 * m_factors[1]) + ";" + Math.round(10 * m_factors[2]));
    }
}
