package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.Point;

import core.Calc;
import ui.FilamentSensor;

public class FilterLineGauss implements FilterPreprocessing
{

    private double m_sigma;
    private boolean[][] m_cell_area;
    private int m_masked;
    public static final String TYPE = "D";

    public FilterLineGauss(double sigma, boolean[][] cell_area)
    {
        m_sigma = sigma;
        m_cell_area = cell_area;
        m_masked = (m_cell_area == null ? 0 : 1);
    }

    public int[][] apply(int[][] image)
    {
        final int range = (int)Math.ceil(3 * m_sigma);
        final double[][] mask = Calc.gaussianQuarterMask(range, m_sigma);
        final Point[][] lines = latticeInit(range);
        final int width = image.length,
                  height = image[0].length;
        int tmp[][] = new int[width][height];

        // In case cell area is given as null, apply filter everywhere.
        if (m_cell_area == null)
        {
            m_cell_area = new boolean[width][height];
            for (int x = range; x < width-range; x++)
            {
                for (int y = range; y < height-range; y++)
                {
                    m_cell_area[x][y] = true;
                }
            }
        }

        for (int x = range; x < width-range; x++)
        {
            for (int y = range; y < height-range; y++)
            {
                double max_ori_mean = -1;

                if(m_cell_area[x][y])
                {
                    for (int ori=0; ori<4*range; ori++)
                    {
                        double mean = image[x][y] * mask[0][0],
                               gauss_sum = mask[0][0];

                        for (int dist=1; dist<=range; dist++)
                        {
                            int dx = lines[ori][dist].x, dy = lines[ori][dist].y,
                                mx = Math.abs(dx), my = Math.abs(dy);

                            mean += (image[x+dx][y+dy] + image[x-dx][y-dy]) * mask[mx][my];
                            gauss_sum += 2 * mask[mx][my];
                        }

                        // TODO: Check for "median" above mean?
                        mean /= gauss_sum;

                        if(mean > max_ori_mean)
                        {
                            max_ori_mean = mean;
                        }
                    }
                    tmp[x][y] = (max_ori_mean >= 0 ? (int)max_ori_mean : image[x][y]);
                }
                else
                {
                    tmp[x][y] = image[x][y];
                }
            }
        }
        return tmp;
    }

    private static Point[][] latticeInit(int radius)
    {
        Point lines[][] = new Point[4*radius][radius+1];

        FilamentSensor.debugMessage("Creating " + 4*radius + " lines.");

        for (int direction = 0; direction < radius; direction++)
        {
            double slope = (double)direction / (double)radius;

            for(int x=0; x<=radius; x++)
            {
                lines[direction][x] = new Point(x, (int)Math.round(slope * x));
            }
        }

        for (int direction=radius; direction<3*radius; direction++)
        {
            double epols = ( 2 * (double)radius - direction ) / radius;
            for(int y=0; y<=radius; y++)
            {
                lines[direction][y] = new Point((int)Math.round(epols * y), y);
            }
        }

        for (int direction=3*radius; direction<4*radius; direction++)
        {
            double slope = (direction - 4 * (double)radius) / radius;
            for(int x=0; x<=radius; x++)
            {
                lines[direction][x] = new Point(x, (int)Math.round(slope * x));
            }
        }

        return lines;
    }

    @Override
    public String type()
    {
        return TYPE;
    }

    @Override
    public String toXML()
    {
        return TYPE + ";" + m_masked + ";" + Math.round(10 * m_sigma);
    }

}
