package filters;
/*
 * This file is part of ElementSensor - A tool for element detection on images
 *
 * Copyright (C) 2016 Benjamin Eltzner
 *
 * ElementSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import core.Calc;

public class FilterSemiLocalContrast implements FilterPreprocessing
{

    public static final String TYPE = "SLC";
    private int m_size, m_min_diff;

    public FilterSemiLocalContrast(int size, int min_diff)
    {
        m_size = size;
        m_min_diff = min_diff;
    }

    @Override
    public int[][] apply(int[][] image)
    {
        final int width = image.length, height = image[0].length;
        int[][][] fields = fields(image);
        int[][] out = new int[width][height];
        
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                out[x][y] = (image[x][y] < fields[0][x][y] ? 0 :
                                (image[x][y] > fields[1][x][y] ? 255 :
                                    (int)Math.round(255 * (image[x][y] - fields[0][x][y]) /
                                                    (float)(fields[1][x][y] - fields[0][x][y]))));
            }
        }
        return out;
    }

    private int[][][] fields(int[][] image)
    {
        final int width = image.length, height = image[0].length,
                  x_blocks = (width - 1) / m_size + 1, y_blocks = (height - 1) / m_size + 1,
                  x_offset = (width - x_blocks * m_size) / 2,
                  y_offset = (height - y_blocks * m_size) / 2,
                  x_1st_center = (width - (x_blocks - 1) * m_size) / 2,
                  y_1st_center = (height - (y_blocks - 1) * m_size) / 2;
        int[][][] seeds = new int[2][x_blocks][y_blocks];

        // Make seed image for black/white fields.
        for (int x = 0; x < x_blocks; x ++)
        {
            for (int y = 0; y < y_blocks; y ++)
            {
                int x0 = x_offset + x * m_size, y0 = y_offset + y * m_size,
                    min_brightness = Integer.MAX_VALUE, max_brightness = 0, sum = 0, count = 0;

                for (int dx = 0; dx < m_size; dx++)
                {
                    int x_val = x0 + dx;
                    if (x_val < 0 || x_val >= width)
                    {
                        continue;
                    }
                    for (int dy = 0; dy < m_size; dy++)
                    {
                        int y_val = y0 + dy;
                        if (y_val < 0 || y_val >= height)
                        {
                            continue;
                        }
                        sum += image[x_val][y_val];
                        count++;
                        if (image[x_val][y_val] > max_brightness)
                        {
                            max_brightness = image[x_val][y_val];
                        }
                        if (image[x_val][y_val] < min_brightness)
                        {
                            min_brightness = image[x_val][y_val];
                        }
                    }
                }
                max_brightness = Math.min(max_brightness,
                                          (int)Math.round(2.0 * sum / ((double) count) - min_brightness));
                if (max_brightness - min_brightness < m_min_diff)
                {
                    int mean = (min_brightness + max_brightness) / 2;
                    if (mean < m_min_diff / 2)
                    {
                        min_brightness = 0;
                        max_brightness = 50;
                    }
                    else if (mean > 255 - m_min_diff / 2)
                    {
                        min_brightness = 205;
                        max_brightness = 255;
                    }
                    else
                    {
                        min_brightness = mean - 25;
                        max_brightness = mean + 25;
                    }
                }
                seeds[0][x][y] = min_brightness;
                seeds[1][x][y] = max_brightness;
            }
        }

        int range = 5 * m_size;
        double[][] mask = Calc.gaussianMask(range, m_size);

        int[][][] out = new int[2][width][height];

        // Make smooth orientation fields.
        double gauss_min = 0.0, gauss_max = 0.0, sum_min = 0.0, sum_max = 0.0;
        final int x_step = m_size - 1 - x_1st_center - range,
                  y_step = m_size -1 - y_1st_center - range;
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                int x_block = Math.max(0, ((x + x_step) / m_size)),
                    y0_block = Math.max(0, ((y + y_step) / m_size)),
                    left = x_1st_center + x_block * m_size - x, right = Math.min(range, width - x),
                    top = y_1st_center + y0_block * m_size - y, bottom = Math.min(range, height - y);
                gauss_min = gauss_max = sum_min = sum_max = 0.0;
                for (int dx = left; dx < right; dx += m_size)
                {
                    int y_block = y0_block;
                    for (int dy = top; dy < bottom; dy += m_size)
                    {
                        gauss_min += seeds[0][x_block][y_block] * mask[range + dx][range + dy];
                        sum_min += mask[range + dx][range + dy];
                        gauss_max += seeds[1][x_block][y_block++] * mask[range + dx][range + dy];
                        sum_max += mask[range + dx][range + dy];
                    }
                    x_block++;
                }

                out[0][x][y] = (sum_min <= 0 ? 0 : (int) Math.round(gauss_min / sum_min));
                out[1][x][y] = (sum_max <= 0 ? 0 : (int) Math.round(gauss_max / sum_max));
            }
        }
        return out;
    }

    @Override
    public String type()
    {
        return TYPE;
    }

    @Override
    public String toXML()
    {
        return TYPE;
    }

    
}
