package filters;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2014-2016 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.Point;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import core.Calc;
import core.Const;
import ui.FilamentSensor;

public class PluginCellShape
{

 // Orientation, aspect ratio, and order parameter are scaled by Const.M.
    private long m_orientation, m_new_orientation, m_aspect_ratio,
                 m_order_parameter, m_new_order_parameter, m_excursion;
    private int m_area;
    private Point m_center;
    private boolean[][] m_binary_image, m_binary_image_tri;
    private int[][] m_excursion_image;

    PluginCellShape(boolean[][] binary_image, boolean[][] larger_mask, boolean[][] mask)
    {
        binary_image = Calc.close(binary_image, Calc.circleMask(30));
        final int width = binary_image.length,
                  height = binary_image[0].length;
        if (mask != null)
        {
            final int mask_width = mask.length,
                      mask_height = mask[0].length;
            if (mask_width == width && mask_height == height)
            {
                for (int x = 0; x < width; x++)
                {
                    for (int y = 0; y < height; y++)
                    {
                        binary_image[x][y] = binary_image[x][y] && mask[x][y];
                    }
                }
            }
        }
        binary_image = FilterCloseHoles.apply(clearEdges(largestObject(binary_image)));

        double sum_x = 0, sum_y = 0, sum_xx = 0, sum_xy = 0, sum_yy = 0;
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    m_area++;
                    sum_x += x;
                    sum_y += y;
                    sum_xx += x * x;
                    sum_xy += x * y;
                    sum_yy += y * y;
                }
            }
        }
        double area = (double) m_area;
        if (m_area == 0)
        {
            m_orientation = m_new_orientation = m_aspect_ratio = m_order_parameter =
                m_new_order_parameter= 0;
            m_center = new Point(-1, -1);
            m_binary_image_tri = m_binary_image = binary_image;
            FilamentSensor.debugError("No cell found!");
            return;
        }
        int mean_x = (int)Math.round(sum_x / area);
        int mean_y = (int)Math.round(sum_y / area);
        double cov_xx = sum_xx / area - sum_x * sum_x / (area * area);
        double cov_xy = sum_xy / area - sum_x * sum_y / (area * area);
        double cov_yy = sum_yy / area - sum_y * sum_y / (area * area);
        double ew_hi = (cov_xx + cov_yy) / 2.0 +
                       Math.sqrt((cov_xx - cov_yy) * (cov_xx - cov_yy) / 4.0 + cov_xy * cov_xy);
        double ew_lo = (cov_xx + cov_yy) / 2.0 -
                       Math.sqrt((cov_xx - cov_yy) * (cov_xx - cov_yy) / 4.0 + cov_xy * cov_xy);
        double phi = 180 - Math.atan2(ew_hi - cov_xx, cov_xy) / Const.RAD;

        // Calculate excursion ratio.
        double cos_phi = -Math.cos(phi * Const.RAD), sin_phi = Math.sin(phi * Const.RAD),
               aspect_ratio = Math.sqrt(ew_hi / ew_lo),
               a2 = area * aspect_ratio / Math.PI,
               b2 = area / (aspect_ratio * Math.PI);
        double excursion = 0;
        m_excursion_image = new int[width][height];
        for (int x = 0; x < width; x++)
        {
            double x_cent = x - sum_x / area, cx = cos_phi * x_cent, sx = -sin_phi * x_cent;
            for (int y = 0; y < height; y++)
            {
                double y_cent = y - sum_y / area,
                       x_rot = (cx + sin_phi * y_cent),
                       y_rot = (sx + cos_phi * y_cent);
                boolean in_ellipse = (x_rot * x_rot / a2 + y_rot * y_rot / b2 <= 1);

                if (binary_image[x][y])
                {
                    if (in_ellipse)
                    {
                        m_excursion_image[x][y] = 2;
                    }
                    else
                    {
                        excursion++;
                        m_excursion_image[x][y] = 3;
                    }
                }
                else if (in_ellipse)
                {
                    m_excursion_image[x][y] = 1;
                }
            }
        }
        excursion /= area;

        FilamentSensor.debugMessage("Cell center: (" + mean_x + ", " + mean_y +")" +
                                    String.format("Covariance matrix: %n%.3f %.3f%n%.3f %.3f%n",
                                                  cov_xx, cov_xy, cov_xy, cov_yy) +
                                    String.format("Eigenvalues: %.3f %.3f%nAspect ratio: %.3f%nOrientation: %.3f%n",
                                                  ew_hi, ew_lo, Math.sqrt(ew_hi / ew_lo), phi));

        m_orientation = (long) (Const.M * phi);
        m_new_orientation = -1000 * Const.M;
        m_aspect_ratio = (long) (Const.M * Math.sqrt(ew_hi / ew_lo));
        m_center = new Point(mean_x, mean_y);
        m_order_parameter = -1000 * Const.M;
        m_new_order_parameter = -1000 * Const.M;
        m_excursion = (long) (Const.M * excursion);
        m_binary_image = binary_image;
        m_binary_image_tri = largestObject(Calc.close(larger_mask, Calc.circleMask(20)));
    }

    private PluginCellShape(long phi, Long new_phi, long aspect_ratio, Long order_parameter,
                            Long new_order_parameter, int area, Point center,
                            boolean[][] binary_image, boolean[][] binary_image_tri)
    {
        m_orientation = phi;
        m_new_orientation = (new_phi == null ? -1000 * Const.M : new_phi);
        m_aspect_ratio = aspect_ratio;
        m_order_parameter = (order_parameter == null ? -1000 * Const.M : order_parameter);
        m_new_order_parameter = (new_order_parameter == null ? -1000 * Const.M : new_order_parameter);
        m_area = area;
        m_center = (Point)center.clone();
        if (binary_image_tri != null)
        {
            final int width = binary_image_tri.length,
                      height = binary_image_tri[0].length;
            m_binary_image_tri = new boolean[width][height];
            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    m_binary_image_tri[x][y] = binary_image_tri[x][y];
                }
            }
        }
        if (binary_image != null)
        {
            final int width = binary_image.length,
                      height = binary_image[0].length;
            m_binary_image = new boolean[width][height];
            for (int x = 0; x < width; x++)
            {
                for (int y = 0; y < height; y++)
                {
                    m_binary_image[x][y] = binary_image[x][y];
                }
            }
        }
    }

    public static PluginCellShape fromXML(Element element)
    {
        for (String s : new String[]{"orientation", "aspect_ratio", "area", "center_x", "center_y"})
        {
            if (!element.hasAttribute(s))
            {
                return null;
            }
        }
        // FIXME: Redo area mask! Here, or in a dedicated function?
        return new PluginCellShape(getLong(element, "orientation"),
                                   getLong(element, "new_orientation"),
                                   getLong(element, "aspect_ratio"),
                                   getLong(element, "order_parameter"),
                                   getLong(element, "new_order_parameter"),
                                   getInt(element, "area"),
                                   new Point(getInt(element, "center_x"), 
                                             getInt(element, "center_y")), null, null);
    }

    public Element toXML(Document document)
    {
        Element element = document.createElement("cell_data");
        element.setAttribute("center_x", "" + getCenter().x);
        element.setAttribute("center_y", "" + getCenter().y);
        element.setAttribute("area", "" + getArea());
        element.setAttribute("aspect_ratio", "" + getAspectRatio());
        element.setAttribute("long_half_axis", "" + getLongHalfAxis());
        element.setAttribute("orientation", "" + getOrientation());
        element.setAttribute("order_parameter", "" + getOrderParameter());
        element.setAttribute("new_orientation", "" + getNewOrientation());
        element.setAttribute("new_order_parameter", "" + getNewOrderParameter());
        return element;
    }

    public static final PluginCellShape getCellData(int[][] image, boolean[][] mask, int m_orig_range, int min_range)
    {
        final int width = image.length,
                  height = image[0].length;
        if (m_orig_range < min_range)
        {
            return new PluginCellShape(new boolean[width][height], new boolean[width][height], null);
        }
        int[][] tmp = image;
        FilterGauss gauss1 = new FilterGauss(0.1); FilterGauss gauss2 = new FilterGauss(2.0);
        tmp = gauss2.apply(gauss1.apply(tmp)); // Strangely, this gives a 1.5 second speed up.

        int[][] tmp2 = new int[width][height];
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                tmp2[x][y] = tmp[x][y];
            }
        }

        FilterHistogramMethods li = new FilterHistogramMethods(tmp, true, FilterHistogramMethods.LI);
        while (tooWhite(li.binaryImage(), 0.5) && li.threshold() > 0)
        {
            new FilterHistogramMethods(tmp, li.threshold(), 255); // Rescales tmp!
            li = new FilterHistogramMethods(tmp, true, FilterHistogramMethods.LI);
        }

        FilterHistogramMethods triangle = new FilterHistogramMethods(tmp2, true, FilterHistogramMethods.TRIANGLE);
        while (tooWhite(triangle.binaryImage(), 0.5) && triangle.threshold() > 0)
        {
            new FilterHistogramMethods(tmp2, triangle.threshold(), 255); // Rescales tmp2!
            triangle = new FilterHistogramMethods(tmp2, true, FilterHistogramMethods.TRIANGLE);
        }
        return new PluginCellShape(li.binaryImage(), triangle.binaryImage(), mask);
    }

    private static Long getLong(Element element, String attribute)
    {
        if (element.hasAttribute(attribute))
        {
            return Long.parseLong(element.getAttribute(attribute));
        }
        return null;
    }

    private static Integer getInt(Element element, String attribute)
    {
        if (element.hasAttribute(attribute))
        {
            return Integer.parseInt(element.getAttribute(attribute));
        }
        return null;
    }

    public void calculateOrderParameter (int[][] orientation_field)
    {
        double sin_sum = 0.0,
               cos_sum = 0.0,
               old_numerator = 0.0,
               new_numerator = 0.0,
               pixel_count = 0.0;
        final int width = orientation_field.length,
                  height = orientation_field[0].length;

        final int[] histogram = new int[181];
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (orientation_field[x][y] > -1)
                {
                    histogram[180 - orientation_field[x][y]]++;
                    pixel_count += 1;
                }
            }
        }
        for (int phi = 0; phi < 181; phi++)
        {
            sin_sum += histogram[phi] * Math.sin(2.0 * phi * Const.RAD);
            cos_sum += histogram[phi] * Math.cos(2.0 * phi * Const.RAD);
        }
        double best_ori = (0.5 * Math.atan(sin_sum / cos_sum) / Const.RAD + 180) % 180;
        m_new_orientation = Math.round(best_ori * Const.M);
        for (int phi = 0; phi < 181; phi++)
        {
            old_numerator += histogram[phi] * Math.cos(2.0 * (phi - m_orientation / Const.MF) * Const.RAD);
            new_numerator += histogram[phi] * Math.cos(2.0 * (phi - best_ori) * Const.RAD);
        }
        m_order_parameter = Math.round(Const.MF * old_numerator / pixel_count);
        m_new_order_parameter = Math.round(Const.MF * new_numerator / pixel_count);
        if (m_new_order_parameter < 0)
        {
            m_new_order_parameter *= -1;
            best_ori = (best_ori + 90) % 180;
        }
        m_new_orientation = Math.round(best_ori * Const.M);
    }

    public void shutdown()
    {
        m_binary_image = null;
        m_binary_image_tri = null;
        m_excursion_image = null;
    }

    public Point getCenter()
    {
        return m_center;
    }

    public long getOrientation()
    {
        return m_orientation;
    }

    public long getNewOrientation()
    {
        return m_new_orientation;
    }

    public long getAspectRatio()
    {
        return m_aspect_ratio;
    }

    public long getOrderParameter()
    {
        return m_order_parameter;
    }

    public long getNewOrderParameter()
    {
        return m_new_order_parameter;
    }

    public long getExcursion()
    {
        return m_excursion;
    }

    public long getLongHalfAxis()
    {
        return (long)Math.sqrt(m_area * Const.MF * m_aspect_ratio / Math.PI);
    }

    public int getArea()
    {
        return m_area;
    }

    public boolean[][] getPaddedMask(int padding)
    {
        if (m_binary_image == null)
        {
            return null;
        }
        final int width = m_binary_image.length,
                  height = m_binary_image[0].length;
        if (padding > 0)
        {
            return Calc.dilate(m_binary_image, Calc.circleMask(padding));
        }
        else if (padding < 0)
        {
            return Calc.erode(m_binary_image, Calc.circleMask(-padding));
        }
        boolean[][] mask = new boolean[width][height];
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                mask[x][y] = m_binary_image[x][y];
            }
        }
        return mask;
    }

    public boolean[][] getExtendedMask()
    {
        if (m_binary_image_tri == null)
        {
            return null;
        }
        final int width = m_binary_image_tri.length,
                  height = m_binary_image_tri[0].length;
        boolean[][] mask = new boolean[width][height];
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                mask[x][y] = m_binary_image_tri[x][y];
            }
        }
        return mask;
    }

    public int[][] getExcursionImage()
    {
        return m_excursion_image;
    }

    /**
     * MAY CHANGE INPUT ARRAY!
     */
    private static boolean[][] largestObject(boolean[][] binary_image)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0)
        {
            System.err.println("FilamentSensor CellData.largestObject: " +
                               "Image must exist and have non-zero size.");
            return null;
        }

        final int width = binary_image.length,
                  height = binary_image[0].length;

        int[][] map = new int[width][height];
        Map<Integer,List<Point>> fragments = new HashMap<>();
        int index = 0;
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    if (x > 0 && binary_image[x-1][y])
                    {
                        int label = map[x][y] = map[x-1][y];
                        fragments.get(map[x][y]).add(new Point(x,y));
                        if (y > 0 && label != map[x][y-1] && fragments.containsKey(map[x][y-1]))
                        {
                            List<Point> to_dissolve = fragments.remove(map[x][y-1]);
                            for (Point p : to_dissolve)
                            {
                                map[p.x][p.y] = label;
                            }
                            fragments.get(map[x][y]).addAll(to_dissolve);
                        }
                    }
                    else if (y > 0 && binary_image[x][y-1])
                    {
                        map[x][y] = map[x][y-1];
                        fragments.get(map[x][y]).add(new Point(x,y));
                    }
                    else
                    {
                        fragments.put(++index, new ArrayList<Point>());
                        map[x][y] = index;
                        fragments.get(map[x][y]).add(new Point(x,y));
                    }
                }
            }
        }

        int largest_fragment_size = 0;
        for (List<Point> fragment : fragments.values())
        {
            if (largest_fragment_size < fragment.size())
            {
                largest_fragment_size = fragment.size();
            }
        }

        for (List<Point> fragment : fragments.values())
        {
            if (fragment.size() < largest_fragment_size)
            {
                for (Point p : fragment)
                {
                    binary_image[p.x][p.y] = false;
                }
            }
        }
        return binary_image;
    }

    private static boolean[][] clearEdges(boolean[][] binary_image)
    {
        if (binary_image == null || binary_image.length == 0 || binary_image[0].length == 0)
        {
            System.err.println("FilamentSensor CellData.clearEdges: " +
                               "Image must exist and have non-zero size.");
            return null;
        }

        final int width = binary_image.length,
                  height = binary_image[0].length;

        // Check upper and lower edge
        boolean[] upper_edge = new boolean[width],
                  lower_edge = new boolean[width];
        int upper_stain = 0,
            lower_stain = 0;
        for (int x = 0; x < width; x++)
        {
            upper_edge[x] = binary_image[x][0];
            upper_stain += (binary_image[x][0] ? 1 : 0);
            lower_edge[x] = binary_image[x][height-1];
            lower_stain += (binary_image[x][height-1] ? 1 : 0);
        }

        // Remove stains from upper and lower edge
        binary_image = clearOneEdge(binary_image, upper_edge, upper_stain, height, false, true);
        binary_image = clearOneEdge(binary_image, lower_edge, lower_stain, height, true, true);

        // Check left and right edge
        boolean[] left_edge = new boolean[height],
                  right_edge = new boolean[height];
        int left_stain = 0,
            right_stain = 0;
        for (int y = 0; y < height; y++)
        {
            left_edge[y] = binary_image[0][y];
            left_stain += (binary_image[0][y] ? 1 : 0);
            right_edge[y] = binary_image[width-1][y];
            right_stain += (binary_image[width-1][y] ? 1 : 0);
        }

        // Remove stains from left and right edge
        binary_image = clearOneEdge(binary_image, left_edge, left_stain, width, false, false);
        binary_image = clearOneEdge(binary_image, right_edge, right_stain, width, true, false);

        return binary_image;
    }

    private static boolean[][] clearOneEdge(boolean[][] binary_image, boolean[] edge, int stain,
                                            int size, boolean reverse, boolean row)
    {
        if (stain <= 0)
        {
            return binary_image;
        }

        final int length = edge.length;
        int[] sequence = new int[size];
        if (reverse)
        {
            for (int y = 0; y < size; y++)
            {
                sequence[y] = size - 1 - y;
            }
        }
        else
        {
            for (int y = 0; y < size; y++)
            {
                sequence[y] = y;
            }
        }

        if (row)
        {
            for (int y = 0; y < size-1; y++)
            {
                boolean[] new_edge = new boolean[length];
                int new_stain = 0;
                for (int x = 0; x < length; x++)
                {
                    if (edge[x] || (x > 0 && edge[x-1]) || (x > 1 && edge[x-2]) ||
                        (x < length-1 && edge[x+1]) || (x < length-2 && edge[x+2]))
                    {
                        new_edge[x] = binary_image[x][sequence[y+1]];
                        new_stain += (binary_image[x][sequence[y+1]] ? 1 : 0);
                    }
                }
                if (new_stain > stain + 1)
                {
                    break;
                }
                for (int x = 0; x < length; x++)
                {
                    if (edge[x])
                    {
                        binary_image[x][sequence[y]] = false;
                    }
                }
                stain = new_stain;
                edge = new_edge;
                if (stain <= 0)
                {
                    break;
                }
            }
        }
        else
        {
            for (int x = 0; x < size; x++)
            {
                boolean[] new_edge = new boolean[length];
                int new_stain = 0;
                for (int y = 0; y < length; y++)
                {
                    if (edge[y] || (y > 0 && edge[y-1]) || (y > 1 && edge[y-2]) ||
                        (y < length-1 && edge[y+1]) || (y < length-2 && edge[y+2]))
                    {
                        new_edge[y] = binary_image[sequence[x+1]][y];
                        new_stain += (binary_image[sequence[x+1]][y] ? 1 : 0);
                    }
                }
                if (new_stain > stain + 1)
                {
                    break;
                }
                for (int y = 0; y < length; y++)
                {
                    if (edge[y])
                    {
                        binary_image[sequence[x]][y] = false;
                    }
                }
                stain = new_stain;
                edge = new_edge;
                if (stain == 0)
                {
                    break;
                }
            }
        }
        return binary_image;
    }

    private static boolean tooWhite (boolean[][] binary_image, double ratio)
    {
        final int width = binary_image.length,
                  height = binary_image[0].length,
                  white_threshold = (int)(width * height * ratio),
                  black_threshold = width * height - white_threshold;
        int white_pixels = 0;
        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                if (binary_image[x][y])
                {
                    white_pixels++;
                }
            }
            if (white_pixels > white_threshold)
            {
                return true;
            }
            if (x * height - white_pixels > black_threshold)
            {
                return false;
            }
        }
        return false;
    }

    @Override
    public Object clone()
    {
        return new PluginCellShape(m_orientation, m_new_orientation, m_aspect_ratio,
                                   m_order_parameter, m_new_order_parameter, m_area, m_center,
                                   m_binary_image, m_binary_image_tri);
    }

}
