package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2015 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.FileDialog;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileFilter;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;

import javax.swing.JOptionPane;
import javax.swing.filechooser.FileNameExtensionFilter;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import core.DataImage;
import core.DataParameters;
import core.DataParameters.Any;
import core.DataProject;

/**
 * This class manages loading and saving of project data. The constructor
 * calls the appropriate loader method while the save method must be called
 * for an existing object. The class is instantiated by the ThumbnailView.
 */
public class IOProjectData extends DataProject
{

    private Document m_xml_source_document;
    private DocumentBuilder m_document_builder;
    private Element m_xml_root_name;

    private String m_path_separator;

    private File m_project_file;

    private boolean m_valid;

    /**
     * The constructor launches a dialog to open an existing project file or to
     * create a new project from a folder (which should contain image files).
     *
     * @param load
     *            If true, the load dialog for an existing folder is launched,
     *            otherwise the dialog for a new project is launched.
     */
    public IOProjectData(boolean load)
    {
        m_processed_images = new ArrayList<DataImage>();
        m_path_separator = System.getProperty("file.separator");

        // These values may be overwritten during loading.
        m_default_parameters = new DataParameters();
        m_project_file = null;
        m_files = new ArrayList<Files>();
        m_force_defaults = false;

        try
        {
            DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
            m_document_builder = docFactory.newDocumentBuilder();
        }
        catch (ParserConfigurationException e)
        {
            System.err.println("FilamentSensor IOProjectData: " +
            		           "Failed to create DocumentBuilder: " + e.getMessage());
            m_document_builder = null;
        }

        if (load)
        {
            m_valid = loadProject();
            if (!m_valid)
            {
                FilamentSensor.debugError("Failed loading project. Image list remains empty.");
            }
        }
        else
        {
            m_valid = newProject();
            if (!m_valid)
            {
                FilamentSensor.debugError("Failed creating new project. Image list remains empty.");
            }
        }
    }

    public IOProjectData(File project_file, File image_folder, File data_folder)
    {
        m_processed_images = new ArrayList<DataImage>();
        m_path_separator = System.getProperty("file.separator");

        // These values may be overwritten during loading.
        m_default_parameters = new DataParameters();
        m_project_file = null;
        m_files = new ArrayList<Files>();
        m_force_defaults = false;

        try
        {
            DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
            m_document_builder = docFactory.newDocumentBuilder();
        }
        catch (ParserConfigurationException e)
        {
            System.err.println("FilamentSensor IOProjectData: " +
                               "Failed to create DocumentBuilder: " + e.getMessage());
            m_document_builder = null;
        }

        m_valid = quickLoadProject(project_file, image_folder, data_folder);
        if (!m_valid)
        {
            FilamentSensor.debugError("Failed loading project. Image list remains empty.");
        }
    }

    public IOProjectData(File image_folder)
    {
        m_processed_images = new ArrayList<>();
        m_path_separator = System.getProperty("file.separator");

        // These values may be overwritten during loading.
        m_default_parameters = new DataParameters();
        m_project_file = null;
        m_force_defaults = false;

        if (image_folder == null)
        {
            return;
        }

        FileFilter image_filter = FileExtensionFilter.image();

        List<File> image_list = new ArrayList<File>(Arrays.asList(image_folder.listFiles(image_filter)));

        if (image_list == null || image_list.isEmpty())
        {
            return;
        }

        List<Files> result_list = new ArrayList<Files>();

        for (File image_file : image_list)
        {
            result_list.add(new Files(image_file, null));
        }

        try
        {
            DocumentBuilderFactory docFactory = DocumentBuilderFactory.newInstance();
            m_document_builder = docFactory.newDocumentBuilder();
        }
        catch (ParserConfigurationException e)
        {
            System.err.println("FilamentSensor IOProjectData: " +
                               "Failed to create DocumentBuilder: " + e.getMessage());
            m_document_builder = null;
        }

        m_files = result_list;
        Collections.sort(m_files);
    }

    public boolean contains(File file)
    {
        for (Files pair : m_files)
        {
            if (file.equals(pair.image()) || file.equals(pair.data()))
            {
                return true;
            }
        }
        return false;
    }

    public int getLength()
    {
        return m_files.size();
    }

    public File getImageFile(int number)
    {
        return m_files.get(number).image();
    }

    public File getDataFile(int number)
    {
        return m_files.get(number).data();
    }

    public File getDataForImage(File image_file)
    {
        for (Files pair : m_files)
        {
            if (image_file.equals(pair.image()))
            {
                return pair.data();
            }
        }

        return null;
    }

    private boolean setDataForImage(File image_file, File data_file)
    {
        for (Files pair : m_files)
        {
            if (image_file.equals(pair.image()))
            {
                pair.setData(data_file);
                return true;
            }
        }
        return false;
    }

    public String getName()
    {
        return (m_project_file == null ? "New Project" : m_project_file.getName().replaceAll("\\.xml$", ""));
    }

    public File[] getLocations()
    {
        return new File[]{m_project_file, getImagePath(), getDataPath(),
                          Dialogs.save(true, "Folder to save images with filaments."),
                          Dialogs.save(true, "Folder to save csv.")};
    }

    public File getImagePath()
    {
        if (m_files == null || m_files.isEmpty())
        {
            return null;
        }
        return m_files.get(0).image().getParentFile();
    }

    public File getDataPath()
    {
        return m_project_file.getParentFile();
    }

    public boolean valid()
    {
        return m_valid;
    }

    /**
     * THIS METHOD CHANGES THE LISTS PASSED TO IT!
     */
    private void loadDataFiles(NodeList element_list, File image_folder, File data_folder,
                               List<Files> result_list, List<File> result_list_images,
                               List<File> result_list_data)
    {
        for (Node n = element_list.item(0).getFirstChild(); n != null; n = n.getNextSibling())
        {
            try
            {
                if (n.getNodeType() != Node.ELEMENT_NODE)
                {
                    continue;
                }

                Element element = (Element) n;
                File image_file = new File(image_folder.getCanonicalPath() + File.separator +
                                           element.getAttribute("image"));

                if (!image_file.exists() || !image_file.canRead() || !image_file.isFile())
                {
                    FilamentSensor.debugMessage("Image " + image_file + " not found. Try again!");
                    image_file = Dialogs.load(false, image_folder, "No image found! Determine image file for " + element.getAttribute("data"));
                    if (!image_file.exists() || !image_file.canRead() || !image_file.isFile())
                    {
                        FilamentSensor.debugMessage("Image " + image_file + " not found.");
                        continue;
                    }
                }

                try
                {
                    image_file = image_file.getCanonicalFile();
                }
                catch (IOException e)
                {
                    FilamentSensor.debugMessage("Converting '" + image_file.getPath() + "' to canonical path failed:\n" + e.getMessage());
                }

                File data_file = null;

                if (data_folder != null)
                {
                    data_file = (element.getAttribute("data").equals("") ? null :
                                 new File(data_folder.getCanonicalPath() + File.separator +
                                          element.getAttribute("data")));
                }

                result_list.add(new Files(image_file, data_file));
                result_list_images.add(image_file);
                if (data_file != null)
                {
                    result_list_data.add(data_file);
                }
            }
            catch (Exception e)
            {
                FilamentSensor.debugMessage("Error parsing Filament '" + n + "': " + e.getMessage());
            }
        }
    }

    /**
     * Check whether all data files in the data folder are listed in the
     * project file. For excess data files, try to find their image file
     * and add them. Otherwise discard them.
     *
     * THIS METHOD CHANGES THE LISTS PASSED TO IT!
     */
    private void checkDataFiles(File data_folder, File image_folder, List<Files> result_list,
                                List<File> result_list_images, List<File> result_list_data)
    {
        FileFilter xml_filter = FileExtensionFilter.xml();

        List<File> data_list = (data_folder == null ? new ArrayList<File>() : new ArrayList<File>(
                Arrays.asList(data_folder.listFiles(xml_filter))));
        if (data_list.contains(m_project_file))
        {
            FilamentSensor.debugMessage("Project file found in data folder...");
            data_list.remove(m_project_file);
        }

        for (File data_file : data_list)
        {
            if (!result_list_data.contains(data_file))
            {
                IOImageData image_data = new IOImageData(data_file);
                File image_file = image_data.getImage();

                if (image_file == null || !image_file.getParentFile().equals(image_folder))
                {
                    FilamentSensor.debugError("Data file specifies no or invalid image.");

                    image_file = Dialogs.load(false, image_folder,
                                              "Choose image file for " + data_file.toString());
                    if (image_file == null || !image_file.exists() ||
                        !image_file.canRead() || !image_file.isFile())
                    {
                        FilamentSensor.debugError("No image file given. Skipping data file.");
                        continue;
                    }

                    try
                    {
                        image_file = image_file.getCanonicalFile();
                    }
                    catch (IOException e)
                    {
                        FilamentSensor.debugMessage("Converting '" + image_file.getPath() +
                                                    "' to canonical path failed:\n" + e.getMessage());
                    }
                }

                result_list.add(new Files(image_file, data_file));
                result_list_images.add(image_file);
            }
        }
    }

    /**
     * Check whether all data files in the data folder are listed in the
     * project file. Discard excess data files.
     *
     * THIS METHOD CHANGES THE LISTS PASSED TO IT!
     */
    private void discardDataFiles(File data_folder, File image_folder, List<Files> result_list,
                                  List<File> result_list_images, List<File> result_list_data)
    {
        FileFilter xml_filter = FileExtensionFilter.xml();

        List<File> data_list = (data_folder == null ? new ArrayList<File>() : new ArrayList<File>(
                Arrays.asList(data_folder.listFiles(xml_filter))));
        if (data_list.contains(m_project_file))
        {
            FilamentSensor.debugMessage("Project file found in data folder...");
            data_list.remove(m_project_file);
        }

        for (File data_file : data_list)
        {
            if (!result_list_data.contains(data_file))
            {
                IOImageData image_data = new IOImageData(data_file);
                File image_file = image_data.getImage();

                if (image_file == null || !image_file.getParentFile().equals(image_folder) ||
                    !image_file.exists() || !image_file.canRead() || !image_file.isFile())
                {
                    FilamentSensor.debugError("No image file given. Skipping data file: " + data_file);
                    continue;
                }

                try
                {
                    image_file = image_file.getCanonicalFile();
                }
                catch (IOException e)
                {
                    FilamentSensor.debugMessage("Converting '" + image_file.getPath() +
                                                "' to canonical path failed:\n" + e.getMessage());
                }

                result_list.add(new Files(image_file, data_file));
                result_list_images.add(image_file);
            }
        }
    }

    private boolean quickLoadProject(File project_file, File image_folder, File data_folder)
    {
        m_project_file = project_file;

        if (m_project_file == null || !m_project_file.exists() || !m_project_file.canRead())
        {
            FilamentSensor.debugError("Specified project file is invalid: " + m_project_file);
            return false;
        }
        if (image_folder == null || !image_folder.exists() || !image_folder.canRead() || !image_folder.isDirectory())
        {
            FilamentSensor.debugError("Specified image folder is invalid:" + image_folder);
            return false;
        }

        if (data_folder == null || !data_folder.exists() || !data_folder.canRead() || !data_folder.isDirectory())
        {
            FilamentSensor.debugError("Specified data folder is invalid: " + data_folder);
            data_folder = null;
        }

        if (m_document_builder == null)
        {
            FilamentSensor.debugError("Fatal: Can't open, docBuilder == null!");
            return false;
        }

        try
        {
            m_xml_source_document = m_document_builder.parse(m_project_file);
            m_xml_root_name = m_xml_source_document.getDocumentElement();
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            FilamentSensor.debugError("Could not read '" + m_project_file + "':\n" + e.getMessage());
            return false;
        }

        NodeList element_list = m_xml_source_document.getElementsByTagName("FilamentSensorProjectDataFile");
        if (element_list.getLength() == 0)
        {
            element_list = m_xml_source_document.getElementsByTagName("FilTrackerProjectDataFile");
            if (element_list.getLength() == 0)
            {
                System.err.println("The specified file is valid XML, but does not seem to be " +
                                   "a valid FilamentSensor project file...");
                return false;
            }
        }

        Element element;

        // Get paths and do a ton of integrity checks.
        element_list = m_xml_source_document.getElementsByTagName("paths");

        if (!(element_list.getLength() == 0))
        {
            element = (Element) element_list.item(0);
            m_force_defaults = Boolean.parseBoolean(element.getAttribute("force_defaults"));
        }

        m_default_parameters = DataParameters.fromXML(m_xml_source_document, new DataParameters());

        List<Files> result_list = new ArrayList<Files>();
        List<File> result_list_images = new ArrayList<File>();
        List<File> result_list_data = new ArrayList<File>();

        // Read file pairs from project file.
        element_list = m_xml_source_document.getElementsByTagName("file_pairs");

        if (element_list.getLength() != 0)
        {
            loadDataFiles(element_list, image_folder, data_folder, result_list, result_list_images, result_list_data);
        }

        if (data_folder != null)
        {
            discardDataFiles(data_folder, image_folder, result_list, result_list_images, result_list_data);
        }

        // Check whether all image files in the image folder are listed
        // in the project file. List excess image files without data files.
        FileFilter image_filter = FileExtensionFilter.image();

        List<File> image_list = new ArrayList<File>(Arrays.asList(image_folder.listFiles(image_filter)));

        if (image_list != null)
        {
            for (File image_file : image_list)
            {
                if (!result_list_images.contains(image_file))
                {
                    result_list.add(new Files(image_file, null));
                }
            }
        }

        m_files = result_list;
        Collections.sort(m_files);
        m_processed_images.clear();

        return true;
    }

    private boolean loadProject()
    {
        m_project_file = Dialogs.load(false, "Load FilamentSensor project");

        if (m_project_file == null)
        {
            return false;
        }

        if (m_document_builder == null)
        {
            FilamentSensor.debugError("Fatal: Can't open, docBuilder == null!");
            return false;
        }

        try
        {
            m_xml_source_document = m_document_builder.parse(m_project_file);
            m_xml_root_name = m_xml_source_document.getDocumentElement();
        }
        catch (Exception e) // Catch generic Exception to avoid redundant code
        { // Can be: IO-, SAX- (parse error) or IllegalArgumentException
            FilamentSensor.debugError("Could not read '" + m_project_file + "':\n" + e.getMessage());
            return false;
        }

        NodeList element_list = m_xml_source_document.getElementsByTagName("FilamentSensorProjectDataFile");
        if (element_list.getLength() == 0)
        {
            element_list = m_xml_source_document.getElementsByTagName("FilTrackerProjectDataFile");
            if (element_list.getLength() == 0)
            {
                System.err.println("The specified file is valid XML, but does not seem to be " +
                                   "a valid FilamentSensor project file...");
                return false;
            }
        }

        Element element;

        // Get paths and do a ton of integrity checks.
        element_list = m_xml_source_document.getElementsByTagName("paths");
        File image_folder = null;
        File data_folder = null;

        if (element_list.getLength() == 0)
        {
            FilamentSensor.debugError("Project file does not specify image folder.");

            image_folder = Dialogs.load(true, "Locate image files");

            if (image_folder == null)
            {
                return false;
            }
        }
        else
        {
            element = (Element) element_list.item(0);
            image_folder = new File(element.getAttribute("image_path"));
            data_folder = new File(element.getAttribute("data_path"));
            m_force_defaults = Boolean.parseBoolean(element.getAttribute("force_defaults"));
        }

        while (!image_folder.exists() || !image_folder.canRead() || !image_folder.isDirectory())
        {
            FilamentSensor.debugError("Specified image folder is invalid.");

            image_folder = Dialogs.load(true, image_folder, "Locate image files");

            if (image_folder == null)
            {
                return false;
            }
        }

        if (!data_folder.equals(m_project_file.getParentFile()))
        {
            FilamentSensor.debugError("Project file specifies data path\n'" + data_folder + "'\n" +
                                      "but is itself located at\n'" + m_project_file.getParent() + "'.");
            data_folder = Dialogs.load(true, data_folder, "Locate data files");
        }

        if (data_folder != null)
        {
            while (!data_folder.exists() || !data_folder.canRead() || !data_folder.isDirectory())
            {
                FilamentSensor.debugError("Specified data folder is invalid.");

                data_folder = Dialogs.load(true, data_folder, "Locate data files");

                if (data_folder == null)
                {
                    break;
                }
            }
        }

        m_default_parameters = DataParameters.fromXML(m_xml_source_document, new DataParameters());

        List<Files> result_list = new ArrayList<Files>();
        List<File> result_list_images = new ArrayList<File>();
        List<File> result_list_data = new ArrayList<File>();

        // Read file pairs from project file.
        element_list = m_xml_source_document.getElementsByTagName("file_pairs");

        if (element_list.getLength() != 0)
        {
            loadDataFiles(element_list, image_folder, data_folder, result_list, result_list_images, result_list_data);
            // Check whether at least one image file was found
            while (result_list_images.isEmpty())
            {
                image_folder = Dialogs.load(true, image_folder, "No image found! Locate images");
                loadDataFiles(element_list, image_folder, data_folder, result_list, result_list_images, result_list_data);
            }
        }

        if (data_folder != null)
        {
            checkDataFiles(data_folder, image_folder, result_list, result_list_images, result_list_data);
        }

        // Check whether all image files in the image folder are listed
        // in the project file. List excess image files without data files.
        FileFilter image_filter = FileExtensionFilter.image();

        List<File> image_list = new ArrayList<File>(Arrays.asList(image_folder.listFiles(image_filter)));

        if (image_list != null)
        {
            for (File image_file : image_list)
            {
                if (!result_list_images.contains(image_file))
                {
                    result_list.add(new Files(image_file, null));
                }
            }
        }

        m_files = result_list;
        Collections.sort(m_files);
        m_processed_images.clear();

        return true;
    }

    private boolean newProject()
    {
        File image_folder = Dialogs.load(true, "Determine image folder");

        if (image_folder == null)
        {
            return false;
        }
        FilamentSensor.debugMessage("Path: " + image_folder.getAbsolutePath());

        while (!image_folder.exists() || !image_folder.canRead() || !image_folder.isDirectory())
        {
            FilamentSensor.debugError("Specified image folder is invalid.");

            image_folder = Dialogs.load(true, image_folder, "Determine image folder");

            if (image_folder == null)
            {
                return false;
            }
        }

        FileFilter image_filter = FileExtensionFilter.image();

        List<File> image_list = new ArrayList<File>(Arrays.asList(image_folder.listFiles(image_filter)));

        if (image_list == null || image_list.isEmpty())
        {
            return false;
        }

        List<Files> result_list = new ArrayList<Files>();
        List<File> result_list_images = new ArrayList<File>();

        // Option to import data files.
        if (JOptionPane.showOptionDialog(MainWindow.frame(), "Do you want to import data files "
                + "related to the images?", "Browse for data folder?", JOptionPane.YES_NO_OPTION,
                JOptionPane.QUESTION_MESSAGE, null, null, null) == JOptionPane.YES_OPTION)
        {
            File data_folder = Dialogs.load(true, image_folder, "Locate data folder");

            if (data_folder != null)
            {
                while (!data_folder.exists() || !data_folder.canRead() || !data_folder.isDirectory())
                {
                    FilamentSensor.debugError("Specified data folder is invalid.");

                    data_folder = Dialogs.load(true, data_folder, "Locate data folder");

                    if (data_folder == null)
                    {
                        break;
                    }
                }
            }

            if (data_folder != null)
            {
                m_project_file = data_folder;

                FileFilter xml_filter = FileExtensionFilter.xml();

                List<File> data_list = (data_folder == null ? new ArrayList<File>() : new ArrayList<File>(
                        Arrays.asList(data_folder.listFiles(xml_filter))));

                for (File data_file : data_list)
                {
                    FilamentSensor.debugMessage("Processing data file " + data_file.getAbsolutePath());
                    IOImageData image_data = new IOImageData(data_file);
                    File image_file = image_data.getImage();

                    if (image_file == null || !image_file.getParentFile().equals(image_folder))
                    {
                        FilamentSensor.debugError("Data file " + data_file.getAbsolutePath() +
                                                  " specifies no or invalid image.");

                        image_file = Dialogs.load(false, image_folder,
                                "Choose image file for " + data_file.toString());
                        if (image_file == null)
                        {
                            FilamentSensor.debugError("No image file given. Skipping data file.");
                        }
                    }

                    if (image_file == null)
                    {
                        continue;
                    }

                    FilamentSensor.debugMessage("Adding: image = " + image_file.getAbsolutePath() +
                                                ", data = " + data_file.getAbsolutePath());
                    result_list.add(new Files(image_file, data_file));
                    result_list_images.add(image_file);
                }
            }
        }

        for (File image_file : image_list)
        {
            if (!result_list_images.contains(image_file))
            {
                result_list.add(new Files(image_file, null));
            }
        }

        m_files = result_list;
        Collections.sort(m_files);
        m_processed_images.clear();

        return true;
    }

    private boolean saveImageData()
    {
        boolean no_errors = true;
        for (DataImage data : m_processed_images)
        {
            no_errors = no_errors && saveImage(data);
        }

        m_processed_images.clear();

        return no_errors;
    }

    private boolean saveImage(DataImage data)
    {
        String path = m_project_file.getParent();
        String name = data.getImageName();
        File image_file = data.getFile();

        if (name == "")
        {
            FilamentSensor.debugError("DataImage object lacks image file name.");

            FileDialog fd = new FileDialog(MainWindow.frame(), "Save image data file", FileDialog.SAVE);
            fd.setDirectory(path);
            fd.setVisible(true);

            if (fd.getFile() == null)
            {
                FilamentSensor.debugError("Invalid file name.");
                return false;
            }

            path = fd.getDirectory() + File.separator + fd.getFile();
        }
        else
        {
            name = name.replaceAll("\\.", "_") + ".xml";
            path += m_path_separator + name;
        }

        File data_file = new File(path);
        IOImageData save_file = new IOImageData(data_file);

        try
        {
            if (!save_file.writeImageDataFile(data))
            {
                FilamentSensor.debugError("Saving file '" + name + "' to '" + path + "' failed.");
                return false;
            }
        }
        catch (IOException ex)
        {
            JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + data.getImageName(),
                                          "Image could not be saved!", JOptionPane.WARNING_MESSAGE, null);
            return false;
        }

        if (!setDataForImage(image_file, data_file))
        {
            FilamentSensor.debugError("The image file " + image_file.getAbsolutePath() +
                                      " listed by the data object is not part of the project!");
            return false;
        }
        return true;
    }

    /**
     * Launches a dialog to determine a folder and file name, to which project
     * data is saved in the XML format. Project data consists of the path of the
     * image files, the path of the data files (which is the path chosen in the
     * dialog, to which the project XML file is saved) and the names of all
     * image and data files.
     *
     * Additionally, data files for all images that were processed are saved to
     * the chosen target folder.
     *
     * @return A boolean signaling success of the operation.
     */
    public boolean saveProject()
    {
        boolean no_errors = true;
        m_project_file = Dialogs.save(false, "Save FilamentSensor project");

        if (m_document_builder == null)
        {
            FilamentSensor.debugError("Fatal: Can't save, docBuilder == null!");
            return false;
        }

        if (!saveImageData())
        {
            FilamentSensor.debugError("Failed saving image data.");
            no_errors = false;
        }

        if (m_project_file.exists() && JOptionPane.showOptionDialog(MainWindow.frame(),
                "Warning! You might be overwriting an image data file with the project file!\n"
                + "(If you derived the project file name from an image file name.)\n"
                + "This would result in loss of data! Continue anyway?","File name collision",
                JOptionPane.YES_NO_OPTION, JOptionPane.WARNING_MESSAGE, null, null, null)
                == JOptionPane.NO_OPTION)
        {
            m_project_file = Dialogs.save(false, "Save FilamentSensor project");
        }

        return writeData() && no_errors;
    }

    /**
     * Saves project data in the XML format to the given File. Project data
     * consists of the path of the image files, the path of the data files
     * (which is the path given as an argument, to which the project XML file
     * is saved) and the names of all image and data files.
     *
     * Additionally, data files for all images that were processed are saved to
     * the chosen target folder.
     *
     * @return A boolean signaling success of the operation.
     */
    public boolean saveProject(File project_file)
    {
        m_project_file = project_file;

        if (m_document_builder == null)
        {
            System.err.println("Fatal: Can't save, docBuilder == null!");
            return false;
        }

        if (m_files == null || m_files.isEmpty())
        {
            System.err.println("Fatal: Can't save, no files associated to project!");
            return false;
        }

        if (!saveImageData())
        {
            System.err.println("Failed saving image data.");
            return false;
        }

        return writeData();
    }

    public boolean overwrite()
    {
        return saveProject(m_project_file);
    }

    private boolean writeData()
    {
        m_xml_source_document = m_document_builder.newDocument();
        m_xml_root_name = m_xml_source_document.createElement("FilamentSensorProjectDataFile");
        m_xml_source_document.appendChild(m_xml_root_name);

        // Version information
        Element element = m_xml_source_document.createElement("program");
        m_xml_root_name.appendChild(element);
        element.setAttribute("name", FilamentSensor.NAME);
        element.setAttribute("version", FilamentSensor.VERSION);

        // Paths of image and data files.
        element = m_xml_source_document.createElement("paths");
        m_xml_root_name.appendChild(element);
        element.setAttribute("image_path", m_files.get(0).image().getParent());
        element.setAttribute("data_path", m_project_file.getParent());
        element.setAttribute("force_defaults", "" + m_force_defaults);

        // Parameters
        if (m_default_parameters != null)
        {
            for (int type : DataParameters.ALL)
            {
                element = m_xml_source_document.createElement(DataParameters.getName(type));
                m_xml_root_name.appendChild(element);
                for (Any key : m_default_parameters.keySet(type))
                {
                    element.setAttribute(key.name(), "" + m_default_parameters.getParameter(key));
                }
            }
        }

        // Files
        Element pairs = m_xml_source_document.createElement("file_pairs");
        m_xml_root_name.appendChild(pairs);

        for (int i = 0; i < m_files.size(); i++)
        {
            Files pair = m_files.get(i);
            element = m_xml_source_document.createElement("files" + i);
            pairs.appendChild(element);

            element.setAttribute("image", pair.image().getName());
            element.setAttribute("data", (pair.data() == null ? "" : pair.data().getName()));
        }

        try
        {
            // write the created document object to file
            Transformer transformer = TransformerFactory.newInstance().newTransformer();
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");

            BufferedWriter writer = new BufferedWriter(new FileWriter(m_project_file));
            StreamResult result = new StreamResult(writer);
            DOMSource source = new DOMSource(m_xml_source_document);
            transformer.transform(source, result);
            writer.close();
        }
        catch (Exception e) // IOException or
                            // TransformerConfigurationException
        {
            FilamentSensor.debugError("Error while saving '" + m_project_file + "': " + e.getMessage());
            return false;
        }

        return Config.saveToDatabase(m_project_file.getAbsolutePath(), Config.PROJECT);
    }

    /**
     * This class wraps a FileNameExtensionFilter for use as a
     * java.io.FileFilter.
     */
    private static class FileExtensionFilter implements FileFilter
    {
        private FileNameExtensionFilter filter;

        private FileExtensionFilter(String description, String... extensions)
        {
            filter = new FileNameExtensionFilter(description, extensions);
        }

        public static FileExtensionFilter image()
        {
            return new FileExtensionFilter("Image File", "png", "bmp", "jpg", "jpeg", "gif", "tif", "tiff");
        }

        public static FileExtensionFilter xml()
        {
            return new FileExtensionFilter("XML File", "xml");
        }

        @Override
        public boolean accept(File file)
        {
            return (filter.accept(file) && !(file.isDirectory()));
        }
    }

}
