package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.Graphics2D;
import java.awt.image.BufferedImage;
import java.awt.image.ColorModel;
import java.awt.image.DataBuffer;
import java.awt.image.DataBufferInt;
import java.awt.image.Raster;
import java.awt.image.WritableRaster;

import core.Misc.Int2D;
import core.Misc.Pair;

public abstract class ImageFactory
{

    /**
     * Converts the array of pixels to BufferedImage.
     */
    public static BufferedImage makeBinaryImage(boolean[][] binary_image, boolean invert)
    {
        if (binary_image == null || binary_image.length <= 0 || binary_image[0].length <= 0)
        {
            return null;
        }

        int width = binary_image.length;
        int height = binary_image[0].length;
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        WritableRaster raster = image.getRaster();

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                raster.setSample(i, j, 0, (binary_image[i][j] == invert ? 0 : 255));
            }
        }

        return image;
    }

    public static BufferedImage makeTwoLevelRedGreenImage(int[][] int_image,
                                                          boolean[][] mask_in,
                                                          boolean[][] mask_out)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0 ||
            mask_in == null || mask_in.length != int_image.length || mask_in[0].length != int_image[0].length ||
            mask_out == null || mask_out.length != int_image.length || mask_out[0].length != int_image[0].length)
        {
            return null;
        }

        final int width = int_image.length,
                  height = int_image[0].length;
        int[] out = new int[width * height];

        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                int tmp = 20 + (int)Math.round(235. * int_image[x][y] / 255.);
                out[width * y + x] = 0xFF000000 | (mask_out[x][y] ? tmp << (mask_in[x][y] ? 8 : 16) :
                                                   (tmp << 16 | tmp << 8 | tmp));
            }
        }

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(out, width * height);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        BufferedImage image = new BufferedImage(cm, raster, false, null);
        return image;
    }
    
    public static BufferedImage makeGreyImage(int[][] int_image, boolean filter)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            return null;
        }

        int width = int_image.length;
        int height = int_image[0].length;
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        WritableRaster raster = image.getRaster();

        if (filter)
        {
            for (int i = 0; i < width; i++)
            {
                for (int j = 0; j < height; j++)
                {
                    int pixel = int_image[i][j] + 128;
                    raster.setSample(i, j, 0, (pixel < 0 ? 0 :
                                              (pixel > 255 ? 255 :
                                                  pixel)));
                }
            }
            return image;
        }

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                raster.setSample(i, j, 0, (int_image[i][j] < 0 ? 0 :
                                          (int_image[i][j] > 255 ? 255 :
                                               int_image[i][j])));
            }
        }
        return image;
    }

    public static BufferedImage makeDeepGreyImage(int[][] int_image)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            return null;
        }

        final int width = int_image.length,
                  height = int_image[0].length,
                  white = (1 << 16) - 1;
        int max = 0;
        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                if (int_image[i][j] > max)
                {
                    max = int_image[i][j];
                }
            }
        }
        if (max < white)
        {
            max = white;
        }
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_USHORT_GRAY);
        WritableRaster raster = image.getRaster();

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                raster.setSample(i, j, 0, (long)int_image[i][j] * white / max);
            }
        }
        return image;
    }

    public static BufferedImage makeColorImage(int[][] int_image, int color_count, boolean scale)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            return null;
        }

        final int width = int_image.length,
                  height = int_image[0].length,
                  padding = 20;
        int[] out = new int[width * height],
              colors = makeColors(color_count);

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[width * j + i] = (int_image[i][j] < 0 ? 0xFF000000 :
                                     (int_image[i][j] >= color_count ? 0xFFFFFFFF :
                                         colors[int_image[i][j]]));
            }
        }

        int end = padding, start = color_count + padding;
        if (scale)
        {
            for (int i = padding; i < 2 * padding; i++)
            {
                for (int j = start; j > end; j--)
                {
                    out[width * j + i] = colors[j - end - 1];
                }
            }
        }

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(out, width * height);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        BufferedImage image = new BufferedImage(cm, raster, false, null);
        if (scale)
        {
            Graphics2D canvas = image.createGraphics();
            String [] labels = new String[]{"0°", "90°", "180°"};
            start += (color_count + canvas.getFontMetrics().getHeight()) / 2;
            for(String l : labels){
                canvas.drawString(l, 5 * padding / 2, start -= color_count / 2);
            }
            /* Right-aligned text
            start += (color_count + canvas.getFontMetrics().getHeight()) / 2;
            for(String l : labels){
                int label_width = canvas.getFontMetrics().stringWidth(l);
                int x = width - label_width - 5 * padding / 2;
                canvas.drawString(l, x, start -= color_count / 2);
            }*/
            canvas.dispose();
        }
        
        return image;
    }

    public static BufferedImage makeThreeColorImage(int[][] int_image)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0)
        {
            return null;
        }

        final int width = int_image.length,
                  height = int_image[0].length;
        int[] out = new int[width * height],
              colors = new int[]{0xFF000000, 0xFF0000AA, 0xFFFF6600, 0xFFFF0000};

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[width * j + i] = colors[int_image[i][j]];
            }
        }

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(out, width * height);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        BufferedImage image = new BufferedImage(cm, raster, false, null);
        return image;
    }

    public static BufferedImage makeRedGreenImage(int[][] int_image, boolean[][] mask)
    {
        if (int_image == null || int_image.length <= 0 || int_image[0].length <= 0 ||
            mask == null || mask.length != int_image.length || mask[0].length != int_image[0].length)
        {
            return null;
        }

        final int width = int_image.length,
                  height = int_image[0].length;
        int[] out = new int[width * height];

        for (int x = 0; x < width; x++)
        {
            for (int y = 0; y < height; y++)
            {
                int tmp = 20 + (int)Math.round(235. * int_image[x][y] / 255.);
                out[width * y + x] = 0xFF000000 | tmp << (mask[x][y] ? 8 : 16);
            }
        }

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(out, width * height);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        BufferedImage image = new BufferedImage(cm, raster, false, null);
        return image;
    }

    public static BufferedImage makeWidthMap(Pair<Int2D, Integer> pair)
    {
        int[][] width_map = pair.a.the;
        final int color_count = pair.b - 1,
                  width = width_map.length,
                  height = width_map[0].length,
                  offset = 5 * (color_count + 1) / 3 + 1;
        int[] out = new int[width * height],
              colors = makeColors(color_count);

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                out[width * j + i] = (width_map[i][j] == 0 ? 0xFF000000 :
                                     (width_map[i][j] == 1 ||
                                      width_map[i][j] > color_count + 1 ? 0xFFFFFFFF :
                                         colors[((offset - width_map[i][j]) % color_count)]));
            }
        }

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(out, width * height);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        return new BufferedImage(cm, raster, false, null);
    }

    public static BufferedImage greyWidthMap(Pair<Int2D, Integer> pair)
    {
        int[][] width_map = pair.a.the;
        final int color_count = pair.b,
                  width = width_map.length,
                  height = width_map[0].length;
        int[] grey = makeGreyLevels(color_count);
        BufferedImage image = new BufferedImage(width, height, BufferedImage.TYPE_BYTE_GRAY);
        WritableRaster raster = image.getRaster();

        for (int i = 0; i < width; i++)
        {
            for (int j = 0; j < height; j++)
            {
                raster.setSample(i, j, 0, (width_map[i][j] == 0 ? 0 :
                                          (width_map[i][j] > color_count ? 255 :
                                           grey[width_map[i][j] - 1])));
            }
        }
        return image;
    }

    public static BufferedImage makeColorImage(int[][] int_image, boolean scale)
    {
        return makeColorImage(int_image, 180, scale);
    }

    public static BufferedImage makeCustomImage(int[] int_image, final int width)
    {
        if (int_image == null || int_image.length <= 0)
        {
            return null;
        }

        final int height = int_image.length / width;

        ColorModel cm = ColorModel.getRGBdefault();
        DataBuffer buffer = new DataBufferInt(int_image, int_image.length);
        WritableRaster raster = Raster.createWritableRaster(cm.createCompatibleSampleModel(width, height), buffer, null);
        return new BufferedImage(cm, raster, false, null);
    }

    public static int[] makeColors(int count)
    {
        int current;
        int[] rainbow = new int[count];
        double step = 1530.0 / count;

        for (int i = 0; i < count; i++)
        {
            current = (int) (i * step);

            int r = (current < 511 ? 255 :
                    (current < 766 ? 765 - current :
                    (current < 1021 ? 0 :
                    (current < 1276 ? current - 1020 :
                        255))));
            int g = (current < 511 ? current/2 :
                    (current < 766 ? 255 :
                    (current < 1021 ? 1020 - current :
                        0)));
            int b = (current < 766 ? 0 :
                    (current < 1021 ? current - 765 :
                    (current < 1276 ? 255 :
                        1530 - current)));

            rainbow[i] = 0xFF000000 | (r << 16) | (g << 8) | b;
        }

        return rainbow;
    }

    public static int[] makeGreyLevels(int count)
    {
        int[] grey = new int[count];
        double step = 200.0 / count;

        for (int i = 0; i < count; i++)
        {
            grey[i] = 55 + (int) ((i + 1) * step);
        }

        return grey;
    }
}
