package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2014 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */


import java.awt.BasicStroke;
import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.InputEvent;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.event.MouseWheelEvent;
import java.awt.event.MouseWheelListener;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.swing.Box;
import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSpinner;
import javax.swing.JToggleButton;
import javax.swing.JToolBar;
import javax.swing.JViewport;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

import core.AbstractFilament;
import core.Calc;
import core.Const;
import core.CurvedFilament;
import core.DataImage;
import core.DataParameters;
import core.FilamentChain;
import core.SingleFilament;


class PanelImage extends JPanel implements ActionListener, ChangeListener
{
    private JScrollPane m_scroll_pane;
    private ImageComponent m_image_component;

    private JToolBar m_toolbar;
    private JButton m_button_zoom1, m_button_width_zoom1;
    private JToggleButton m_button_draw_filaments, m_button_show_filaments;
    private JComboBox<String> m_spinbox_image;
    private Map<Integer, JToggleButton> m_orientation_field_buttons;
    private JSpinner m_spinner_zoom, m_spinner_line_width;
    private JLabel m_orientation_field_label;
    private DataImage m_data;
    private BufferedImage m_image, m_image_filaments;

    PanelImage()
    {
        super();
        m_data = null;
        m_image = null;
        m_image_filaments = null;

        m_toolbar = new JToolBar("Image Panel Tools");
        m_toolbar.setLayout(new BoxLayout(m_toolbar, BoxLayout.X_AXIS));
        m_toolbar.setPreferredSize(new Dimension(5000, 30));
        m_toolbar.setFloatable(false);
        m_toolbar.setRollover(true);
        m_toolbar.setVisible(false);

        m_button_draw_filaments = new JToggleButton(IconCollection.EDIT);
        m_button_draw_filaments.addActionListener(this);
        m_toolbar.add(m_button_draw_filaments);

        m_button_show_filaments = new JToggleButton(IconCollection.HIDE, true);
        m_button_show_filaments.addActionListener(this);
        m_toolbar.add(m_button_show_filaments);

        m_toolbar.add(Box.createRigidArea(new Dimension(10, 0)));

        m_toolbar.add(new JLabel("Zoom: "));
        m_spinner_zoom = new JSpinner(new SpinnerNumberModel(100.0, 1.0, 999.0, 1.0));
        m_spinner_zoom.setMaximumSize(new Dimension(85, 30));
        m_spinner_zoom.addChangeListener(this);
        m_toolbar.add(m_spinner_zoom);

        m_button_zoom1 = new JButton(IconCollection.ZOOMORG);
        m_button_zoom1.addActionListener(this);
        m_toolbar.add(m_button_zoom1);

        m_toolbar.add(Box.createRigidArea(new Dimension(10, 0)));

        m_toolbar.add(new JLabel("Width: "));
        m_spinner_line_width = new JSpinner(new SpinnerNumberModel(100.0, 1.0, 100.0, 1.0));
        m_spinner_line_width.setMaximumSize(new Dimension(85, 30));
        m_spinner_line_width.addChangeListener(this);
        m_toolbar.add(m_spinner_line_width);

        m_button_width_zoom1 = new JButton(IconCollection.ZOOMORG);
        m_button_width_zoom1.addActionListener(this);
        m_toolbar.add(m_button_width_zoom1);

        m_toolbar.add(Box.createRigidArea(new Dimension(10, 0)));

        m_toolbar.add(new JLabel("Image: "));
        m_spinbox_image = new JComboBox<String>(new String[]{ "Original", "Preprocessed", "Binarized" });
        m_spinbox_image.setMaximumSize(m_spinbox_image.getPreferredSize());
        m_spinbox_image.setEditable(false);
        m_spinbox_image.addActionListener(this);
        m_toolbar.add(m_spinbox_image);

        m_toolbar.add(Box.createRigidArea(new Dimension(10, 0)));

        m_orientation_field_label = new JLabel("Orientation Field: ");
        m_orientation_field_buttons = new HashMap<>();

        m_image_component = new ImageComponent();
        m_scroll_pane = new JScrollPane(m_image_component);

        setLayout(new BorderLayout());
        this.add(m_toolbar, BorderLayout.PAGE_START);
        this.add(m_scroll_pane, BorderLayout.CENTER);
    }

    BufferedImage getImage()
    {
        return m_image;
    }

    void setData(DataImage data) throws IOException
    {
        m_data = data;
        m_image_filaments = m_data.makeImage(DataImage.ORIG);
    }

    void update(int view) throws IOException
    {
        if (m_data == null)
        {
            return;
        }

        switch (view)
        {
        case 0:
            int type = (m_data.getParameter(DataParameters.Pre.preview) != 0 ? DataImage.PRE
                                                                             : DataImage.ORIG);
            m_image = m_data.makeImage(type);
            setFilamentDraw(false);
            break;
        case 1:
            m_image = m_data.makeImage(DataImage.BIN);
            setFilamentDraw(false);
            break;
        case 2:
            m_image = m_image_filaments;
            fieldButtons(m_data.countOrientationFields());
            setFilamentDraw(true);
        default:
            return;
        }

        m_image_component.update();
    }

    void setImage(BufferedImage bi, boolean reset)
    {
        setFilamentDraw(false);
        m_image = bi;

        if (reset)
        {
            resetView();
        }
        else
        {
            m_image_component.update();
        }
    }

    void showFilaments(boolean show)
    {
        m_image_component.draw_filaments = show;
        m_image_component.repaint();
    }

    void showField(int n)
    {
        m_image_component.field = n;
    }

    void resetView()
    {
        m_image_component.setScale(1.0);
    }

    void zoomIn()
    {
        m_image_component.zoomIn();
    }

    void zoomOut()
    {
        m_image_component.zoomOut();
    }

    private void setFilamentDraw(boolean enabled)
    {
        m_toolbar.setVisible(enabled);
        m_image_component.draw_filaments = enabled;
        m_button_show_filaments.setSelected(enabled);
    }

    private void fieldButtons(int n)
    {
        for (JToggleButton button : m_orientation_field_buttons.values())
        {
            m_toolbar.remove(button);
        }
        m_orientation_field_buttons = new HashMap<>();
        m_toolbar.remove(m_orientation_field_label);
        if (n < 1)
        {
            m_toolbar.validate();
            return;
        }
        m_toolbar.add(m_orientation_field_label);
        for (int i = 0; i < n; i++)
        {
            JToggleButton button = new JToggleButton("" + i);
            button.addActionListener(this);
            button.setActionCommand("field" + i);
            m_orientation_field_buttons.put(i, button);
            m_toolbar.add(button);
        }
        m_toolbar.validate();
    }

    @Override
    public void actionPerformed(ActionEvent e)
    {
        Object src = e.getSource();
        String com = e.getActionCommand();

        if (src == m_button_show_filaments)
        {
            showField(-1);
            for (JToggleButton button : m_orientation_field_buttons.values())
            {
                button.setSelected(false);
            }
            showFilaments(m_button_show_filaments.isSelected());
            if (!m_button_show_filaments.isSelected())
            {
                m_button_draw_filaments.setSelected(false);
                m_image_component.dragging = true;
            }
        }
        else if (src == m_button_draw_filaments)
        {
            showField(-1);
            for (JToggleButton button : m_orientation_field_buttons.values())
            {
                button.setSelected(false);
            }
            m_image_component.dragging = !m_button_draw_filaments.isSelected();
            if (m_button_draw_filaments.isSelected())
            {
                m_button_show_filaments.setSelected(true);
                showFilaments(true);
            }
        }
        else if (src == m_button_zoom1)
        {
            m_image_component.setScale(1.0);
        }
        else if (src == m_button_width_zoom1)
        {
            m_spinner_line_width.setValue(new Double(100.0));
            m_image_component.m_width_scale = 1.0;
        }
        else if (src == m_spinbox_image)
        {
            try
            {
                switch (m_spinbox_image.getSelectedIndex())
                {
                case 0:
                    m_image_component.fil_color = Color.orange;
                    m_image = m_image_filaments = m_data.makeImage(DataImage.ORIG);
                    break;
                case 1:
                    m_data.updatePreprocessing();
                    m_image_component.fil_color = Color.orange;
                    m_image = m_image_filaments = m_data.makeImage(DataImage.PRE);
                    break;
                case 2:
                    m_data.updateBinarization();
                    m_image_component.fil_color = new Color(255, 140, 0); // dark orange
                    m_image = m_image_filaments = m_data.makeImage(DataImage.BIN);
                    break;
                default:
                    m_image_component.fil_color = Color.orange;
                    m_image = m_image_filaments = m_data.makeImage(DataImage.ORIG);
                    break;
                }
            }
            catch (IOException ex)
            {
                JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + m_data.getImageName(),
                                              "Image skipped!", JOptionPane.WARNING_MESSAGE, null);
            }
        }
        else if (com.startsWith("field"))
        {
            int n = Integer.parseInt(com.substring(5));
            for (JToggleButton button : m_orientation_field_buttons.values())
            {
                button.setSelected(false);
            }
            m_orientation_field_buttons.get(n).setSelected(true);
            showFilaments(true);
            showField(n);
            m_data.clearSelection();
            m_button_draw_filaments.setSelected(false);
            m_button_show_filaments.setSelected(true);
            m_image_component.dragging = true;
        }

        m_image_component.update();
    }

    @Override
    public void stateChanged(ChangeEvent e)
    {
        Object src = e.getSource();

        if (src == m_spinner_zoom)
        {
            m_image_component.setScale((Double) m_spinner_zoom.getValue() / 100.0);
        }
        else if (src == m_spinner_line_width)
        {
            m_image_component.m_width_scale = (Double) m_spinner_line_width.getValue() / 100.0;
        }
        m_image_component.setScale((Double) m_spinner_zoom.getValue() / 100.0);
        
    }

    /**
     * Internal class providing the surface to display image data, handle mouse
     * input and draw the filaments.
     */
    private class ImageComponent extends JLabel implements MouseListener, MouseWheelListener, MouseMotionListener,
            KeyListener
    {
        private boolean scaling, dragging, draw_filaments;
        private int field;

        private Color fil_color, sel_color;

        private Point m_last_mouse_click, m_last_mouse_press, m_last_mouse_drag, m_last_mouse_move;

        private double m_scale, m_width_scale;

        private static final double minScale = 0.03125, maxScale = 30.0;
        private static final double zoomInVelocity = 1.0 / 0.95 - 1, zoomOutVelocity = 0.05;

        private ImageComponent()
        {
            super();
            scaling = dragging = true;
            draw_filaments = false;
            m_last_mouse_click = m_last_mouse_press = m_last_mouse_drag = m_last_mouse_move = new Point(-1,-1);
            m_scale = 1.0;
            m_width_scale = 1.0;
            field = -1;

            fil_color = Color.orange;
            sel_color = Color.cyan;

            setFocusable(true);
            addMouseListener(this);
            addMouseWheelListener(this);
            addMouseMotionListener(this);
            addKeyListener(this);
        }

        @Override
        public Dimension getPreferredSize()
        {
            if (m_image == null)
            {
                return super.getPreferredSize();
            }
            return getScaledDimension();
        }

        private int getScaledWidth()
        {
            return (int) Math.ceil(m_scale * m_image.getWidth(this));
        }

        private int getScaledHeight()
        {
            return (int) Math.ceil(m_scale * m_image.getHeight(this));
        }

        private Dimension getScaledDimension()
        {
            return new Dimension(getScaledWidth(), getScaledHeight());
        }

        private int getScaledCoord(int c)
        {
            return (int)Math.round(m_scale * c); // TODO: + (scale > 1 ? scale / 2.1 : 0)
        }

        private Point unscale(Point scaled)
        {
            return new Point((int) (scaled.x / m_scale), (int) (scaled.y / m_scale));
        }

        /**
         * This method handles scaling of the image (e.g. by mouse wheel).
         *
         * @param scale The scale to which the image should be set.
         */
        private void setScale(double scale)
        {
            if (scale < minScale)
            {
                scale = minScale;
            }
            if (scale > maxScale)
            {
                scale = maxScale;
            }

            Rectangle view = m_scroll_pane.getViewport().getViewRect();

            double mouse_location_unscaled_x = m_last_mouse_move.x / m_scale;
            double mouse_location_unscaled_y = m_last_mouse_move.y / m_scale;

            double mouse_location_in_view_x = (m_last_mouse_move.x - view.x) / (double) view.width;
            double mouse_location_in_view_y = (m_last_mouse_move.y - view.y) / (double) view.height;

            m_spinner_zoom.setValue(new Double(scale * 100.0));

            m_scale = scale;
            this.update();

            if (!m_scroll_pane.getVerticalScrollBar().isVisible() && !m_scroll_pane.getHorizontalScrollBar().isVisible())
            {
                return;
            }

            view = m_scroll_pane.getViewport().getViewRect();

            view.x = Math.round((float) (mouse_location_unscaled_x * scale - mouse_location_in_view_x * view.width));
            view.y = Math.round((float) (mouse_location_unscaled_y * scale - mouse_location_in_view_y * view.height));

            if (view.x + view.width > getScaledWidth())
            {
                view.x = getScaledWidth() - view.width;
            }
            if (view.x < 0)
            {
                view.x = 0;
            }

            if (view.y + view.height > getScaledHeight())
            {
                view.y = getScaledHeight() - view.height;
            }
            if (view.y < 0)
            {
                view.y = 0;
            }

            m_scroll_pane.getViewport().setViewPosition(new Point(view.x, view.y));
        }

        private void update()
        {
            m_scroll_pane.getViewport().setView(m_image_component);
            this.repaint();
        }

        /**
         * This method paints the image panel. It is called e.g. by update().
         */
        @Override
        public void paint(Graphics graphics)
        {
            if (m_image == null)
            {
                return;
            }
            if (m_data != null && field > m_data.countOrientationFields() - 1)
            {
                field = -1;
            }

            graphics.drawImage(m_image, 0, 0, getScaledWidth(), getScaledHeight(), 0, 0,
                               m_image.getWidth(), m_image.getHeight(), this);

            if (!draw_filaments || m_data == null)
            {
                graphics.dispose();
                return;
            }

            Graphics2D graphics2d = (Graphics2D) graphics;

            graphics2d.setColor(fil_color);

            // Draw filaments.
            if (field > -1 && m_data != null)
            {
                drawFilaments(m_data.getFilamentsByOrientationField(field), graphics2d);
            }
            else
            {
                drawFilaments(m_data.getFilaments(), graphics2d);
            }            

            graphics2d.setColor(sel_color);

            // Draw selection.
            drawFilaments(m_data.getSelectedFilaments(), graphics2d);

            // Draw manually drawn filaments.
            if (m_last_mouse_drag.x != -1 && m_last_mouse_drag.y != -1)
            {
                graphics2d.setStroke(new BasicStroke());
                graphics2d.drawLine(m_last_mouse_press.x, m_last_mouse_press.y, m_last_mouse_drag.x, m_last_mouse_drag.y);
            }

            graphics2d.dispose();
        }

        /*private void drawFilaments (List<Filament> list, Graphics2D graphics2d)
        {
            for (Filament fil : list)
            {
                graphics2d.setStroke(new BasicStroke((float) (scale_width ? fil.getWidth() * m_scale / Const.MF :
                                                              fil.getWidth() / Const.MF), BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
                graphics2d.drawLine(getScaledCoord(fil.getHead().x), getScaledCoord(fil.getHead().y),
                                    getScaledCoord(fil.getTail().x), getScaledCoord(fil.getTail().y));
            }
        }*/

        private void drawFilaments (List<SingleFilament> list, Graphics2D graphics2d)
        {
            for (SingleFilament fil : list)
            {
                graphics2d.setStroke(new BasicStroke((float) (fil.getWidth() * m_scale * m_width_scale / Const.MF),
                                                     BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND));
                List<Point> points = fil.getPoints();
                for (int i = 0; i < points.size() - 1; i++)
                {
                    graphics2d.drawLine(getScaledCoord(points.get(i).x), getScaledCoord(points.get(i).y),
                                        getScaledCoord(points.get(i+1).x), getScaledCoord(points.get(i+1).y));
                }
            }
        }

        private void zoomIn()
        {
            if (scaling)
            {
                setScale(m_scale + zoomInVelocity * m_scale);
            }
        }

        private void zoomOut()
        {
            if (scaling)
            {
                setScale(m_scale - zoomOutVelocity * m_scale);
            }
        }

        @Override
        public void mouseClicked(MouseEvent e)
        {
            // TODO: Use shift modifier to draw curved filaments.
            m_last_mouse_click = e.getPoint();

            if (draw_filaments && (e.getModifiers() & InputEvent.BUTTON1_MASK) != 0) // !dragging &&
            {
                Point click = new Point( (int)Math.round(10 * m_last_mouse_click.x / m_scale),
                                         (int)Math.round(10 * m_last_mouse_click.y / m_scale));
                FilamentSensor.debugMessage("Point (" + click.x/10 + ", " + click.y/10 + ") clicked.");
                long minimal_distance = Long.MAX_VALUE;
                AbstractFilament selected = null;
                List<SingleFilament> list = (field == -1 ? m_data.getFilaments() :
                                             m_data.getFilamentsByOrientationField(field));

                // find the nearest center of a visible filament to where the user clicked
                for (SingleFilament filament : list)
                {
                    long distance = Calc.distanceM(click, filament.getCenter());
                    if (distance < minimal_distance)
                    {
                        minimal_distance = distance;
                        selected = m_data.wrap(filament);
                    }
                }

                // did the user hold CTRL?
                boolean ctrl = ((e.getModifiers() & InputEvent.CTRL_MASK) != 0);
 
                if (!ctrl)
                {
                    m_data.clearSelection();
                }

                if (minimal_distance < 500 * Const.M && selected != null)
                {
                    List<AbstractFilament> selection = m_data.getSelection();

                    // when CTRL is held down, add/remove nearest filament within range to/from selection
                    if (ctrl)
                    {
                        FilamentSensor.debugMessage("Control modifier applied");
                        if (selection.contains(selected))
                        {
                            selection.remove(selected);
                        }
                        else
                        {
                            selection.add(selected);
                        }
                    }
                    else
                    {
                        // without CTRL, just select the clicked filament normally, or select nothing
                        // if the click was not in proximity of valid filaments
                        selection.add(selected);
                    }
                    m_data.setSelection(selection);

                    this.repaint();

                    FilamentSensor.debugMessage("Filament de-/selected at distance " + (minimal_distance / Const.M));
                }
                else if (!ctrl)
                {
                    FilamentSensor.debugMessage("Selection cleared.");
                }
            }
        }

        @Override
        public void mouseEntered(MouseEvent e)
        {
        }

        @Override
        public void mouseExited(MouseEvent e)
        {
        }

        @Override
        public void mousePressed(MouseEvent e)
        {
            m_last_mouse_press = e.getPoint();

            if (!dragging)
            {
                m_data.clearSelection();
            }

            this.requestFocusInWindow();
        }

        /**
         * This method adds manually drawn filaments to the data object.
         */
        @Override
        public void mouseReleased(MouseEvent e)
        {
            if (e.getButton() == MouseEvent.BUTTON1 && m_last_mouse_drag.x != -1)
            {
                if (!dragging)
                {
                    List<Point> points = new ArrayList<>();
                    points.add(unscale(m_last_mouse_press));
                    points.add(unscale(m_last_mouse_drag));
                    CurvedFilament fil = new CurvedFilament(points);
                    m_data.addFilamentChain(new FilamentChain(fil), true);
                    this.requestFocusInWindow();
                    this.repaint();
                }

                m_last_mouse_drag = new Point(-1, -1);
            }
        }

        @Override
        public void mouseWheelMoved(MouseWheelEvent e)
        {
            m_last_mouse_move = e.getPoint();

            if (scaling)
            {
                for (int i = 0; i < e.getWheelRotation(); i++)
                {
                    zoomOut();
                }
                for (int i = 0; i > e.getWheelRotation(); i--)
                {
                    zoomIn();
                }
            }
        }

        @Override
        public void mouseMoved(MouseEvent e)
        {
            m_last_mouse_move = e.getPoint();
        }

        /**
         * This method handles drag and pan as well as the line preview when
         * manually drawing filaments.
         */
        @Override
        public void mouseDragged(MouseEvent e)
        {
            if (!draw_filaments || dragging)
            {
                JViewport jvp_scroll = m_scroll_pane.getViewport();
                Rectangle rect = jvp_scroll.getViewRect();

                if (m_scroll_pane.getHorizontalScrollBar().isVisible())
                {
                    rect.x += m_last_mouse_press.x - e.getX();
                    if (rect.x < 0)
                    {
                        rect.x = 0;
                    }
                    if (rect.x + rect.width > getScaledWidth())
                    {
                        rect.x = getScaledWidth() - rect.width;
                    }
                }
                if (m_scroll_pane.getVerticalScrollBar().isVisible())
                {
                    rect.y += m_last_mouse_press.y - e.getY();
                    if (rect.y < 0)
                    {
                        rect.y = 0;
                    }
                    if (rect.y + rect.height > getScaledHeight())
                    {
                        rect.y = getScaledHeight() - rect.height;
                    }
                }

                jvp_scroll.setViewPosition(new Point(rect.x, rect.y));
            }
            else if ((e.getModifiers() & InputEvent.BUTTON1_MASK) != 0)
            {
                m_last_mouse_drag = e.getPoint();

                // draw preview of new filament
                this.repaint();
            }

        }

        @Override
        public void keyTyped(KeyEvent e)
        {
        }

        @Override
        public void keyPressed(KeyEvent e)
        {
        }

        @Override
        public void keyReleased(KeyEvent e)
        {
            switch (e.getKeyCode())
            {
            case KeyEvent.VK_ESCAPE:
                m_last_mouse_press = new Point(m_last_mouse_drag);
                m_last_mouse_drag = new Point(-1, -1);

                this.repaint();
                break;
            case KeyEvent.VK_DELETE:
                FilamentSensor.debugMessage("DEL pressed, dispatching delete event.");
                dispatchEvent(new KeyEvent(this, KeyEvent.KEY_PRESSED, System.currentTimeMillis(), 0,
                        KeyEvent.VK_DELETE, KeyEvent.CHAR_UNDEFINED));
                break;
            case KeyEvent.VK_CONTROL:
                break;
            default:
                FilamentSensor.debugMessage("Unhandled KeyCode: " + e.getKeyCode());
            }

        }
    }
}
