package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2011-2013 Julian Rüger
 *               2013-2014 Benjamin Eltzner
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import javax.swing.table.AbstractTableModel;

import core.AbstractFilament;
import core.Const;
import core.DataImage;

public class TableModelFilaments extends AbstractTableModel
{
    private static final String[] column_names =
    { "Keep", "Center X", "Center Y", "Length", "Angle", "Width" };
    // TODO: Check whether it is sensible to use only filament data here.
    private DataImage data;

    int scroll_min = Integer.MAX_VALUE, scroll_max = 0;

    void setData(DataImage data)
    {
        this.data = data;
    }

    @Override
    public int getColumnCount()
    {
        return column_names.length;
    }

    @Override
    public String getColumnName(int col)
    {
        return column_names[col];
    }

    @Override
    public int getRowCount()
    {
        return (data == null) ? 0 : data.countRows();
    }

    @Override
    public Object getValueAt(int row, int col)
    {
        AbstractFilament fil = data.getEntryInRow(row);

        if (fil == null)
        {
            FilamentSensor.debugMessage("Tried to access row=" + row);

            return null;
        }

        switch (col)
        {
        case 0:
            return fil.getKeep();
        case 1:
            return fil.getCenter().x / 10.0;
        case 2:
            return fil.getCenter().y / 10.0;
        case 3:
            return Math.round(fil.getLength() / 1000.0) / 1000.0;
        case 4:
            return Math.round(fil.getOrientation() / 1000.0) / 1000.0;
        case 5:
            return Math.round(fil.getWidth() / 1000.0) / 1000.0;
        default:
            FilamentSensor.debugMessage("Tried to access col=" + col);
            return null;
        }
    }

    @Override
    public Class<?> getColumnClass(int col)
    {
        if (getRowCount() == 0)
        {
            return Object.class;
        }

        return getValueAt(0, col).getClass();
    }

    @Override
    public boolean isCellEditable(int row, int col)
    {
        AbstractFilament fil = data.getEntryInRow(row);

        if (fil == null)
        {
            FilamentSensor.debugMessage("Checking row = " + row);

            return false;
        }

        switch (col)
        {
        case 0:
        case 5:
            return true;
        default:
            FilamentSensor.debugMessage("Checking col = " + col);
            return false;
        }
    }

    @Override
    public void setValueAt(Object value, int row, int col)
    {
        if (!isCellEditable(row, col))
        {
            return;
        }

        switch (col)
        {
        case 0:
            if (value instanceof Boolean)
            {
                data.changeEntryInRow(row).setKeep(((Boolean) value).booleanValue());
            }
            fireTableDataChanged();
            return;
        case 5:
            if (value instanceof Double)
            {
                AbstractFilament before = data.getEntryInRow(row);
                AbstractFilament after = (AbstractFilament) before.clone();
                after.setWidth((long) (((Double) value).doubleValue() * Const.MF));
                data.modifyAbstractFilament(before, after);
            }
            fireTableDataChanged();
            return;
        default:
            FilamentSensor.debugMessage("Tried to access col=" + col);
        }
    }

}
