package ui;
/*
 * This file is part of FilamentSensor - A tool for filament tracking from cell images
 *
 * Copyright (C) 2013 Patricia Burger
 *
 * FilamentSensor is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA
 * or see <http://www.gnu.org/licenses/>.
 */

import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.GridLayout;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
//import java.awt.geom.AffineTransform;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.BorderFactory;
import javax.swing.ImageIcon;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.ScrollPaneConstants;
import javax.swing.SwingConstants;

import core.Const;
import core.DataProject;

public class ThumbnailView extends AbstractTab implements MouseListener
{
    private IOProjectData m_data;
    // ThumbnailView Elements
    public JScrollPane scroller;
    private JPanel thumbnail_panel;
    private List<JLabel> thumbnail_list;

    public ThumbnailView(TabBar bar, IOProjectData data)
    {

        // ////////Tab labels (project data) //////////
        super(new FlowLayout(FlowLayout.LEFT));
        m_tab_bar = bar;
        thumbnail_list = new ArrayList<>();

        this.add(new JLabel(IconCollection.THUMBNAIL));
        m_tab_label = new JLabel();
        m_tab_label.setBorder(BorderFactory.createEmptyBorder(1, 5, 0, 5));
        this.add(m_tab_label);

        setBorder(BorderFactory.createEmptyBorder(3, 3, 0, -1));
        setOpaque(false);

        // Add a scrollable panel
        thumbnail_panel = new JPanel(new GridLayout(0, 3, 10, 10));
        scroller = new JScrollPane(thumbnail_panel, ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED,
                                   ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);
        scroller.getVerticalScrollBar().setUnitIncrement(Const.SCROLL);
        scroller.setPreferredSize(new Dimension(800, 600));

        this.add(scroller, FlowLayout.CENTER);

        reset(data);
    }

    public boolean reset(IOProjectData data)
    {

        // ////////Tab labels (project data) //////////
        m_data = data;
        m_tab_title = m_data.getName();
        m_tab_label.setText(m_data.getName());

        // //////////////ThumbnailView ////////////////

        thumbnail_panel.removeAll();
        for (JLabel label : thumbnail_list)
        {
            label.removeAll();
        }
        thumbnail_list = new ArrayList<>();

        File folder = m_data.getImagePath();

        for (int i = 0; i < m_data.getLength(); i++)
        {
            File file = m_data.getImageFile(i);
            String full_path = folder.getAbsolutePath() + File.separator + file.getName();
            ImageIcon icon = makeThumbnail(new File(full_path));
            if (icon != null)
            {
                JLabel label = new JLabel(icon);
                label.setText(file.getName());
                label.setName(folder + File.separator + file.getName());
                label.setHorizontalTextPosition(SwingConstants.CENTER);
                label.setVerticalTextPosition(SwingConstants.BOTTOM);
                label.addMouseListener(this);
                thumbnail_list.add(label);
                thumbnail_panel.add(label);
                thumbnail_panel.revalidate();
            }
        }

        return true;
    }

    public boolean saveData()
    {
        return m_data.saveProject();
    }

    private static ImageIcon makeThumbnail(File file)
    {
        BufferedImage image = null;
        
        try
        {
            image = ImageFileChooser.readImage(file);
        }
        catch (IOException e)
        {
            System.out.println(e.getMessage());
            System.out.println("Not Created: " + file.getAbsolutePath());
        }

        if (image == null)
        {
            System.out.println("Image created from " + file.getAbsolutePath() + " is null.");
            return null;
        }
        
        int height = Const.THUMBNAIL_SIZE,
            width = (int) (image.getWidth() * height / (double) image.getHeight());
        if (image.getWidth() > image.getHeight())
        {
            width = Const.THUMBNAIL_SIZE;
            height = (int) (image.getHeight() * width / (double) image.getWidth());
        }

        BufferedImage thumbnail_image = makeCanvas(width, height);
        Graphics2D canvas = thumbnail_image.createGraphics();

        canvas.drawImage(image, 0, 0, width, height, null);
        canvas.dispose();

        return new ImageIcon(thumbnail_image);
    }

    private static BufferedImage makeCanvas(int width, int height)
    {
        return GraphicsEnvironment.getLocalGraphicsEnvironment().getDefaultScreenDevice()
                                  .getDefaultConfiguration().createCompatibleImage(width, height);
    }

    @Override
    public boolean readyToClose(boolean quiet)
    {
        for (JLabel label : thumbnail_list)
        {
            label.setIcon(null);
        }
        thumbnail_panel.revalidate();
        thumbnail_panel.repaint();
        thumbnail_panel.removeAll();
        thumbnail_list.clear();
        if (quiet)
        {
            return true;
        }
        if (m_data != null)
        {
            switch (JOptionPane.showConfirmDialog(MainWindow.frame(), "Do you want to save your changes?", m_tab_title,
                                                  JOptionPane.YES_NO_CANCEL_OPTION, JOptionPane.PLAIN_MESSAGE,
                                                  IconCollection.CLOSE16))
            {
            case JOptionPane.YES_OPTION:
                if (!saveData())
                {
                    return false;
                }
                // fall through
            case JOptionPane.NO_OPTION:
                return true;
            case JOptionPane.CANCEL_OPTION:
                return false;
            }
        }
        return true;
    }

    @Override
    public void mouseClicked(MouseEvent e)
    {
        String name = e.getComponent().getName();

        if (e.getClickCount() == 2 && name != null)
        {
            FilamentSensor.debugMessage("Label " + name + " was double clicked. Opening PickingView.");

            File image_file = new File(name);
            File data_file = m_data.getDataForImage(image_file);

            if (!m_data.contains(image_file))
            {
                System.out.println("File not contained in project data: " + name);
            }
            else
            {
                try
                {
                    m_tab_bar.newPickingView(image_file, data_file, m_data.getImagePath().getName(),
                                             DataProject.getDefaultParameters(), DataProject.getForce());
                }
                catch (IOException ex)
                {
                    JOptionPane.showMessageDialog(MainWindow.frame(), "This image could not be loaded: " + image_file,
                                                  "Invalid image!", JOptionPane.WARNING_MESSAGE, null);
                }
            }
        }
        else
        {
            FilamentSensor.debugMessage("Name of clicked label: " + name);
        }
    }

    @Override
    public void mousePressed(MouseEvent e)
    {
    }

    @Override
    public void mouseReleased(MouseEvent e)
    {
    }

    @Override
    public void mouseEntered(MouseEvent e)
    {
    }

    @Override
    public void mouseExited(MouseEvent e)
    {
    }

}
